
    (*********************************************************************
     *  MonkeyDecoder.pas                                                *
     *                                                                   *
     *  author    : Milenko Mitrovic                                     *
     *  email     : dcoder@dsp-worx.de                                   *
     *  web       : http://dsp-worx.de                                   *
     *  date      : 20-08-2003                                           *
     *  version   : 1.00                                                 *
     *                                                                   *
     *  The contents of this file are used with permission, subject to   *
     *  the Mozilla Public License Version 1.1 (the "License"); you may  *
     *  not use this file except in compliance with the License. You may *
     *  obtain a copy of the License at                                  *
     *  http://www.mozilla.org/MPL/MPL-1.1.html                          *
     *                                                                   *
     *  Software distributed under the License is distributed on an      *
     *  "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or   *
     *  implied. See the License for the specific language governing     *
     *  rights and limitations under the License.                        *
     *                                                                   *
     *  (C) 2003 Milenko Mitrovic <dcoder@dsp-worx.de>                   *
     *                                                                   *
     *********************************************************************)

unit MonkeyDecoder;

interface

uses
  Windows, SysUtils, Dialogs, waUtils, dmoConst;

type
  TAPE_DECOMPRESS_FIELDS = (
    APE_INFO_SAMPLE_RATE = 1003,
    APE_INFO_BITS_PER_SAMPLE = 1004,
    APE_INFO_CHANNELS = 1006,
    APE_DECOMPRESS_TOTAL_BLOCKS = 2002
  );

  TMonkeyDecoder = class
  private
    c_APEDecompress_Create : function (Filename : PWChar; var ErrorCode : Integer) : Pointer; stdcall;
    c_APEDecompress_GetData : function (hAPEDecompress : Pointer; Buffer : Pointer; nBlocks: Integer; var nBlocksRetrieved: Integer): integer; stdcall;
    c_APEDecompress_GetInfo : function (hAPEDecompress : Pointer; Field : integer; nParam1 : integer = 0; nParam2 : integer = 0): integer; stdcall;
    c_APEDecompress_Seek : function (hAPEDecompress : Pointer; nBlockOffset : integer): integer; stdcall;
    c_APEDecompress_Destroy : procedure (hAPEDecompress : Pointer); stdcall;
    c_APEDecompress_GetTag : function (hAPEDecompress : Pointer; tagfield: PWideChar; tagname: PWideChar): integer; stdcall;
    fLib : hWnd;

    fDuration : Int64;
    fBitsPerSample : integer;
    fBytesPerSample : integer;
    fFrequency : integer;
    fChannels : integer;
    fBlockSize : integer;
    fDecoder : Pointer;

    function GetInfo(InfoType : TAPE_DECOMPRESS_FIELDS) : integer;
    procedure GetStreamInfos;
    procedure SetPosition(Pos : Int64);
    function InitMacDLL : Boolean;
    procedure FreeMacDLL;

    function GetApeTag(FieldName: WideString): WideString;
  public
    constructor Create(Filename : PWCHAR; out hr : HRESULT);
    destructor Destroy; override;
    function GetBuffer(Buffer: Pointer; Blocks: Integer) : integer;

    function getAuthorName: WideString;
    function getTitle: WideString;
    function getRating: WideString;
    function getDescription: WideString;
    function getCopyright: WideString;
    function getBaseURL: WideString;
    function getLogoURL: WideString;
    function getLogoIconURL: WideString;
    function getWatermarkURL: WideString;
    function getMoreInfoURL: WideString;
    function getMoreInfoBannerImage: WideString;
    function getMoreInfoBannerURL: WideString;
    function getMoreInfoText: WideString;

  published
    property Duration : Int64 read fDuration;
    property BitsPerSample : integer read fBitsPerSample;
    property BytesPerSample : integer read fBytesPerSample;
    property Frequency : integer read fFrequency;
    property Channels : integer read fChannels;
    property BlockSize : integer read fBlockSize;
    property Position : Int64 write SetPosition;
  end;

implementation

function ReadString(Section, Name, Default, FileName: WideString): WideString;
var
  wBuf: array[0..MAX_PATH -1] of WideChar;
  aBuf: array[0..MAX_PATH -1] of Char;
begin
  if Win32Platform = VER_PLATFORM_WIN32_NT then
  begin
    GetPrivateProfileStringW(PWChar(Section),PWChar(Name),PWChar(Default),wBuf,MAX_PATH,PWChar(FileName));
    Result := wBuf;
  end else
  begin
    GetPrivateProfileString(PChar(String(Section)),PChar(String(Name)),PChar(String(Default)),aBuf,MAX_PATH,PChar(String(FileName)));
    Result := aBuf;
  end;
end;

function ReadInteger(Section, Name: WideString; Default: integer; FileName: WideString): integer;
begin
  if Win32Platform = VER_PLATFORM_WIN32_NT
    then Result := GetPrivateProfileIntW(PWChar(Section),PWChar(Name),Default,PWChar(FileName))
    else Result := GetPrivateProfileInt(PChar(String(Section)),PChar(String(Name)),Default,PChar(String(FileName)));
end;

// Taken from TntUnicodeControls
function StrScanW(const Str: PWideChar; Chr: WideChar): PWideChar;
begin
  Result := Str;
  while Result^ <> Chr do
  begin
    if Result^ = #0 then
    begin
      Result := nil;
      Exit;
    end;
    Inc(Result);
  end;
end;

function WideLastDelimiter(const Delimiters, S: WideString): Integer;
var
  P: PWideChar;
begin
  Result := Length(S);
  P := PWideChar(Delimiters);
  while Result > 0 do
  begin
    if (S[Result] <> #0) and (StrScanW(P, S[Result]) <> nil) then
      Exit;
    Dec(Result);
  end;
end;

function WideExtractFilePath(const FileName: WideString): WideString;
var
  I: Integer;
begin
  I := WideLastDelimiter('\:', FileName);
  Result := Copy(FileName, 1, I);
end;


constructor TMonkeyDecoder.Create(Filename : PWCHAR; out hr : HRESULT);
var
  Res : integer;
begin
  inherited Create;
  if not InitMacDLL then
  begin
    hr := E_FAIL;
    Exit;
  end;

  Res := S_OK;

  // check if its an .apl File
  fDecoder := c_APEDecompress_Create(Filename,Res);
  if (Res <> 0) or (fDecoder = nil) then
  begin
    hr := E_FAIL;
    Exit;
  end else
  begin
    hr := S_OK;
    GetStreamInfos;
  end;
end;

destructor TMonkeyDecoder.Destroy;
begin
  FreeMacDLL;
  inherited Destroy;
end;

function TMonkeyDecoder.GetInfo(InfoType : TAPE_DECOMPRESS_FIELDS) : integer;
begin
  Result := c_APEDecompress_GetInfo(fDecoder,integer(InfoType),0,0);
end;

procedure TMonkeyDecoder.GetStreamInfos;
begin
  fBitsPerSample := GetInfo(APE_INFO_BITS_PER_SAMPLE);
  fBytesPerSample := fBitsPerSample div 8;
  fChannels := GetInfo(APE_INFO_CHANNELS);
  fFrequency := GetInfo(APE_INFO_SAMPLE_RATE);
  fBlockSize := fBytesPerSample * fChannels;
  fDuration := Int64(GetInfo(APE_DECOMPRESS_TOTAL_BLOCKS)) * One_Second div fFrequency;
end;

procedure TMonkeyDecoder.SetPosition(Pos : Int64);
begin
  c_APEDecompress_Seek(fDecoder,Pos);
end;

function TMonkeyDecoder.GetBuffer(Buffer: Pointer; Blocks : Integer) : integer;
begin
  c_APEDecompress_GetData(fDecoder,Buffer,Blocks,Result);
end;

function TMonkeyDecoder.InitMacDLL : Boolean;
var
  Buf : array[0..MAX_PATH -1] of Char;
  Path : String;
begin
  Result := False;
  GetModuleFileName(HInstance,Buf,MAX_PATH);
  Path := ExtractFilePath(Buf) + 'MACDec.dll';
  if FileExists(Path) then fLib := LoadLibrary(PChar(Path))
                      else fLib := LoadLibrary('MACDec.dll');
  if fLib = 0 then Exit;
  @c_APEDecompress_Create := GetProcAddress(fLib,'c_APEDecompress_Create');
  @c_APEDecompress_GetData := GetProcAddress(fLib,'c_APEDecompress_GetData');
  @c_APEDecompress_GetInfo := GetProcAddress(fLib,'c_APEDecompress_GetInfo');
  @c_APEDecompress_Seek := GetProcAddress(fLib,'c_APEDecompress_Seek');
  @c_APEDecompress_Destroy := GetProcAddress(fLib,'c_APEDecompress_Destroy');
  @c_APEDecompress_GetTag := GetProcAddress(fLib,'c_APEDecompress_GetTag');
  Result := True;
end;

procedure TMonkeyDecoder.FreeMacDLL;
begin
  if fDecoder <> nil then c_APEDecompress_Destroy(fDecoder);
  if (fLib > 0) then
  begin
    FreeLibrary(fLib);
    fLib := 0;
    @c_APEDecompress_Create := nil;
    @c_APEDecompress_GetData := nil;
    @c_APEDecompress_GetInfo := nil;
    @c_APEDecompress_Seek := nil;
    @c_APEDecompress_Destroy := nil;
    @c_APEDecompress_GetTag := nil;
  end;
end;

function TMonkeyDecoder.GetApeTag(FieldName: WideString): WideString;
var
  tmp: array[0..1023] of WideChar;
  res: integer;
begin
  res := c_APEDecompress_GetTag(fDecoder, PWideChar(FieldName), @tmp);
  if res = -1
    then Result := ''
    else Result := tmp;
end;

function TMonkeyDecoder.getAuthorName: WideString;
begin
  Result := GetApetag('artist');
end;

function TMonkeyDecoder.getTitle: WideString;
begin
  Result := GetApetag('title');
end;

function TMonkeyDecoder.getRating: WideString;
begin
  Result := GetApetag('rating');
end;

function TMonkeyDecoder.getDescription: WideString;
begin
  Result := GetApetag('album');
end;

function TMonkeyDecoder.getCopyright: WideString;
begin
  Result := GetApetag('copyright');
end;

function TMonkeyDecoder.getBaseURL: WideString;
begin
  Result := GetApetag('artisturl');
end;

function TMonkeyDecoder.getLogoURL: WideString;
begin
  Result := GetApetag('logourl');
end;

function TMonkeyDecoder.getLogoIconURL: WideString;
begin
  Result := GetApetag('logoiconurl');
end;

function TMonkeyDecoder.getWatermarkURL: WideString;
begin
  Result := GetApetag('copyrighturl');
end;

function TMonkeyDecoder.getMoreInfoURL: WideString;
begin
  Result := GetApetag('publisherurl');
end;

function TMonkeyDecoder.getMoreInfoBannerImage: WideString;
begin
  Result := GetApetag('moreinfobannerimage');
end;

function TMonkeyDecoder.getMoreInfoBannerURL: WideString;
begin
  Result := GetApetag('moreinfobannerurl');
end;

function TMonkeyDecoder.getMoreInfoText: WideString;
begin
  Result := GetApetag('comment');
end;

end.
