{*******************************************************************}
{                                                                   }
{ MSLU API Interface Unit for Delphi 5+                             }
{ Copyright (c) 2002, CoCo Systems Ltd.                             }
{ Author: S.P.Moss, CoCo Systems Ltd. (mailto:spm@coco.co.uk)       }
{ Unit download URL:                                                }
{    http://www.coco.co.uk/pubftp/Developers/Delphi/unicows.zip     }
{                                                                   }
{                                                                   }
{ Release History:                                                  }
{ ===============                                                   }
{ 1st November, 2002:                                               }
{    - Added support for new functions exported by unicows.dll      }
{      3703: BeginUpdateResourceA, EndUpdateResourceA,              }
{      UpdateResourceA, and RasGetProjectionInfoW.                  }
{    - Added support for the remaining functions exported by        }
{      unicows.dll (IsDestinationReachable and he OleUI functions). }
{    - Renamed USEJWA define to UNIJWA.                             }
{    - Added UNIJRA to conditionally compile-in the RAS API         }
{      functions.                                                   }
{ 17th October, 2002:                                               }
{    - Minor corrections.                                           }
{ 11th October, 2002:                                               }
{    - First release                                                }
{                                                                   }
{ For unicows.dll build: 3665 or later                              }
{ NB: Unless otherwise stated, all functions here are supported by  }
{ unicows.dll build 3665 or later. Where API functions were         }
{ introduced in later builds, they are marked here with comments    }
{ such as this:                                                     }
{                                                                   }
{    // 3703                                                        }
{                                                                   }
{ CAUTION: If you attempt to call a function which is supported     }
{ only by a later builds than that which is installed at run-time   }
{ will result in an access violation!                               }
{                                                                   }
{ The contents of this file are used with permission, subject to    }
{ the Mozilla Public License Version 1.1 (the "License"); you may   }
{ not use this file except in compliance with the License. You may  }
{ obtain a copy of the License at:                                  }
{                                                                   }
{    http://www.mozilla.org/MPL/MPL-1.1.html                        }
{                                                                   }
{ Software distributed under the License is distributed on a strict }
{ "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or    }
{ implied. See the License for the specific language governing      }
{ rights and limitations under the License.                         }
{                                                                   }
{ How to use this unit:                                             }
{ ====================                                              }
{ 1. Obtain unicows.dll. At the time of writing this is available   }
{    from MSDN, at:                                                 }
{    http://www.microsoft.com/downloads/release.asp?releaseid=30039 }
{ 2. Place the dll in the same folder as your application, or       }
{    elsewhere on your PATH.                                        }
{ 3. Include this unit in your Delphi project.                      }
{ 4. Before calling functions in this unit, have your code call     }
{    IsMSLULoaded to determine whether the MSLU was loaded OK.      }
{ 5. Call functions from other units by prefixing them with this    }
{    unit's name, thus: UnicoWS.ExtTextOutW(...).                   }
{                                                                   }
{ Notes:                                                            }
{ =====                                                             }
{ 1. This unit provides an interface to all of the functions        }
{    exported by unicows.dll.                                       }
{ 2. The type declarations required to use some of the functions    }
{    available here are not defined in the Delphi run-time library, }
{    but are defined in some third party libraries, such as the     }
{    JEDI WinAPI library. In order that you can use this unit       }
{    without the need for these other libraries, the API functions  }
{    concerned are conditionally compiled here, thus:               }
{                                                                   }
{    - Define UNIJWA if you have the JEDI WinAPI library installed  }
{    - Define UNIJRA if you have the JEDI RAS API library installed }
{                                                                   }
{ 3. This version of the UnicoWS.pas unit was written to support    }
{    the unicows.dll build 3665 or later. If you have an earlier    }
{    build of unicows.dll, you are strongly advised to upgrade to   }
{    the indicated build (or a later one). If you have a later      }
{    build of unicows.dll this unit should work OK, but any         }
{    functions added to unicows.dll since the indicated build will  }
{    not be accessible via this unit.                               }
{                                                                   }
{ The author welcomes any feedback you may have (to the e-mail      }
{ address given above), and I will endeavour to include any bug-    }
{ fixes reported, or add support for further API functions that are }
{ added to new builds of unicows.dll.                               }
{                                                                   }
{*******************************************************************}

unit UnicoWS;

interface

uses
{$IFDEF UNIJWA}
   JwaSspi, JwaWinCrypt, JwaWinGDI, JwaWinNLS, JwaWinType,
{$ENDIF}
{$IFDEF UNIJRA}
   Ras,
{$ENDIF}
   Windows, ActiveX, CommCtrl, CommDlg, DDEml, MMSystem, MultiMon, OleDlg, ShellAPI,
   ShlObj, WinSpool;

const
  { Flags used in TQOCInfo.dwFlags}
  NETWORK_ALIVE_LAN = $00000001; {The computer has one or more active LAN cards}
  NETWORK_ALIVE_WAN = $00000002; {The computer has one or more active RAS connections}
  NETWORK_ALIVE_AOL = $00000004; {This flag is only valid in Windows 95 and Windows 98. The computer is connected to the America Online network}

type
  PQOCInfo = ^TQOCInfo;
  {$EXTERNALSYM tagQOCINFO}
  tagQOCINFO = record
    dwSize: Longint;
    dwFlags: Longint;
    dwInSpeed: Longint;
    dwOutSpeed: Longint;
  end;
  TQOCInfo = tagQOCINFO;

type
  POleUIBusyW = ^TOleUIBusyW;
  {$EXTERNALSYM tagOLEUIBUSYW}
  tagOLEUIBUSYW = record
    cbStruct: Longint;
    dwFlags: Longint;
    hWndOwner: HWnd;
    lpszCaption: PWideChar;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    hInstance: THandle;
    lpszTemplate: PWideChar;
    hResource: HRsrc;
    task: HTask;
    lphWndDialog: ^HWnd;
  end;
  TOleUIBusyW = tagOLEUIBUSYW;

  POleUIChangeIconW = ^TOleUIChangeIconW;
  {$EXTERNALSYM tagOLEUICHANGEICONW}
  tagOLEUICHANGEICONW = record
    cbStruct: Longint;
    dwFlags: Longint;
    hWndOwner: HWnd;
    lpszCaption: PWideChar;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    hInstance: THandle;
    lpszTemplate: PWideChar;
    hResource: HRsrc;
    hMetaPict: HGlobal;
    clsid: TCLSID;
    szIconExe: array[0..MAX_PATH - 1] of WideChar;
    cchIconExe: Integer;
  end;
  TOleUIChangeIconW = tagOLEUICHANGEICONW;

  POleUIChangeSourceW = ^TOleUIChangeSourceW;
  {$EXTERNALSYM tagOLEUICHANGESOURCEW}
  tagOLEUICHANGESOURCEW = record
    cbStruct: Longint;
    dwFlags: Longint;
    hWndOwner: HWnd;
    lpszCaption: PWideChar;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    hInstance: THandle;
    lpszTemplate: PWideChar;
    hResource: HRsrc;
    lpOFN: POpenFileNameW;
    dwReserved1: array[0..3] of Longint;
    lpOleUILinkContainer: IOleUILinkContainer;
    dwLink: Longint;
    lpszDisplayName: PWideChar;
    nFileLength: ULONG;
    lpszFrom: PWideChar;
    lpszTo: PWideChar;
  end;
  TOleUIChangeSourceW = tagOLEUICHANGESOURCEW;

  POleUIConvertW = ^TOleUIConvertW;
  {$EXTERNALSYM tagOLEUICONVERTW}
  tagOLEUICONVERTW = record
    cbStruct: Longint;
    dwFlags: Longint;
    hWndOwner: HWnd;
    lpszCaption: PWideChar;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    hInstance: THandle;
    lpszTemplate: PWideChar;
    hResource: HRsrc;
    clsid: TCLSID;
    clsidConvertDefault: TCLSID;
    clsidActivateDefault: TCLSID;
    clsidNew: TCLSID;
    dvAspect: Longint;
    wFormat: Word;
    fIsLinkedObject: BOOL;
    hMetaPict: HGlobal;
    lpszUserType: PWideChar;
    fObjectsIconChanged: BOOL;
    lpszDefLabel: PWideChar;
    cClsidExclude: Integer;
    lpClsidExclude: PCLSID;             
  end;
  TOleUIConvertW = tagOLEUICONVERTW;

  POleUIEditLinksW = ^TOleUIEditLinksW;
  {$EXTERNALSYM tagOLEUIEDITLINKSW}
  tagOLEUIEDITLINKSW = record
    cbStruct: Longint;
    dwFlags: Longint;
    hWndOwner: HWnd;
    lpszCaption: PWideChar;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    hInstance: THandle;
    lpszTemplate: PWideChar;
    hResource: HRsrc;
    OleUILinkContainer: IOleUILinkContainer; 

  end;
  TOleUIEditLinksW = tagOLEUIEDITLINKSW;

  POleUIGnrlPropsW = ^TOleUIGnrlPropsW;
  {$EXTERNALSYM tagOLEUIGNRLPROPSW}
  tagOLEUIGNRLPROPSW = record
    cbStruct: Longint;
    dwFlags: Longint;
    dwReserved1: array[1..2] of Longint;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    dwReserved2: array[1..3] of Longint;
    lpOP: Pointer;
  end;
  TOleUIGnrlPropsW = tagOLEUIGNRLPROPSW;

  POleUIInsertObjectW = ^TOleUIInsertObjectW;
  {$EXTERNALSYM tagOLEUIINSERTOBJECTW}
  tagOLEUIINSERTOBJECTW = record
    cbStruct: Longint;
    dwFlags: Longint;
    hWndOwner: HWnd;
    lpszCaption: PWideChar;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    hInstance: THandle;
    lpszTemplate: PWideChar;
    hResource: HRsrc;
    clsid: TCLSID;
    lpszFile: PChar;
    cchFile: Integer;
    cClsidExclude: Integer;
    lpClsidExclude: PCLSID;
    iid: TIID;
    oleRender: Longint;
    lpFormatEtc: PFormatEtc;
    lpIOleClientSite: IOleClientSite;
    lpIStorage: IStorage;
    ppvObj: Pointer;
    sc: HResult;
    hMetaPict: HGlobal;                 
  end;
  TOleUIInsertObjectW = tagOLEUIINSERTOBJECTW;

  POleUILinkPropsW = ^TOleUILinkPropsW;
  {$EXTERNALSYM tagOLEUILINKPROPSW}
  tagOLEUILINKPROPSW = record
    cbStruct: Longint;
    dwFlags: Longint;
    dwReserved1: array[1..2] of Longint;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    dwReserved2: array[1..3] of Longint;
    lpOP: Pointer;
  end;
  TOleUILinkPropsW = tagOLEUILINKPROPSW;

  POleUIViewPropsW = ^TOleUIViewPropsW;
  {$EXTERNALSYM tagOLEUIVIEWPROPSW}
  tagOLEUIVIEWPROPSW = record
    cbStruct: Longint;
    dwFlags: Longint;
    dwReserved1: array[1..2] of Longint;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    dwReserved2: array[1..3] of Longint;
    lpOP: Pointer;
    nScaleMin: Integer;
    nScaleMax: Integer;
  end;
  TOleUIViewPropsW = tagOLEUIVIEWPROPSW;

  POleUIObjectPropsW = ^TOleUIObjectPropsW;
  {$EXTERNALSYM tagOLEUIOBJECTPROPSW}
  tagOLEUIOBJECTPROPSW = record
    cbStruct: Longint;
    dwFlags: Longint;
    lpPS: PPropSheetHeaderW;
    dwObject: Longint;
    lpObjInfo: IOleUIObjInfo;
    dwLink: Longint;
    lpLinkInfo: IOleUILinkInfo;
    lpGP: POleUIGnrlPropsW;
    lpVP: POleUIViewPropsW;
    lpLP: POleUILinkPropsW;
  end;
  TOleUIObjectPropsW = tagOLEUIOBJECTPROPSW;

  POleUIPasteEntryW = ^TOleUIPasteEntryW;
  {$EXTERNALSYM tagOLEUIPASTEENTRYW}
  tagOLEUIPASTEENTRYW = record
    fmtetc: TFormatEtc;
    lpstrFormatName: PWideChar;
    lpstrResultText: PWideChar;
    dwFlags: Longint;
    dwScratchSpace: Longint;
  end;
  TOleUIPasteEntryW = tagOLEUIPASTEENTRYW;

  POleUIPasteSpecialW = ^TOleUIPasteSpecialW;
  {$EXTERNALSYM tagOLEUIPASTESPECIALW}
  tagOLEUIPASTESPECIALW = record
    cbStruct: Longint;
    dwFlags: Longint;
    hWndOwner: HWnd;
    lpszCaption: PWideChar;
    lpfnHook: TFNOleUIHook;
    lCustData: Longint;
    hInstance: THandle;
    lpszTemplate: PWideChar;
    hResource: HRsrc;
    lpSrcDataObj: IDataObject;
    arrPasteEntries: POleUIPasteEntryW;
    cPasteEntries: Integer;
    arrLinkTypes: PLongint;
    cLinkTypes: Integer;
    cClsidExclude: Integer;
    lpClsidExclude: PCLSID;
    nSelectedIndex: Integer;
    fLink: BOOL;
    hMetaPict: HGlobal;
    sizel: TSize;
  end;
  TOleUIPasteSpecialW = tagOLEUIPASTESPECIALW;

type

{$IFDEF UNIJWA}
  TAcquireCredentialsHandleW = function(pszPrincipal, pszPackage: PSecWChar;
    fCredentialUse: Cardinal; pvLogonId, pAuthData: Pointer;
    pGetKeyFn: SEC_GET_KEY_FN; pvGetKeyArgument: Pointer; phCredential: PCredHandle;
    var ptsExpiry: TTimeStamp): SECURITY_STATUS; stdcall;
  {$EXTERNALSYM TAcquireCredentialsHandleW}
{$ENDIF}
  TAddAtomW = function(lpString: PWideChar): ATOM; stdcall;
  {$EXTERNALSYM TAddAtomW}
  TAddFontResourceW = function(FileName: PWideChar): Integer; stdcall;
  {$EXTERNALSYM TAddFontResourceW}
  TAddJobW = function(hPrinter: THandle; Level: DWORD; pData: Pointer; cbBuf: DWORD; var pcbNeeded: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TAddJobW}
  TAddMonitorW = function(pName: PWideChar; Level: DWORD; pMonitors: Pointer): BOOL; stdcall;
  {$EXTERNALSYM TAddMonitorW}
  TAddPortW = function(pName: PWideChar; hWnd: HWND; pMonitorName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TAddPortW}
  TAddPrintProcessorW = function(pName, pEnvironment, pPathName, pPrintProcessorName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TAddPrintProcessorW}
  TAddPrintProvidorW = function(pName: PWideChar; level: DWORD; pProvidorInfo: Pointer): BOOL; stdcall;
  {$EXTERNALSYM TAddPrintProvidorW}
  TAddPrinterDriverW = function(pName: PWideChar; Level: DWORD; pDriverInfo: Pointer): BOOL; stdcall;
  {$EXTERNALSYM TAddPrinterDriverW}
  TAddPrinterW = function(pName: PWideChar; Level: DWORD; pPrinter: Pointer): THandle; stdcall;
  {$EXTERNALSYM TAddPrinterW}
  TAdvancedDocumentPropertiesW = function(hWnd: HWND; hPrinter: THandle; pDeviceName: PWideChar;
    pDevModeOutput, pDevModeInput: PDeviceModeW): Longint; stdcall;
  {$EXTERNALSYM TAdvancedDocumentPropertiesW}
  TAppendMenuW = function(hMenu: HMENU; uFlags, uIDNewItem: DWORD;
    lpNewItem: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TAppendMenuW}
  TBeginUpdateResourceA = function(pFileName: PChar; bDeleteExistingResources: BOOL): THandle; stdcall; // 3703
  {$EXTERNALSYM TBeginUpdateResourceA}
  TBeginUpdateResourceW = function(pFileName: PWideChar; bDeleteExistingResources: BOOL): THandle; stdcall;
  {$EXTERNALSYM TBeginUpdateResourceW}
  TBroadcastSystemMessageW = function(Flags: DWORD; Recipients: PDWORD;
    uiMessage: DWORD; wParam: WPARAM; lParam: LPARAM): Longint; stdcall;
  {$EXTERNALSYM TBroadcastSystemMessageW}
  TBuildCommDCBAndTimeoutsW = function(lpDef: PWideChar; var lpDCB: TDCB;
    var lpCommTimeouts: TCommTimeouts): BOOL; stdcall;
  {$EXTERNALSYM TBuildCommDCBAndTimeoutsW}
  TBuildCommDCBW = function(lpDef: PWideChar; var lpDCB: TDCB): BOOL; stdcall;
  {$EXTERNALSYM TBuildCommDCBW}
  TCallMsgFilterW = function(var lpMsg: TMsg; nCode: Integer): BOOL; stdcall;
  {$EXTERNALSYM TCallMsgFilterW}
  TCallNamedPipeW = function(lpNamedPipeName: PWideChar; lpInBuffer: Pointer;
    nInBufferSize: DWORD; lpOutBuffer: Pointer; nOutBufferSize: DWORD;
    var lpBytesRead: DWORD; nTimeOut: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCallNamedPipeW}
  TCallWindowProcA = function(lpPrevWndFunc: TFNWndProc; hWnd: HWND; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  {$EXTERNALSYM TCallWindowProcA}
  TCallWindowProcW = function(lpPrevWndFunc: TFNWndProc; hWnd: HWND; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  {$EXTERNALSYM TCallWindowProcW}
  TChangeDisplaySettingsExW = function(lpszDeviceName: PWideChar; var lpDevMode: TDeviceModeW;
        wnd: HWND; dwFlags: DWORD; lParam: Pointer): Longint; stdcall;
  {$EXTERNALSYM TChangeDisplaySettingsExW}
  TChangeDisplaySettingsW = function(var lpDevMode: TDeviceModeW; dwFlags: DWORD): Longint; stdcall;
  {$EXTERNALSYM TChangeDisplaySettingsW}
  TChangeMenuW = function(hMenu: HMENU; cmd: DWORD; lpszNewItem: PWideChar;
    cmdInsert: DWORD; flags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TChangeMenuW}
  TCharLowerBuffW = function(lpsz: PWideChar; cchLength: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TCharLowerBuffW}
  TCharLowerW = function(lpsz: PWideChar): PWideChar; stdcall;
  {$EXTERNALSYM TCharLowerW}
  TCharNextW = function(lpsz: PWideChar): PWideChar; stdcall;
  {$EXTERNALSYM TCharNextW}
  TCharPrevW = function(lpszStart: PWideChar; lpszCurrent: PWideChar): PWideChar; stdcall;
  {$EXTERNALSYM TCharPrevW}
  TCharToOemBuffW = function(lpszSrc: PWideChar; lpszDst: PWideChar; cchDstLength: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCharToOemBuffW}
  TCharToOemW = function(lpszSrc: PWideChar; lpszDst: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TCharToOemW}
  TCharUpperBuffW = function(lpsz: PWideChar; cchLength: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TCharUpperBuffW}
  TCharUpperW = function(lpsz: PWideChar): PWideChar; stdcall;
  {$EXTERNALSYM TCharUpperW}
  TChooseColorWFunc = function(var CC: TChooseColorW): BOOL; stdcall;
  {$EXTERNALSYM TChooseColorWFunc}
  TChooseFontWFunc = function(var ChooseFont: TChooseFontW): BOOL; stdcall;
  {$EXTERNALSYM TChooseFontWFunc}
  TCommConfigDialogW = function(lpszName: PWideChar; hWnd: HWND; var lpCC: TCommConfig): BOOL; stdcall;
  {$EXTERNALSYM TCommConfigDialogW}
  TCompareStringW = function(Locale: LCID; dwCmpFlags: DWORD; lpString1: PWideChar;
   cchCount1: Integer; lpString2: PWideChar; cchCount2: Integer): Integer; stdcall;
  {$EXTERNALSYM TCompareStringW}
  TConfigurePortW = function(pName: PWideChar; hWnd: HWND; pPortName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TConfigurePortW}
  TCopyAcceleratorTableW = function(hAccelSrc: HACCEL; var lpAccelDst; cAccelEntries: Integer): Integer; stdcall;
  {$EXTERNALSYM TCopyAcceleratorTableW}
  TCopyEnhMetaFileW = function(p1: HENHMETAFILE; p2: PWideChar): HENHMETAFILE; stdcall;
  {$EXTERNALSYM TCopyEnhMetaFileW}
  TCopyFileExW = function(lpExistingFileName, lpNewFileName: PWideChar;
    lpProgressRoutine: TFNProgressRoutine; lpData: Pointer; pbCancel: PBOOL;
    dwCopyFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCopyFileExW}
  TCopyFileW = function(lpExistingFileName, lpNewFileName: PWideChar; bFailIfExists: BOOL): BOOL; stdcall;
  {$EXTERNALSYM TCopyFileW}
  TCopyMetaFileW = function(p1: HMETAFILE; p2: PWideChar): HMETAFILE; stdcall;
  {$EXTERNALSYM TCopyMetaFileW}
  TCreateAcceleratorTableW = function(var Accel; Count: Integer): HACCEL; stdcall;
  {$EXTERNALSYM TCreateAcceleratorTableW}
  TCreateColorSpaceW = function(var ColorSpace: TLogColorSpaceW): HCOLORSPACE; stdcall;
  {$EXTERNALSYM TCreateColorSpaceW}
  TCreateDCW = function(lpszDriver, lpszDevice, lpszOutput: PWideChar;
    lpdvmInit: PDeviceModeW): HDC; stdcall;
  {$EXTERNALSYM TCreateDCW}
  TCreateDialogIndirectParamW = function(hInstance: HINST; const lpTemplate: TDlgTemplate;
    hWndParent: HWND; lpDialogFunc: TFNDlgProc; dwInitParam: LPARAM): HWND; stdcall;
  {$EXTERNALSYM TCreateDialogIndirectParamW}
  TCreateDialogParamW = function(hInstance: HINST; lpTemplateName: PWideChar;
    hWndParent: HWND; lpDialogFunc: TFNDlgProc; dwInitParam: LPARAM): HWND; stdcall;
  {$EXTERNALSYM TCreateDialogParamW}
  TCreateDirectoryExW = function(lpTemplateDirectory, lpNewDirectory: PWideChar;
    lpSecurityAttributes: PSecurityAttributes): BOOL; stdcall;
  {$EXTERNALSYM TCreateDirectoryExW}
  TCreateDirectoryW = function(lpPathName: PWideChar;
    lpSecurityAttributes: PSecurityAttributes): BOOL; stdcall;
  {$EXTERNALSYM TCreateDirectoryW}
  TCreateEnhMetaFileW = function(DC: HDC; p2: PWideChar; p3: PRect; p4: PWideChar):
    HDC; stdcall;
  {$EXTERNALSYM TCreateEnhMetaFileW}
  TCreateEventW = function(lpEventAttributes: PSecurityAttributes;
    bManualReset, bInitialState: BOOL; lpName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TCreateEventW}
  TCreateFileMappingW = function(hFile: THandle; lpFileMappingAttributes: PSecurityAttributes;
    flProtect, dwMaximumSizeHigh, dwMaximumSizeLow: DWORD; lpName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TCreateFileMappingW}
  TCreateFileW = function(lpFileName: PWideChar; dwDesiredAccess, dwShareMode: DWORD;
    lpSecurityAttributes: PSecurityAttributes; dwCreationDisposition, dwFlagsAndAttributes: DWORD;
    hTemplateFile: THandle): THandle; stdcall;
  {$EXTERNALSYM TCreateFileW}
  TCreateFontIndirectW = function(const p1: TLogFontW): HFONT; stdcall;
  {$EXTERNALSYM TCreateFontIndirectW}
  TCreateFontW = function(nHeight, nWidth, nEscapement, nOrientaion, fnWeight: Integer;
    fdwItalic, fdwUnderline, fdwStrikeOut, fdwCharSet, fdwOutputPrecision,
    fdwClipPrecision, fdwQuality, fdwPitchAndFamily: DWORD; lpszFace: PWideChar): HFONT; stdcall;
  {$EXTERNALSYM TCreateFontW}
  TCreateICW = function(lpszDriver, lpszDevice, lpszOutput: PWideChar; lpdvmInit: PDeviceModeW): HDC; stdcall;
  {$EXTERNALSYM TCreateICW}
  TCreateMDIWindowW = function(lpClassName, lpWindowName: PWideChar;
    dwStyle: DWORD; X, Y, nWidth, nHeight: Integer;
    hWndParent: HWND; hInstance: HINST; lParam: LPARAM): HWND; stdcall;
  {$EXTERNALSYM TCreateMDIWindowW}
  TCreateMailslotW = function(lpName: PWideChar; nMaxMessageSize: DWORD;
    lReadTimeout: DWORD; lpSecurityAttributes: PSecurityAttributes): THandle; stdcall;
  {$EXTERNALSYM TCreateMailslotW}
  TCreateMetaFileW = function(p1: PWideChar): HDC; stdcall;
  {$EXTERNALSYM TCreateMetaFileW}
  TCreateMutexW = function(lpMutexAttributes: PSecurityAttributes; bInitialOwner: BOOL; lpName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TCreateMutexW}
  TCreateNamedPipeW = function(lpName: PWideChar;
    dwOpenMode, dwPipeMode, nMaxInstances, nOutBufferSize, nInBufferSize, nDefaultTimeOut: DWORD;
    lpSecurityAttributes: PSecurityAttributes): THandle; stdcall;
  {$EXTERNALSYM TCreateNamedPipeW}
  TCreateProcessW = function(lpApplicationName: PWideChar; lpCommandLine: PWideChar;
    lpProcessAttributes, lpThreadAttributes: PSecurityAttributes;
    bInheritHandles: BOOL; dwCreationFlags: DWORD; lpEnvironment: Pointer;
    lpCurrentDirectory: PWideChar; const lpStartupInfo: TStartupInfo;
    var lpProcessInformation: TProcessInformation): BOOL; stdcall;
  {$EXTERNALSYM TCreateProcessW}
  TCreateScalableFontResourceW = function(p1: DWORD; p2, p3, p4: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TCreateScalableFontResourceW}
  TCreateSemaphoreW = function(lpSemaphoreAttributes: PSecurityAttributes;
    lInitialCount, lMaximumCount: Longint; lpName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TCreateSemaphoreW}
  TCreateStdAccessibleProxyW = procedure(hWnd: HWND; pszClassName: PWideChar;
    idObject: LongInt; const riidInterface: TIID; var ppvObject: Pointer); stdcall;
  {$EXTERNALSYM TCreateStdAccessibleProxyW}
  TCreateWaitableTimerW = function(lpTimerAttributes: PSecurityAttributes; bManualReset: BOOL; lpTimerName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TCreateWaitableTimerW}
  TCreateWindowExW = function(dwExStyle: DWORD; lpClassName: PWideChar;
    lpWindowName: PWideChar; dwStyle: DWORD; X, Y, nWidth, nHeight: Integer;
    hWndParent: HWND; hMenu: HMENU; hInstance: HINST; lpParam: Pointer): HWND; stdcall;
  {$EXTERNALSYM TCreateWindowExW}
{$IFDEF UNIJWA}
  TCryptAcquireContextW = function(var phProv: HCRYPTPROV; pszContainer: PWideChar;
    pszProvider: PWideChar; dwProvType: DWORD; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCryptAcquireContextW}
  TCryptEnumProviderTypesW = function(dwIndex: DWORD; pdwReserved: LPDWORD;
    dwFlags: DWORD; var pdwProvType: DWORD; pszTypeName: PWideChar;
    var pcbTypeName: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCryptEnumProviderTypesW}
  TCryptEnumProvidersW = function(dwIndex: DWORD; pdwReserved: LPDWORD; dwFlags: DWORD;
    var pdwProvType: DWORD; pszProvName: PWideChar; var pcbProvName: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCryptEnumProvidersW}
  TCryptGetDefaultProviderW = function(dwProvType: DWORD; pdwReserved: LPDWORD;
    dwFlags: DWORD; pszProvName: PWideChar; var pcbProvName: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCryptGetDefaultProviderW}
  TCryptSetProviderExW = function(pszProvName: PWideChar; dwProvType: DWORD;
    pdwReserved: LPDWORD; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCryptSetProviderExW}
  TCryptSetProviderW = function(pszProvName: PWideChar; dwProvType: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCryptSetProviderW}
  TCryptSignHashW = function(hHash: HCRYPTHASH; dwKeySpec: DWORD; sDescription: PWideChar;
    dwFlags: DWORD; pbSignature: LPBYTE; var pdwSigLen: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCryptSignHashW}
  TCryptVerifySignatureW = function(hHash: HCRYPTHASH; pbSignature: LPBYTE;
    dwSigLen: DWORD; hPubKey: HCRYPTKEY; sDescription: PWideChar; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TCryptVerifySignatureW}
{$ENDIF}
  TDdeConnect = function(Inst: DWORD; Service, Topic: HSZ;
    CC: PConvContext): HConv; stdcall;
  {$EXTERNALSYM TDdeConnect}
  TDdeConnectList = function(Inst: DWORD; Service, Topic: HSZ;
    ConvList: HConvList; CC: PConvContext): HConvList; stdcall;
  {$EXTERNALSYM TDdeConnectList}
  TDdeCreateStringHandleW = function(Inst: DWORD; psz: PWideChar;
    CodePage: Integer): HSZ; stdcall;
  {$EXTERNALSYM TDdeCreateStringHandleW}
  TDdeInitializeW = function(var Inst: Longint; Callback: TFNCallback;
    Cmd, Res: Longint): Longint; stdcall;
  {$EXTERNALSYM TDdeInitializeW}
  TDdeQueryConvInfo = function(Conv: HConv; Transaction: DWORD;
   ConvInfo: PConvInfo): DWORD; stdcall;
  {$EXTERNALSYM TDdeQueryConvInfo}
  TDdeQueryStringW = function(Inst: DWORD; HSZ: HSZ; psz: PWideChar;
    Max: DWORD; CodePage: Integer): DWORD; stdcall;
  {$EXTERNALSYM TDdeQueryStringW}
  TDefDlgProcW = function(hDlg: HWND; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  {$EXTERNALSYM TDefDlgProcW}
  TDefFrameProcW = function(hWnd, hWndMDIClient: HWND; uMsg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  {$EXTERNALSYM TDefFrameProcW}
  TDefMDIChildProcW = function(hWnd: HWND; uMsg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  {$EXTERNALSYM TDefMDIChildProcW}
  TDefWindowProcW = function(hWnd: HWND; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  {$EXTERNALSYM TDefWindowProcW}
  TDeleteFileW = function(lpFileName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TDeleteFileW}
  TDeleteMonitorW = function(pName, pEnvironment, pMonitorName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TDeleteMonitorW}
  TDeletePortW = function(pName: PWideChar; hWnd: HWND; pPortName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TDeletePortW}
  TDeletePrintProcessorW = function(pName, pEnvironment, pPrintProcessorName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TDeletePrintProcessorW}
  TDeletePrintProvidorW = function(pName, pEnvironment, pPrintProvidorName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TDeletePrintProvidorW}
  TDeletePrinterDriverW = function(pName, pEnvironment, pDriverName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TDeletePrinterDriverW}
  TDeviceCapabilitiesW = function(pDriverName, pDeviceName, pPort: PWideChar;
    iIndex: Integer; pOutput: PWideChar; DevMode: PDeviceModeW): Integer; stdcall;
  {$EXTERNALSYM TDeviceCapabilitiesW}
  TDialogBoxIndirectParamW = function(hInstance: HINST; const lpDialogTemplate: TDlgTemplate;
    hWndParent: HWND; lpDialogFunc: TFNDlgProc; dwInitParam: LPARAM): Integer; stdcall;
  {$EXTERNALSYM TDialogBoxIndirectParamW}
  TDialogBoxParamW = function(hInstance: HINST; lpTemplateName: PWideChar;
    hWndParent: HWND; lpDialogFunc: TFNDlgProc; dwInitParam: LPARAM): Integer; stdcall;
  {$EXTERNALSYM TDialogBoxParamW}
  TDispatchMessageW = function(const lpMsg: TMsg): Longint; stdcall;
  {$EXTERNALSYM TDispatchMessageW}
  TDlgDirListComboBoxW = function(hDlg: HWND; lpPathSpec: PWideChar;
    nIDComboBox, nIDStaticPath: Integer; uFiletype: DWORD): Integer; stdcall;
  {$EXTERNALSYM TDlgDirListComboBoxW}
  TDlgDirListW = function(hDlg: HWND; lpPathSpec: PWideChar;
    nIDListBox, nIDStaticPath: Integer; uFileType: DWORD): Integer; stdcall;
  {$EXTERNALSYM TDlgDirListW}
  TDlgDirSelectComboBoxExW = function(hDlg: HWND; lpString: PWideChar;
    nCount, nIDComboBox: Integer): BOOL; stdcall;
  {$EXTERNALSYM TDlgDirSelectComboBoxExW}
  TDlgDirSelectExW = function(hDlg: HWND; lpString: PWideChar; nCount, nIDListBox: Integer): BOOL; stdcall;
  {$EXTERNALSYM TDlgDirSelectExW}
  TDocumentPropertiesW = function(hWnd: HWND; hPrinter: THandle; pDeviceName: PWideChar;
    const pDevModeOutput: TDeviceModeW; var pDevModeInput: TDeviceModeW;
    fMode: DWORD): Longint; stdcall;
  {$EXTERNALSYM TDocumentPropertiesW}
  TDragQueryFileW = function(Drop: HDROP; FileIndex: DWORD; FileName: PWideChar; cb: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TDragQueryFileW}
  TDrawStateW = function(DC: HDC; Brush: HBRUSH; CBFunc: TFNDrawStateProc;
    lData: LPARAM; wData: WPARAM; x, y, cx, cy: Integer; Flags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TDrawStateW}
  TDrawTextExW = function(DC: HDC; lpchText: PWideChar; cchText: Integer; var p4: TRect;
    dwDTFormat: DWORD; DTParams: PDrawTextParams): Integer; stdcall;
  {$EXTERNALSYM TDrawTextExW}
  TDrawTextW = function(hDC: HDC; lpString: PWideChar; nCount: Integer;
   var lpRect: TRect; uFormat: DWORD): Integer; stdcall;
  {$EXTERNALSYM TDrawTextW}
  TEnableWindow = function(hWnd: HWND; bEnable: BOOL): BOOL; stdcall;
  {$EXTERNALSYM TEnableWindow}
  TEndUpdateResourceA = function(hUpdate: THandle; fDiscard: BOOL): BOOL; stdcall; // 3703
  {$EXTERNALSYM TEndUpdateResourceA}
  TEndUpdateResourceW = function(hUpdate: THandle; fDiscard: BOOL): BOOL; stdcall;
  {$EXTERNALSYM TEndUpdateResourceW}
{$IFDEF UNIJWA}
  TEnumCalendarInfoExW = function(lpCalInfoEnumProcEx: CALINFO_ENUMPROCEXW;
    Locale: LCID; Calendar: CALID; CalType: CALTYPE): BOOL; stdcall;
  {$EXTERNALSYM TEnumCalendarInfoExW}
{$ENDIF}
  TEnumCalendarInfoW = function(lpCalInfoEnumProc: TFNCalInfoEnumProc; Locale: LCID;
    Calendar: CALID; CalType: CALTYPE): BOOL; stdcall;
  {$EXTERNALSYM TEnumCalendarInfoW}
  TEnumClipboardFormats = function(format: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TEnumClipboardFormats}
{$IFDEF UNIJWA}
  TEnumDateFormatsExW = function(lpDateFmtEnumProcEx: DATEFMT_ENUMPROCEXW;
    Locale: LCID; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumDateFormatsExW}
{$ENDIF}
  TEnumDateFormatsW = function(lpDateFmtEnumProc: TFNDateFmtEnumProc;
    Locale: LCID; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumDateFormatsW}
  TEnumDisplayDevicesW = function(Unused: Pointer; iDevNum: DWORD;
    var lpDisplayDevice: TDisplayDeviceW; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumDisplayDevicesW}
  TEnumDisplaySettingsExW = function(lpszDeviceName: PWideChar; iModeNum: DWORD;
    var lpDevMode: DEVMODEW; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumDisplaySettingsExW}
  TEnumDisplaySettingsW = function(lpszDeviceName: PWideChar; iModeNum: DWORD;
    var lpDevMode: TDeviceModeW): BOOL; stdcall;
  {$EXTERNALSYM TEnumDisplaySettingsW}
  TEnumFontFamiliesExW = function(DC: HDC; var p2: TLogFontW;
    p3: TFNFontEnumProcW; p4: LPARAM; p5: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumFontFamiliesExW}
  TEnumFontFamiliesW = function(DC: HDC; p2: PWideChar; p3: TFNFontEnumProcW; p4: LPARAM): BOOL; stdcall;
  {$EXTERNALSYM TEnumFontFamiliesW}
  TEnumFontsW = function(DC: HDC; lpszFace: PWideChar; fntenmprc: TFNFontEnumProcW;
    lpszData: PChar): Integer; stdcall;
  {$EXTERNALSYM TEnumFontsW}
  TEnumICMProfilesW = function(DC: HDC; ICMProc: TFNICMEnumProc; p3: LPARAM): Integer; stdcall;
  {$EXTERNALSYM TEnumICMProfilesW}
  TEnumMonitorsW = function(pName: PWideChar; Level: DWORD; pMonitors: Pointer; cbBuf: DWORD;
    var pcbNeeded, pcReturned: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumMonitorsW}
  TEnumPortsW = function(pName: PWideChar; Level: DWORD; pPorts: Pointer; cbBuf: DWORD;
    var pcbNeeded, pcReturned: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumPortsW}
  TEnumPrintProcessorDatatypesW = function(pName, pPrintProcessorName: PWideChar; Level: DWORD;
    pDatatypes: Pointer; cbBuf: DWORD; var pcbNeeded, pcReturned: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumPrintProcessorDatatypesW}
  TEnumPrintProcessorsW = function(pName, pEnvironment: PWideChar; Level: DWORD; pPrintProcessorInfo: Pointer;
    cbBuf: DWORD; var pcbNeeded, pcReturned: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumPrintProcessorsW}
  TEnumPrinterDriversW = function(pName, pEnvironment: PWideChar; Level: DWORD;
    pDriverInfo: Pointer; cbBuf: DWORD; var pcbNeeded, pcReturned: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumPrinterDriversW}
  TEnumPrintersW = function(Flags: DWORD; Name: PWideChar; Level: DWORD;
    pPrinterEnum: Pointer; cbBuf: DWORD; var pcbNeeded, pcReturned: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumPrintersW}
  TEnumPropsA = function(hWnd: HWND; lpEnumFunc: TFNPropEnumProc): Integer; stdcall;
  {$EXTERNALSYM TEnumPropsA}
  TEnumPropsExA = function(hWnd: HWND; lpEnumFunc: TFNPropEnumProcEx; lParam: LPARAM): Integer; stdcall;
  {$EXTERNALSYM TEnumPropsExA}
  TEnumPropsExW = function(hWnd: HWND; lpEnumFunc: TFNPropEnumProcEx; lParam: LPARAM): Integer; stdcall;
  {$EXTERNALSYM TEnumPropsExW}
  TEnumPropsW = function(hWnd: HWND; lpEnumFunc: TFNPropEnumProc): Integer; stdcall;
  {$EXTERNALSYM TEnumPropsW}
  TEnumSystemCodePagesW = function(lpCodePageEnumProc: TFNCodepageEnumProc; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumSystemCodePagesW}
  TEnumSystemLocalesW = function(lpLocaleEnumProc: TFNLocaleEnumProc; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumSystemLocalesW}
  TEnumTimeFormatsW = function(lpTimeFmtEnumProc: TFNTimeFmtEnumProc;
    Locale: LCID; dwFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TEnumTimeFormatsW}
{$IFDEF UNIJWA}
  TEnumerateSecurityPackagesW = function(var pcPackages: Cardinal;
    var ppPackageInfo: PSecPkgInfoW): SECURITY_STATUS; stdcall;
  {$EXTERNALSYM TEnumerateSecurityPackagesW}
{$ENDIF}
  TExpandEnvironmentStringsW = function(lpSrc: PWideChar; lpDst: PWideChar;
   nSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TExpandEnvironmentStringsW}
  TExtTextOutW = function(DC: HDC; X, Y: Integer; Options: Longint; Rect: PRect;
    Str: PWideChar; Count: Longint; Dx: PInteger): BOOL; stdcall;
  {$EXTERNALSYM TExtTextOutW}
  TExtractIconExW = function(lpszFile: PWideChar; nIconIndex: Integer;
    var phiconLarge, phiconSmall: HICON; nIcons: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TExtractIconExW}
  TExtractIconW = function(hInst: HINST; lpszExeFileName: PWideChar;
    nIconIndex: DWORD): HICON; stdcall;
  {$EXTERNALSYM TExtractIconW}
  TFatalAppExitW = procedure(uAction: DWORD; lpMessageText: PWideChar); stdcall;
  {$EXTERNALSYM TFatalAppExitW}
  TFillConsoleOutputCharacterW = function(hConsoleOutput: THandle; cCharacter: WideChar;
    nLength: DWORD; dwWriteCoord: TCoord; var lpNumberOfCharsWritten: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TFillConsoleOutputCharacterW}
  TFindAtomW = function(lpString: PWideChar): ATOM; stdcall;
  {$EXTERNALSYM TFindAtomW}
  TFindExecutableW = function(FileName, Directory: PWideChar; Result: PWideChar): HINST; stdcall;
  {$EXTERNALSYM TFindExecutableW}
  TFindFirstChangeNotificationW = function(lpPathName: PWideChar;
    bWatchSubtree: BOOL; dwNotifyFilter: DWORD): THandle; stdcall;
  {$EXTERNALSYM TFindFirstChangeNotificationW}
  TFindFirstFileW = function(lpFileName: PWideChar; var lpFindFileData: TWIN32FindDataW): THandle; stdcall;
  {$EXTERNALSYM TFindFirstFileW}
  TFindNextFileW = function(hFindFile: THandle; var lpFindFileData: TWIN32FindDataW): BOOL; stdcall;
  {$EXTERNALSYM TFindNextFileW}
  TFindResourceExW = function(hModule: HMODULE; lpType, lpName: PWideChar; wLanguage: Word): HRSRC; stdcall;
  {$EXTERNALSYM TFindResourceExW}
  TFindResourceW = function(hModule: HMODULE; lpName, lpType: PWideChar): HRSRC; stdcall;
  {$EXTERNALSYM TFindResourceW}
  TFindTextW = function(var FindReplace: TFindReplaceW): HWND; stdcall;
  {$EXTERNALSYM TFindTextW}
  TFindWindowExW = function(Parent, Child: HWND; ClassName, WindowName: PWideChar): HWND; stdcall;
  {$EXTERNALSYM TFindWindowExW}
  TFindWindowW = function(lpClassName, lpWindowName: PWideChar): HWND; stdcall;
  {$EXTERNALSYM TFindWindowW}
  TFormatMessageW = function(dwFlags: DWORD; lpSource: Pointer; dwMessageId: DWORD; dwLanguageId: DWORD;
    lpBuffer: PWideChar; nSize: DWORD; Arguments: Pointer): DWORD; stdcall;
  {$EXTERNALSYM TFormatMessageW}
{$IFDEF UNIJWA}
  TFreeContextBuffer = function(pvContextBuffer: Pointer): SECURITY_STATUS; stdcall;
  {$EXTERNALSYM TFreeContextBuffer}
{$ENDIF}
  TFreeEnvironmentStringsW = function(EnvBlock: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TFreeEnvironmentStringsW}
  TGetAltTabInfoW = function(hwnd: HWND; iItem: Integer; var pati: TAltTabInfo;
    pszItemText: PWideChar; cchItemText: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetAltTabInfoW}
  TGetAtomNameW = function(nAtom: ATOM; lpBuffer: PWideChar; nSize: Integer): DWORD; stdcall;
  {$EXTERNALSYM TGetAtomNameW}
  TGetCPInfo = function(CodePage: DWORD; var lpCPInfo: TCPInfo): BOOL; stdcall;
  {$EXTERNALSYM TGetCPInfo}
{$IFDEF UNIJWA}
  TGetCPInfoExW = function(CodePage: DWORD; dwFlags: DWORD; var lpCPInfoEx: CPINFOEXW): BOOL; stdcall;
  {$EXTERNALSYM TGetCPInfoExW}
{$ENDIF}
  TGetCalendarInfoW = function(Locale: LCID; Calendar: CALID; CalType: CALTYPE;
    lpCalData: PWideChar; cchData: Integer; lpValue: LPDWORD): Integer; stdcall;
  {$EXTERNALSYM TGetCalendarInfoW}
  TGetCharABCWidthsFloatW = function(DC: HDC; FirstChar, LastChar: DWORD; const ABCFloatSturcts): BOOL; stdcall;
  {$EXTERNALSYM TGetCharABCWidthsFloatW}
  TGetCharABCWidthsW = function(DC: HDC; FirstChar, LastChar: DWORD; const ABCStructs): BOOL; stdcall;
  {$EXTERNALSYM TGetCharABCWidthsW}
  TGetCharWidth32W = function(DC: HDC; FirstChar, LastChar: DWORD; const Widths): BOOL; stdcall;
  {$EXTERNALSYM TGetCharWidth32W}
  TGetCharWidthFloatW = function(DC: HDC; FirstChar, LastChar: DWORD; const Widths): BOOL; stdcall;
  {$EXTERNALSYM TGetCharWidthFloatW}
  TGetCharWidthW = function(DC: HDC; FirstChar, LastChar: DWORD; const Widths): BOOL; stdcall;
  {$EXTERNALSYM TGetCharWidthW}
  TGetCharacterPlacementW = function(DC: HDC; p2: PWideChar; p3, p4: BOOL;
    var p5: TGCPResults; p6: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetCharacterPlacementW}
  TGetClassInfoExW = function(Instance: HINST; Classname: PWideChar; var WndClass: TWndClassExW): BOOL; stdcall;
  {$EXTERNALSYM TGetClassInfoExW}
  TGetClassInfoW = function(hInstance: HINST; lpClassName: PWideChar;
    var lpWndClass: TWndClassW): BOOL; stdcall;
  {$EXTERNALSYM TGetClassInfoW}
  TGetClassLongW = function(hWnd: HWND; nIndex: Integer): DWORD; stdcall;
  {$EXTERNALSYM TGetClassLongW}
  TGetClassNameW = function(hWnd: HWND; lpClassName: PWideChar; nMaxCount: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetClassNameW}
  TGetClipboardData = function(uFormat: DWORD): THandle; stdcall;
  {$EXTERNALSYM TGetClipboardData}
  TGetClipboardFormatNameW = function(format: DWORD; lpszFormatName: PWideChar;
    cchMaxCount: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetClipboardFormatNameW}
  TGetComputerNameW = function(lpBuffer: PWideChar; var nSize: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetComputerNameW}
  TGetConsoleTitleW = function(lpConsoleTitle: PWideChar; nSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetConsoleTitleW}
  TGetCurrencyFormatW = function(Locale: LCID; dwFlags: DWORD; lpValue: PWideChar;
    lpFormat: PCurrencyFmtW; lpCurrencyStr: PWideChar; cchCurrency: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetCurrencyFormatW}
  TGetCurrentDirectoryW = function(nBufferLength: DWORD; lpBuffer: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetCurrentDirectoryW}
  TGetCurrentHwProfileW = function(var lpHwProfileInfo: THWProfileInfoW): BOOL; stdcall;
  {$EXTERNALSYM TGetCurrentHwProfileW}
  TGetDateFormatW = function(Locale: LCID; dwFlags: DWORD; lpDate: PSystemTime;
    lpFormat: PWideChar; lpDateStr: PWideChar; cchDate: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetDateFormatW}
  TGetDefaultCommConfigW = function(lpszName: PWideChar;
    var lpCC: TCommConfig; var lpdwSize: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetDefaultCommConfigW}
  TGetDiskFreeSpaceExW = function(lpDirectoryName: PWideChar;
    var lpFreeBytesAvailableToCaller, lpTotalNumberOfBytes; lpTotalNumberOfFreeBytes: PLargeInteger): BOOL; stdcall;
  {$EXTERNALSYM TGetDiskFreeSpaceExW}
  TGetDiskFreeSpaceW = function(lpRootPathName: PWideChar;
    var lpSectorsPerCluster, lpBytesPerSector, lpNumberOfFreeClusters, lpTotalNumberOfClusters: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetDiskFreeSpaceW}
  TGetDlgItemTextW = function(hDlg: HWND; nIDDlgItem: Integer;
    lpString: PWideChar; nMaxCount: Integer): DWORD; stdcall;
  {$EXTERNALSYM TGetDlgItemTextW}
  TGetDriveTypeW = function(lpRootPathName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetDriveTypeW}
  TGetEnhMetaFileDescriptionW = function(p1: HENHMETAFILE; p2: DWORD; p3: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetEnhMetaFileDescriptionW}
  TGetEnhMetaFileW = function(p1: PWideChar): HENHMETAFILE; stdcall;
  {$EXTERNALSYM TGetEnhMetaFileW}
  TGetEnvironmentStringsW = function: PWideChar; stdcall;
  {$EXTERNALSYM TGetEnvironmentStringsW}
  TGetEnvironmentVariableW = function(lpName: PWideChar; lpBuffer: PWideChar; nSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetEnvironmentVariableW}
  TGetFileAttributesExW = function(lpFileName: PWideChar;
    fInfoLevelId: TGetFileExInfoLevels; lpFileInformation: Pointer): BOOL; stdcall;
  {$EXTERNALSYM TGetFileAttributesExW}
  TGetFileAttributesW = function(lpFileName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetFileAttributesW}
  TGetFileTitleW = function(FileName: PWideChar; Title: PWideChar; TitleSize: Word): Smallint; stdcall;
  {$EXTERNALSYM TGetFileTitleW}
  TGetFileVersionInfoSizeW = function(lptstrFilename: PWideChar; var lpdwHandle: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetFileVersionInfoSizeW}
  TGetFileVersionInfoW = function(lptstrFilename: PWideChar; dwHandle, dwLen: DWORD;
    lpData: Pointer): BOOL; stdcall;
  {$EXTERNALSYM TGetFileVersionInfoW}
  TGetFullPathNameW = function(lpFileName: PWideChar; nBufferLength: DWORD;
    lpBuffer: PWideChar; var lpFilePart: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetFullPathNameW}
  TGetGlyphOutlineW = function(DC: HDC; uChar, uFormat: DWORD;
    const lpgm: TGlyphMetrics; cbBuffer: DWORD; lpvBuffer: Pointer; const lpmat2: TMat2): DWORD; stdcall;
  {$EXTERNALSYM TGetGlyphOutlineW}
  TGetICMProfileW = function(DC: HDC; var Size: DWORD; Name: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TGetICMProfileW}
  TGetJobW = function(hPrinter: THandle; JobId: DWORD; Level: DWORD; pJob: Pointer; cbBuf: DWORD; pcbNeeded: PDWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetJobW}
{$IFDEF UNIJWA}
  TGetKerningPairsW = function(hDc: HDC; nNumPairs: DWORD; lpkrnpair: LPKERNINGPAIR): DWORD; stdcall;
  {$EXTERNALSYM TGetKerningPairsW}
{$ENDIF}
  TGetKeyNameTextW = function(lParam: Longint; lpString: PWideChar; nSize: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetKeyNameTextW}
  TGetKeyboardLayoutNameW = function(pwszKLID: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TGetKeyboardLayoutNameW}
  TGetLocaleInfoW = function(Locale: LCID; LCType: LCTYPE; lpLCData: PWideChar; cchData: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetLocaleInfoW}
  TGetLogColorSpaceW = function(p1: HCOLORSPACE; var ColorSpace: TLogColorSpaceW; Size: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetLogColorSpaceW}
  TGetLogicalDriveStringsW = function(nBufferLength: DWORD; lpBuffer: PAnsiChar): DWORD; stdcall;
  {$EXTERNALSYM TGetLogicalDriveStringsW}
  TGetLongPathNameW = function(lpszShortPath: PWideChar; lpszLongPath: PWideChar;
    cchBuffer: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetLongPathNameW}
  TGetMenuItemInfoW = function(p1: HMENU; p2: DWORD; p3: BOOL; var p4: TMenuItemInfoW): BOOL; stdcall;
  {$EXTERNALSYM TGetMenuItemInfoW}
  TGetMenuStringW = function(hMenu: HMENU; uIDItem: DWORD; lpString: PWideChar;
    nMaxCount: Integer; uFlag: DWORD): Integer; stdcall;
  {$EXTERNALSYM TGetMenuStringW}
  TGetMessageW = function(var lpMsg: TMsg; hWnd: HWND;
    wMsgFilterMin, wMsgFilterMax: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetMessageW}
  TGetMetaFileW = function(p1: PWideChar): HMETAFILE; stdcall;
  {$EXTERNALSYM TGetMetaFileW}
  TGetModuleFileNameW = function(hModule: HINST; lpFilename: PWideChar; nSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetModuleFileNameW}
  TGetModuleHandleW = function(lpModuleName: PWideChar): HMODULE; stdcall;
  {$EXTERNALSYM TGetModuleHandleW}
  TGetMonitorInfoW = function(hMonitor: HMONITOR; lpMonitorInfo: PMonitorInfoW): Boolean; stdcall;
  {$EXTERNALSYM TGetMonitorInfoW}
  TGetNamedPipeHandleStateW = function(hNamedPipe: THandle;
    lpState, lpCurInstances, lpMaxCollectionCount, lpCollectDataTimeout: PDWORD;
    lpUserName: PWideChar; nMaxUserNameSize: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetNamedPipeHandleStateW}
  TGetNumberFormatW = function(Locale: LCID; dwFlags: DWORD; lpValue: PWideChar;
    lpFormat: PNumberFmtW; lpNumberStr: PWideChar; cchNumber: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetNumberFormatW}
  TGetObjectW = function(p1: HGDIOBJ; p2: Integer; p3: Pointer): Integer; stdcall;
  {$EXTERNALSYM TGetObjectW}
  TGetOpenFileNamePreviewW = function(var OpenFile: TOpenFilenameW): BOOL; stdcall;
  {$EXTERNALSYM TGetOpenFileNamePreviewW}
  TGetOpenFileNameW = function(var OpenFile: TOpenFilenameW): BOOL; stdcall;
  {$EXTERNALSYM TGetOpenFileNameW}
  TGetOutlineTextMetricsW = function(DC: HDC; p2: DWORD; OTMetricStructs: Pointer): DWORD; stdcall;
  {$EXTERNALSYM TGetOutlineTextMetricsW}
  TGetPrintProcessorDirectoryW = function(pName, pEnvironment: PWideChar; Level: DWORD;
    pPrintProcessorInfo: Pointer; cbBuf: DWORD; var pcbNeeded: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetPrintProcessorDirectoryW}
  TGetPrinterDataW = function(hPrinter: THandle; pValueName: PWideChar; pType: PDWORD; pData: Pointer;
    nSize: DWORD; var pcbNeeded: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetPrinterDataW}
  TGetPrinterDriverDirectoryW = function(pName, pEnvironment: PWideChar; Level: DWORD;
    pDriverDirectory: Pointer; cbBuf: DWORD; var pcbNeeded: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetPrinterDriverDirectoryW}
  TGetPrinterDriverW = function(hPrinter: THandle; pEnvironment: PWideChar; Level: DWORD;
    pDriverInfo: Pointer; cbBuf: DWORD; var pcbNeeded: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetPrinterDriverW}
  TGetPrinterW = function(hPrinter: THandle; Level: DWORD; pPrinter: Pointer; cbBuf: DWORD; pcbNeeded: PDWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetPrinterW}
  TGetPrivateProfileIntW = function(lpAppName, lpKeyName: PWideChar; nDefault: Integer; lpFileName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetPrivateProfileIntW}
  TGetPrivateProfileSectionNamesW = function(lpszReturnBuffer: PWideChar; nSize: DWORD; lpFileName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetPrivateProfileSectionNamesW}
  TGetPrivateProfileSectionW = function(lpAppName: PWideChar; lpReturnedString: PWideChar; nSize: DWORD; lpFileName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetPrivateProfileSectionW}
  TGetPrivateProfileStringW = function(lpAppName, lpKeyName, lpDefault: PWideChar;
    lpReturnedString: PWideChar; nSize: DWORD; lpFileName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetPrivateProfileStringW}
  TGetPrivateProfileStructW = function(lpszSection, lpszKey: PWideChar;
    lpStruct: Pointer; uSizeStruct: DWORD; szFile: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TGetPrivateProfileStructW}
  TGetProcAddress = function(hModule: HMODULE; lpProcName: PChar): FARPROC; stdcall;
  {$EXTERNALSYM TGetProcAddress}
  TGetProfileIntW = function(lpAppName, lpKeyName: PWideChar; nDefault: Integer): DWORD; stdcall;
  {$EXTERNALSYM TGetProfileIntW}
  TGetProfileSectionW = function(lpAppName: PWideChar; lpReturnedString: PWideChar; nSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetProfileSectionW}
  TGetProfileStringW = function(lpAppName, lpKeyName, lpDefault: PWideChar;
    lpReturnedString: PWideChar; nSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetProfileStringW}
  TGetPropA = function(hWnd: HWND; lpString: PAnsiChar): THandle; stdcall;
  {$EXTERNALSYM TGetPropA}
  TGetPropW = function(hWnd: HWND; lpString: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TGetPropW}
  TGetRoleTextW = function(Role: DWORD; lpszRole: PWideChar; cchRoleMax: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetRoleTextW}
  TGetSaveFileNamePreviewW = function(var OpenFile: TOpenFilenameW): BOOL; stdcall;
  {$EXTERNALSYM TGetSaveFileNamePreviewW}
  TGetSaveFileNameW = function(var OpenFile: TOpenFilenameW): BOOL; stdcall;
  {$EXTERNALSYM TGetSaveFileNameW}
  TGetShortPathNameW = function(lpszLongPath: PWideChar; lpszShortPath: PWideChar;
    cchBuffer: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetShortPathNameW}
  TGetStartupInfoW = procedure (var lpStartupInfo: TStartupInfo); stdcall;
  {$EXTERNALSYM TGetStartupInfoW}
  TGetStateTextW = function(StateBit: DWORD; lpszStateBit: PWideChar;
    cchStateBitMax: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetStateTextW}
  TGetStringTypeExW = function(Locale: LCID; dwInfoType: DWORD;
    lpSrcStr: PWideChar; cchSrc: Integer; var lpCharType): BOOL; stdcall;
  {$EXTERNALSYM TGetStringTypeExW}
  TGetStringTypeW = function(dwInfoType: DWORD; const lpSrcStr: WCHAR; cchSrc: BOOL; var lpCharType: Word): BOOL;
  {$EXTERNALSYM TGetStringTypeW}
  TGetSystemDirectoryW = function(lpBuffer: PWideChar; uSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetSystemDirectoryW}
  TGetSystemWindowsDirectoryW = function(lpBuffer: PWideChar; uSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetSystemWindowsDirectoryW}
  TGetTabbedTextExtentW = function(hDC: HDC; lpString: PWideChar;
    nCount, nTabPositions: Integer; var lpnTabStopPositions): DWORD; stdcall;
  {$EXTERNALSYM TGetTabbedTextExtentW}
  TGetTempFileNameW = function(lpPathName, lpPrefixString: PWideChar;
    uUnique: DWORD; lpTempFileName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetTempFileNameW}
  TGetTempPathW = function(nBufferLength: DWORD; lpBuffer: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TGetTempPathW}
  TGetTextExtentExPointW = function(DC: HDC; p2: PWideChar; p3, p4: Integer;
    p5, p6: PInteger; var p7: TSize): BOOL; stdcall;
  {$EXTERNALSYM TGetTextExtentExPointW}
  TGetTextExtentPoint32W = function(DC: HDC; Str: PWideChar; Count: Integer;
    var Size: TSize): BOOL; stdcall;
  {$EXTERNALSYM TGetTextExtentPoint32W}
  TGetTextExtentPointW = function(DC: HDC; Str: PWideChar; Count: Integer;
    var Size: TSize): BOOL; stdcall;
  {$EXTERNALSYM TGetTextExtentPointW}
  TGetTextFaceW = function(DC: HDC; Count: Integer; Buffer: PWideChar): Integer; stdcall;
  {$EXTERNALSYM TGetTextFaceW}
  TGetTextMetricsW = function(DC: HDC; var TM: TTextMetricW): BOOL; stdcall;
  {$EXTERNALSYM TGetTextMetricsW}
  TGetTimeFormatW = function(Locale: LCID; dwFlags: DWORD; lpTime: PSystemTime;
    lpFormat: PWideChar; lpTimeStr: PWideChar; cchTime: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetTimeFormatW}
  TGetUserNameW = function(lpBuffer: PWideChar; var nSize: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetUserNameW}
  TGetVersionExW = function(var lpVersionInformation: TOSVersionInfo): BOOL; stdcall;
  {$EXTERNALSYM TGetVersionExW}
  TGetVolumeInformationW = function(lpRootPathName: PWideChar;
    lpVolumeNameBuffer: PWideChar; nVolumeNameSize: DWORD; lpVolumeSerialNumber: PDWORD;
    var lpMaximumComponentLength, lpFileSystemFlags: DWORD;
    lpFileSystemNameBuffer: PWideChar; nFileSystemNameSize: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TGetVolumeInformationW}
  TGetWindowLongA = function(hWnd: HWND; nIndex: Integer): Longint; stdcall;
  {$EXTERNALSYM TGetWindowLongA}
  TGetWindowLongW = function(hWnd: HWND; nIndex: Integer): Longint; stdcall;
  {$EXTERNALSYM TGetWindowLongW}
  TGetWindowModuleFileNameW = function(hwnd: HWND; pszFileName: PAnsiChar; cchFileNameMax: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetWindowModuleFileNameW}
  TGetWindowTextLengthW = function(hWnd: HWND): Integer; stdcall;
  {$EXTERNALSYM TGetWindowTextLengthW}
  TGetWindowTextW = function(hWnd: HWND; lpString: PWideChar; nMaxCount: Integer): Integer; stdcall;
  {$EXTERNALSYM TGetWindowTextW}
  TGetWindowsDirectoryW = function(lpBuffer: PWideChar; uSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TGetWindowsDirectoryW}
  TGlobalAddAtomW = function(lpString: PWideChar): ATOM; stdcall;
  {$EXTERNALSYM TGlobalAddAtomW}
  TGlobalFindAtomW = function(lpString: PWideChar): ATOM; stdcall;
  {$EXTERNALSYM TGlobalFindAtomW}
  TGlobalGetAtomNameW = function(nAtom: ATOM; lpBuffer: PWideChar; nSize: Integer): DWORD; stdcall;
  {$EXTERNALSYM TGlobalGetAtomNameW}
  TGrayStringW = function(hDC: HDC; hBrush: HBRUSH; lpOutputFunc: TFNGrayStringProc;
    lpData: LPARAM; nCount, X, Y, nWidth, nHeight: Integer): BOOL; stdcall;
  {$EXTERNALSYM TGrayStringW}
{$IFDEF UNIJWA}
  TInitSecurityInterfaceW = function: PSecurityFunctionTableW; stdcall;
  {$EXTERNALSYM TInitSecurityInterfaceW}
  TInitializeSecurityContextW = function(phCredential: PCredHandle; phContext: PCtxtHandle;
    pszTargetName: PSecWChar; fContextReq, Reserved1, TargetDataRep: Cardinal;
    pInput: PSecBufferDesc; Reserved2: Cardinal; phNewContext: PCtxtHandle;
    pOutput: PSecBufferDesc; var pfContextAttr: Cardinal; ptsExpiry: PTimeStamp): SECURITY_STATUS; stdcall;
  {$EXTERNALSYM TInitializeSecurityContextW}
{$ENDIF}
  TInsertMenuItemW = function(p1: HMENU; p2: DWORD; p3: BOOL; const p4: TMenuItemInfoW): BOOL; stdcall;
  {$EXTERNALSYM TInsertMenuItemW}
  {$EXTERNALSYM TInsertMenuW}
  TInsertMenuW = function(hMenu: HMENU; uPosition, uFlags, uIDNewItem: DWORD;
    lpNewItem: PWideChar): BOOL; stdcall;
  TIsBadStringPtrW = function(lpsz: PWideChar; ucchMax: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TIsBadStringPtrW}
  TIsCharAlphaNumericW = function(ch: WideChar): BOOL; stdcall;
  {$EXTERNALSYM TIsCharAlphaNumericW}
  TIsCharAlphaW = function(ch: WideChar): BOOL; stdcall;
  {$EXTERNALSYM TIsCharAlphaW}
  TIsCharLowerW = function(ch: WideChar): BOOL; stdcall;
  {$EXTERNALSYM TIsCharLowerW}
  TIsCharUpperW = function(ch: WideChar): BOOL; stdcall;
  {$EXTERNALSYM TIsCharUpperW}
  TIsClipboardFormatAvailable = function(format: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TIsClipboardFormatAvailable}
  TIsDestinationReachableW = function(Destination: PWideChar; var QOCInfo: TQOCInfo): BOOL; stdcall;
  {$EXTERNALSYM TIsDestinationReachableW}
  TIsDialogMessageW = function(hDlg: HWND; var lpMsg: TMsg): BOOL; stdcall;
  {$EXTERNALSYM TIsDialogMessageW}
  TIsTextUnicode = function(lpBuffer: Pointer; cb: Integer; lpi: PINT): BOOL; stdcall;
  {$EXTERNALSYM TIsTextUnicode}
  TIsValidCodePage = function(CodePage: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TIsValidCodePage}
  TIsWindowUnicode = function(hWnd: HWND): BOOL; stdcall;
  {$EXTERNALSYM TIsWindowUnicode}
  TLCMapStringW = function(Locale: LCID; dwMapFlags: DWORD; lpSrcStr: PWideChar;
    cchSrc: Integer; lpDestStr: PWideChar; cchDest: Integer): Integer; stdcall;
  {$EXTERNALSYM TLCMapStringW}
  TLoadAcceleratorsW = function(hInstance: HINST; lpTableName: PWideChar): HACCEL; stdcall;
  {$EXTERNALSYM TLoadAcceleratorsW}
  TLoadBitmapW = function(hInstance: HINST; lpBitmapName: PAnsiChar): HBITMAP; stdcall;
  {$EXTERNALSYM TLoadBitmapW}
  TLoadCursorFromFileW = function(lpFileName: PAnsiChar): HCURSOR; stdcall;
  {$EXTERNALSYM TLoadCursorFromFileW}
  TLoadCursorW = function(hInstance: HINST; lpCursorName: PAnsiChar): HCURSOR; stdcall;
  {$EXTERNALSYM TLoadCursorW}
  TLoadIconW = function(hInstance: HINST; lpIconName: PWideChar): HICON; stdcall;
  {$EXTERNALSYM TLoadIconW}
  TLoadImageW = function(hInst: HINST; ImageName: PWideChar; ImageType: DWORD; X, Y: Integer; Flags: DWORD): THandle; stdcall;
  {$EXTERNALSYM TLoadImageW}
  TLoadKeyboardLayoutW = function(pwszKLID: PWideChar; Flags: DWORD): HKL; stdcall;
  {$EXTERNALSYM TLoadKeyboardLayoutW}
  TLoadLibraryExW = function(lpLibFileName: PWideChar; hFile: THandle; dwFlags: DWORD): HMODULE; stdcall;
  {$EXTERNALSYM TLoadLibraryExW}
  TLoadLibraryW = function(lpLibFileName: PWideChar): HMODULE; stdcall;
  {$EXTERNALSYM TLoadLibraryW}
  TLoadMenuIndirectW = function(lpMenuTemplate: Pointer): HMENU; stdcall;
  {$EXTERNALSYM TLoadMenuIndirectW}
  TLoadMenuW = function(hInstance: HINST; lpMenuName: PAnsiChar): HMENU; stdcall;
  {$EXTERNALSYM TLoadMenuW}
  TLoadStringW = function(hInstance: HINST; uID: DWORD; lpBuffer: PWideChar; nBufferMax: Integer): Integer; stdcall;
  {$EXTERNALSYM TLoadStringW}
  TMCIWndCreateW = function(hWndParent: HWND; Instance: HINST; Style: DWORD; FileName: PWideChar): HWND; stdcall;
  {$EXTERNALSYM TMCIWndCreateW}
  TMapVirtualKeyExW = function(uCode, uMapType: DWORD; dwhkl: HKL): DWORD; stdcall;
  {$EXTERNALSYM TMapVirtualKeyExW}
  TMapVirtualKeyW = function(uCode, uMapType: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TMapVirtualKeyW}
  TMessageBoxExW = function(hWnd: HWND; lpText, lpCaption: PWideChar;
    uType: DWORD; wLanguageId: Word): Integer; stdcall;
  {$EXTERNALSYM TMessageBoxExW}
  TMessageBoxIndirectW = function(const MsgBoxParams: TMsgBoxParamsW): BOOL; stdcall;
  {$EXTERNALSYM TMessageBoxIndirectW}
  TMessageBoxW = function(hWnd: HWND; lpText, lpCaption: PWideChar; uType: DWORD): Integer; stdcall;
  {$EXTERNALSYM TMessageBoxW}
  TModifyMenuW = function(hMnu: HMENU; uPosition, uFlags, uIDNewItem: DWORD;
    lpNewItem: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TModifyMenuW}
  TMoveFileW = function(lpExistingFileName, lpNewFileName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TMoveFileW}
  TMultiByteToWideChar = function(CodePage: DWORD; dwFlags: DWORD;
    const lpMultiByteStr: PChar; cchMultiByte: Integer;
    lpWideCharStr: PWideChar; cchWideChar: Integer): Integer; stdcall;
  {$EXTERNALSYM TMultiByteToWideChar}
  TMultinetGetConnectionPerformanceW = function(lpNetResource: PNetResourceW;
    lpNetConnectInfoStruc: PNetConnectInfoStruct): DWORD; stdcall;
  {$EXTERNALSYM TMultinetGetConnectionPerformanceW}
  TOemToCharBuffW = function(lpszSrc: PWideChar; lpszDst: PWideChar; cchDstLength: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TOemToCharBuffW}
  TOemToCharW = function(lpszSrc: PWideChar; lpszDst: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TOemToCharW}
  TOleUIAddVerbMenuW = function(oleObj: IOleObject; pszShortType: PWideChar;
    menu: HMenu; uPos: Integer; uIDVerbMin: Integer; uIDVerbMax: Integer;
    bAddConvert: BOOL; idConvert: Integer; var outMenu: HMenu): BOOL; stdcall;
  {$EXTERNALSYM TOleUIAddVerbMenuW}
  TOleUIBusyWFunc = function(var Info: TOleUIBusyW): UINT; stdcall;
  {$EXTERNALSYM TOleUIBusyWFunc}
  TOleUIChangeIconWFunc = function(var Info: TOleUIChangeIconW): UINT; stdcall;
  {$EXTERNALSYM TOleUIChangeIconWFunc}
  TOleUIChangeSourceWFunc = function(var CS: TOleUIChangeSourceW): UINT; stdcall;
  {$EXTERNALSYM TOleUIChangeSourceWFunc}
  TOleUIConvertWFunc = function(var Info: TOleUIConvertW): UINT; stdcall;
  {$EXTERNALSYM TOleUIConvertWFunc}
  TOleUIEditLinksWFunc = function(var Info: TOleUIEditLinksW): UINT; stdcall;
  {$EXTERNALSYM TOleUIEditLinksWFunc}
  TOleUIInsertObjectWFunc = function(var Info: TOleUIInsertObjectW): UINT; stdcall;
  {$EXTERNALSYM TOleUIInsertObjectWFunc}
  TOleUIObjectPropertiesW = function(var Info: TOleUIObjectPropsW): UINT; stdcall;
  {$EXTERNALSYM TOleUIObjectPropertiesW}
  TOleUIPasteSpecialWFunc = function(var Info: TOleUIPasteSpecialW): UINT; stdcall;
  {$EXTERNALSYM TOleUIPasteSpecialWFunc}
  TOleUIPromptUserW = function(Template: Integer; hWndParent: HWND): Integer; stdcall;
  {$EXTERNALSYM TOleUIPromptUserW}
  TOleUIUpdateLinksW = function(OleUILinkCntr: IOleUILinkContainer; hWndParent: HWND;
     Title: PWideChar; Links: Integer): BOOL; stdcall;
  {$EXTERNALSYM TOleUIUpdateLinksW}
  TOpenEventW = function(dwDesiredAccess: DWORD; bInheritHandle: BOOL; lpName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TOpenEventW}
  TOpenFileMappingW = function(dwDesiredAccess: DWORD; bInheritHandle: BOOL; lpName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TOpenFileMappingW}
  TOpenMutexW = function(dwDesiredAccess: DWORD; bInheritHandle: BOOL; lpName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TOpenMutexW}
  TOpenPrinterW = function(pPrinterName: PWideChar; var phPrinter: THandle; pDefault: PPrinterDefaultsW): BOOL; stdcall;
  {$EXTERNALSYM TOpenPrinterW}
  TOpenSemaphoreW = function(dwDesiredAccess: DWORD; bInheritHandle: BOOL; lpName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TOpenSemaphoreW}
  TOpenWaitableTimerW = function(dwDesiredAccess: DWORD; bInheritHandle: BOOL;
    lpTimerName: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TOpenWaitableTimerW}
  TOutputDebugStringW = procedure(lpOutputString: PWideChar); stdcall;
  {$EXTERNALSYM TOutputDebugStringW}
  TPageSetupDlgWFunc = function(var PgSetupDialog: TPageSetupDlgW): BOOL; stdcall;
  {$EXTERNALSYM TPageSetupDlgWFunc}
  TPeekConsoleInputW = function(hConsoleInput: THandle; var lpBuffer: TInputRecord;
    nLength: DWORD; var lpNumberOfEventsRead: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TPeekConsoleInputW}
  TPeekMessageW = function(var lpMsg: TMsg; hWnd: HWND;
    wMsgFilterMin, wMsgFilterMax, wRemoveMsg: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TPeekMessageW}
  TPlaySoundW = function(pszSound: PWideChar; hmod: HMODULE; fdwSound: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TPlaySoundW}
  TPolyTextOutW = function(DC: HDC; const PolyTextArray; Strings: Integer): BOOL; stdcall;
  {$EXTERNALSYM TPolyTextOutW}
  TPostMessageW = function(hWnd: HWND; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): BOOL; stdcall;
  {$EXTERNALSYM TPostMessageW}
  TPostThreadMessageW = function(idThread: DWORD; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): BOOL; stdcall;
  {$EXTERNALSYM TPostThreadMessageW}
  TPrintDlgWFunc = function(var PrintDlg: TPrintDlgW): BOOL; stdcall;
  {$EXTERNALSYM TPrintDlgWFunc}
{$IFDEF UNIJWA}
  TQueryContextAttributesW = function(phContext: PCtxtHandle; ulAttribute: Cardinal;
    pBuffer: Pointer): SECURITY_STATUS; stdcall;
  {$EXTERNALSYM TQueryContextAttributesW}
  TQueryCredentialsAttributesW = function(phCredential: PCredHandle;
    ulAttribute: Cardinal; pBuffer: Pointer): SECURITY_STATUS; stdcall;
  {$EXTERNALSYM TQueryCredentialsAttributesW}
{$ENDIF}
  TQueryDosDeviceW = function(lpDeviceName: PWideChar; lpTargetPath: PWideChar; ucchMax: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TQueryDosDeviceW}
{$IFDEF UNIJWA}
  TQuerySecurityPackageInfoW = function(pszPackageName: PSecWChar;
    var ppPackageInfo: PSecPkgInfoW): SECURITY_STATUS; stdcall;
  {$EXTERNALSYM TQuerySecurityPackageInfoW}
{$ENDIF}
{$IFDEF UNIJRA}
  TRasConnectionNotificationW = function(hrasconn: HRasConn; hEvent: THandle;
    dwFlags: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TRasConnectionNotificationW}
  TRasCreatePhonebookEntryW = function(hWndParent : HWND; lpszPhonebook : PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TRasCreatePhonebookEntryW}
  TRasDeleteEntryW = function(lpszPhonebook, szEntry: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TRasDeleteEntryW}
  TRasDeleteSubEntryW = function(lpszPhoneBook, lpszEntry: PWideChar;
    dwSubEntryId: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TRasDeleteSubEntryW}
  TRasDialW = function(lpRasDialExtensions: PRasDialExtensions; lpszPhonebook: PChar;
    lpRasDialParams: PRasDialParamsW; dwNotifierType: DWORD; lpvNotifier: Pointer;
    var rasConn: HRASConn): DWORD; stdcall;
  {$EXTERNALSYM TRasDialW}
  TRasEditPhonebookEntryW = function(hWndParent: HWND; lpszPhonebook: PWideChar;
    lpszEntryName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TRasEditPhonebookEntryW}
  TRasEnumConnectionsW = function(RASConnW: PrasConn; var BufSize: DWord;
    var Connections: DWord): DWORD; stdcall;
  {$EXTERNALSYM TRasEnumConnectionsW}
  TRasEnumDevicesW = function(lpBuff: PRasDevInfoW; var lpcbSize: DWORD;
    var lpcDevices: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TRasEnumDevicesW}
  TRasEnumEntriesW = function(reserved: PWideChar; lpszPhonebook: PWideChar;
    lprasentryname: PRasEntryNameW; var lpcb: DWORD; var lpcEntries: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TRasEnumEntriesW}
  TRasGetConnectStatusW = function(RASConn: hrasConn; RASConnStatus: PRASConnStatusW): DWORD; stdcall;
  {$EXTERNALSYM TRasGetConnectStatusW}
  TRasGetEntryDialParamsW = function(lpszPhonebook: PWideChar;
    var lprasdialparams: TRasDialParamsW; var lpfPassword: BOOL): DWORD; stdcall;
  {$EXTERNALSYM TRasGetEntryDialParamsW}
  TRasGetEntryPropertiesW = function(lpszPhonebook, szEntry: PWideChar; lpbEntry: PWideChar;
    lpdwEntrySize: PWideChar; lpbDeviceInfo: PWideChar; lpdwDeviceInfoSize: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TRasGetEntryPropertiesW}
  TRasGetErrorStringW = function(ErrorCode: DWord; szErrorString: PWideChar;
    BufSize: DWord): DWORD; stdcall;
  {$EXTERNALSYM TRasGetErrorStringW}
  TRasGetProjectionInfoW = function(hConn: HRasConn; rasproj: TRasProjection;
   lpProjection: Pointer; var lpcb: DWORD): DWORD; stdcall; // 3703
  {$EXTERNALSYM TRasGetProjectionInfoW}
  TRasHangUpW = function(RASConn: hrasConn): DWORD; stdcall;
  {$EXTERNALSYM TRasHangUpW}
  TRasRenameEntryW = function(lpszPhonebook, szEntryOld, szEntryNew: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TRasRenameEntryW}
  TRasSetEntryDialParamsW = function(lpszPhonebook: PWideChar; var lprasdialparams: TRasDialParamsW;
    fRemovePassword: BOOL): DWORD; stdcall;
  {$EXTERNALSYM TRasSetEntryDialParamsW}
  TRasSetEntryPropertiesW = function(lpszPhonebook, szEntry: PWideChar; lpbEntry: PWideChar;
    dwEntrySize: DWORD; lpbDeviceInfo: PWideChar; dwDeviceInfoSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TRasSetEntryPropertiesW}
  TRasSetSubEntryPropertiesW = function(lpszPhoneBook, lpszEntry: PWideChar; dwSubEntry: DWORD;
    lpRasSubEntry: PWideChar; dwcb: DWORD; lpbDeviceInfo: PChar; dwDeviceInfoSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TRasSetSubEntryPropertiesW}
  TRasValidateEntryNameW = function(lpszPhonebook, szEntry: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TRasValidateEntryNameW}
{$ENDIF}
  TReadConsoleInputW = function(hConsoleInput: THandle; var lpBuffer: TInputRecord;
    nLength: DWORD; var lpNumberOfEventsRead: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TReadConsoleInputW}
  TReadConsoleOutputCharacterW = function(hConsoleOutput: THandle; lpCharacter: PAnsiChar;
    nLength: DWORD; dwReadCoord: TCoord; var lpNumberOfCharsRead: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TReadConsoleOutputCharacterW}
  TReadConsoleOutputW = function(hConsoleOutput: THandle; lpBuffer: Pointer;
    dwBufferSize, dwBufferCoord: TCoord; var lpReadRegion: TSmallRect): BOOL; stdcall;
  {$EXTERNALSYM TReadConsoleOutputW}
  TReadConsoleW = function(hConsoleInput: THandle; lpBuffer: Pointer;
    nNumberOfCharsToRead: DWORD; var lpNumberOfCharsRead: DWORD; lpReserved: Pointer): BOOL; stdcall;
  {$EXTERNALSYM TReadConsoleW}
  TRegConnectRegistryW = function(lpMachineName: PWideChar; hKey: HKEY;
    var phkResult: HKEY): Longint; stdcall;
  {$EXTERNALSYM TRegConnectRegistryW}
  TRegCreateKeyExW = function(hKey: HKEY; lpSubKey: PWideChar; Reserved: DWORD;
    lpClass: PWideChar; dwOptions: DWORD; samDesired: REGSAM;
    lpSecurityAttributes: PSecurityAttributes; var phkResult: HKEY;
    lpdwDisposition: PDWORD): Longint; stdcall;
  {$EXTERNALSYM TRegCreateKeyExW}
  TRegCreateKeyW = function(hKey: HKEY; lpSubKey: PWideChar;
    var phkResult: HKEY): Longint; stdcall;
  {$EXTERNALSYM TRegCreateKeyW}
  TRegDeleteKeyW = function(hKey: HKEY; lpSubKey: PWideChar): Longint; stdcall;
  {$EXTERNALSYM TRegDeleteKeyW}
  TRegDeleteValueW = function(hKey: HKEY; lpValueName: PWideChar): Longint; stdcall;
  {$EXTERNALSYM TRegDeleteValueW}
  TRegEnumKeyExW = function(hKey: HKEY; dwIndex: DWORD; lpName: PWideChar;
    var lpcbName: DWORD; lpReserved: Pointer; lpClass: PWideChar;
    lpcbClass: PDWORD; lpftLastWriteTime: PFileTime): Longint; stdcall;
  {$EXTERNALSYM TRegEnumKeyExW}
  TRegEnumKeyW = function(hKey: HKEY; dwIndex: DWORD; lpName: PWideChar; cbName: DWORD): Longint; stdcall;
  {$EXTERNALSYM TRegEnumKeyW}
  TRegEnumValueW = function(hKey: HKEY; dwIndex: DWORD; lpValueName: PChar;
    var lpcbValueName: DWORD; lpReserved: Pointer; lpType: PDWORD; lpData: PByte;
    lpcbData: PDWORD): Longint; stdcall;
  {$EXTERNALSYM TRegEnumValueW}
  TRegLoadKeyW = function(hKey: HKEY; lpSubKey, lpFile: PWideChar): Longint; stdcall;
  {$EXTERNALSYM TRegLoadKeyW}
  TRegOpenKeyExW = function(hKey: HKEY; lpSubKey: PWideChar; ulOptions: DWORD;
    samDesired: REGSAM; var phkResult: HKEY): Longint; stdcall;
  {$EXTERNALSYM TRegOpenKeyExW}
  TRegOpenKeyW = function(hKey: HKEY; lpSubKey: PWideChar; var phkResult: HKEY): Longint; stdcall;
  {$EXTERNALSYM TRegOpenKeyW}
  TRegQueryInfoKeyW = function(hKey: HKEY; lpClass: PChar; lpcbClass: PDWORD;
    lpReserved: Pointer; lpcSubKeys, lpcbMaxSubKeyLen, lpcbMaxClassLen, lpcValues,
    lpcbMaxValueNameLen, lpcbMaxValueLen, lpcbSecurityDescriptor: PDWORD;
    lpftLastWriteTime: PFileTime): Longint; stdcall;
  {$EXTERNALSYM TRegQueryInfoKeyW}
  TRegQueryMultipleValuesW = function(hKey: HKEY; var ValList;
    NumVals: DWORD; lpValueBuf: PWideChar; var ldwTotsize: DWORD): Longint; stdcall;
  {$EXTERNALSYM TRegQueryMultipleValuesW}
  TRegQueryValueExW = function(hKey: HKEY; lpValueName: PWideChar;
    lpReserved: Pointer; lpType: PDWORD; lpData: PByte; lpcbData: PDWORD): Longint;
    stdcall;
  {$EXTERNALSYM TRegQueryValueExW}
  TRegQueryValueW = function(hKey: HKEY; lpSubKey: PWideChar;
    lpValue: PWideChar; var lpcbValue: Longint): Longint; stdcall;
  {$EXTERNALSYM TRegQueryValueW}
  TRegReplaceKeyW = function(hKey: HKEY; lpSubKey: PWideChar; lpNewFile: PWideChar;
    lpOldFile: PWideChar): Longint; stdcall;
  {$EXTERNALSYM TRegReplaceKeyW}
  TRegSaveKeyW = function(hKey: HKEY; lpFile: PWideChar;
    lpSecurityAttributes: PSecurityAttributes): Longint; stdcall;
  {$EXTERNALSYM TRegSaveKeyW}
  TRegSetValueExW = function(hKey: HKEY; lpValueName: PWideChar; Reserved: DWORD;
    dwType: DWORD; lpData: Pointer; cbData: DWORD): Longint; stdcall;
  {$EXTERNALSYM TRegSetValueExW}
  TRegSetValueW = function(hKey: HKEY; lpSubKey: PWideChar;
    dwType: DWORD; lpData: PWideChar; cbData: DWORD): Longint; stdcall;
  {$EXTERNALSYM TRegSetValueW}
  TRegUnLoadKeyW = function(hKey: HKEY; lpSubKey: PWideChar): Longint; stdcall;
  {$EXTERNALSYM TRegUnLoadKeyW}
  TRegisterClassExW = function(const WndClass: TWndClassExW): ATOM; stdcall;
  {$EXTERNALSYM TRegisterClassExW}
  TRegisterClassW = function(const lpWndClass: TWndClassW): ATOM; stdcall;
  {$EXTERNALSYM TRegisterClassW}
  TRegisterClipboardFormatW = function(lpszFormat: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TRegisterClipboardFormatW}
  TRegisterDeviceNotificationW = function(hRecipient: THandle; NotificationFilter: Pointer; Flags: DWORD): HDEVNOTIFY; stdcall;
  {$EXTERNALSYM TRegisterDeviceNotificationW}
  TRegisterWindowMessageW = function(lpString: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TRegisterWindowMessageW}
  TRemoveDirectoryW = function(lpPathName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TRemoveDirectoryW}
  TRemoveFontResourceW = function(FileName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TRemoveFontResourceW}
  TRemovePropA = function(hWnd: HWND; lpString: PAnsiChar): THandle; stdcall;
  {$EXTERNALSYM TRemovePropA}
  TRemovePropW = function(hWnd: HWND; lpString: PWideChar): THandle; stdcall;
  {$EXTERNALSYM TRemovePropW}
  TReplaceTextW = function(var FindReplace: TFindReplaceW): HWND; stdcall;
  {$EXTERNALSYM TReplaceTextW}
  TResetDCW = function(DC: HDC; const InitData: TDeviceModeW): HDC; stdcall;
  {$EXTERNALSYM TResetDCW}
  TResetPrinterW = function(hPrinter: THandle; pDefault: PPrinterDefaultsW): BOOL; stdcall;
  {$EXTERNALSYM TResetPrinterW}
  TSHBrowseForFolderW = function(var lpbi: TBrowseInfoW): PItemIDList; stdcall;
  {$EXTERNALSYM TSHBrowseForFolderW}
  TSHChangeNotify = procedure(wEventId: Longint; uFlags: DWORD;
    dwItem1, dwItem2: Pointer); stdcall;
  {$EXTERNALSYM TSHChangeNotify}
  TSHFileOperationW = function(const lpFileOp: TSHFileOpStructW): Integer; stdcall;
  {$EXTERNALSYM TSHFileOperationW}
  TShGetFileInfoW = function(pszPath: PWideChar; dwFileAttributes: DWORD;
    var psfi: TSHFileInfoW; cbFileInfo, uFlags: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TShGetFileInfoW}
  TSHGetNewLinkInfoW = function(pszLinkTo, pszDir, pszName: PWideChar; var pfMustCopy: BOOL;
    uFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TSHGetNewLinkInfoW}
  TSHGetPathFromIDListW = function(pidl: PItemIDList; pszPath: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSHGetPathFromIDListW}
  TScrollConsoleScreenBufferW = function(hConsoleOutput: THandle;
    const lpScrollRectangle: TSmallRect; lpClipRectangle: PSmallRect;
    dwDestinationOrigin: TCoord; var lpFill: TCharInfo): BOOL; stdcall;
  {$EXTERNALSYM TScrollConsoleScreenBufferW}
  TSearchPathW = function(lpPath, lpFileName, lpExtension: PWideChar;
    nBufferLength: DWORD; lpBuffer: PWideChar; var lpFilePart: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TSearchPathW}
  TSendDlgItemMessageW = function(hDlg: HWND; nIDDlgItem: Integer;
    Msg: DWORD; wParam: WPARAM; lParam: LPARAM): Longint; stdcall;
  {$EXTERNALSYM TSendDlgItemMessageW}
  TSendMessageCallbackW = function(hWnd: HWND; Msg: DWORD; wParam: WPARAM;
    lParam: LPARAM; lpResultCallBack: TFNSendAsyncProc; dwData: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TSendMessageCallbackW}
  TSendMessageTimeoutW = function(hWnd: HWND; Msg: DWORD; wParam: WPARAM;
    lParam: LPARAM; fuFlags, uTimeout: DWORD; var lpdwResult: DWORD): LRESULT; stdcall;
  {$EXTERNALSYM TSendMessageTimeoutW}
  TSendMessageW = function(hWnd: HWND; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  {$EXTERNALSYM TSendMessageW}
  TSendNotifyMessageW = function(hWnd: HWND; Msg: DWORD; wParam: WPARAM;
    lParam: LPARAM): BOOL; stdcall;
  {$EXTERNALSYM TSendNotifyMessageW}
  TSetCalendarInfoW = function(Locale: LCID; Calendar: CALID; CalType: CALTYPE;
    lpCalData: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetCalendarInfoW}
  TSetClassLongW = function(hWnd: HWND; nIndex: Integer; dwNewLong: Longint): DWORD; stdcall;
  {$EXTERNALSYM TSetClassLongW}
  TSetComputerNameW = function(lpComputerName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetComputerNameW}
  TSetConsoleTitleW = function(lpConsoleTitle: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetConsoleTitleW}
  TSetCurrentDirectoryW = function(lpPathName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetCurrentDirectoryW}
  TSetDefaultCommConfigW = function(lpszName: PWideChar; lpCC: PCommConfig; dwSize: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TSetDefaultCommConfigW}
  TSetDlgItemTextW = function(hDlg: HWND; nIDDlgItem: Integer; lpString: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetDlgItemTextW}
  TSetEnvironmentVariableW = function(lpName, lpValue: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetEnvironmentVariableW}
  TSetFileAttributesW = function(lpFileName: PWideChar; dwFileAttributes: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TSetFileAttributesW}
  TSetICMProfileW = function(DC: HDC; Name: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetICMProfileW}
  TSetJobW = function(hPrinter: THandle; JobId: DWORD; Level: DWORD; pJob: Pointer; Command: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TSetJobW}
  TSetLocaleInfoW = function(Locale: LCID; LCType: LCTYPE; lpLCData: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetLocaleInfoW}
  TSetMenuItemInfoW = function(p1: HMENU; p2: DWORD; p3: BOOL; const p4: TMenuItemInfoW): BOOL; stdcall;
  {$EXTERNALSYM TSetMenuItemInfoW}
  TSetPrinterDataW = function(hPrinter: THandle; pValueName: PWideChar; dwType: DWORD; pData: Pointer; cbData: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TSetPrinterDataW}
  TSetPrinterW = function(hPrinter: THandle; Level: DWORD; pPrinter: Pointer; Command: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TSetPrinterW}
  TSetPropA = function(hWnd: HWND; lpString: PAnsiChar; hData: THandle): BOOL; stdcall;
  {$EXTERNALSYM TSetPropA}
  TSetPropW = function(hWnd: HWND; lpString: PWideChar; hData: THandle): BOOL; stdcall;
  {$EXTERNALSYM TSetPropW}
  TSetVolumeLabelW = function(lpRootPathName: PWideChar; lpVolumeName: PAnsiChar): BOOL; stdcall;
  {$EXTERNALSYM TSetVolumeLabelW}
  TSetWindowLongA = function(hWnd: HWND; nIndex: Integer; dwNewLong: Longint): Longint; stdcall;
  {$EXTERNALSYM TSetWindowLongA}
  TSetWindowLongW = function(hWnd: HWND; nIndex: Integer; dwNewLong: Longint): Longint; stdcall;
  {$EXTERNALSYM TSetWindowLongW}
  TSetWindowTextW = function(hWnd: HWND; lpString: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TSetWindowTextW}
  TSetWindowsHookExW = function(idHook: Integer; lpfn: TFNHookProc; hmod: HINST; dwThreadId: DWORD): HHOOK; stdcall;
  {$EXTERNALSYM TSetWindowsHookExW}
  TSetWindowsHookW = function(nFilterType: Integer; pfnFilterProc: TFNHookProc): HHOOK; stdcall;
  {$EXTERNALSYM TSetWindowsHookW}
  TShellAboutW = function(Wnd: HWND; szApp, szOtherStuff: PWideChar; Icon: HICON): Integer; stdcall;
  {$EXTERNALSYM TShellAboutW}
  TShellExecuteExW = function(lpExecInfo: PShellExecuteInfoW):BOOL; stdcall;
  {$EXTERNALSYM TShellExecuteExW}
  TShellExecuteW = function(hWnd: HWND; Operation, FileName, Parameters,
    Directory: PWideChar; ShowCmd: Integer): HINST; stdcall;
  {$EXTERNALSYM TShellExecuteW}
  TShell_NotifyIconW = function(dwMessage: DWORD; lpData: PNotifyIconDataW): BOOL; stdcall;
  {$EXTERNALSYM TShell_NotifyIconW}
  TStartDocPrinterW = function(hPrinter: THandle; Level: DWORD; pDocInfo: Pointer): DWORD; stdcall;
  {$EXTERNALSYM TStartDocPrinterW}
  TStartDocW = function(DC: HDC; const p2: TDocInfoW): Integer; stdcall;
  {$EXTERNALSYM TStartDocW}
  TSystemParametersInfoW = function(uiAction, uiParam: DWORD;
    pvParam: Pointer; fWinIni: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TSystemParametersInfoW}
  TTabbedTextOutW = function(hDC: HDC; X, Y: Integer; lpString: PWideChar; nCount, nTabPositions: Integer;
    var lpnTabStopPositions; nTabOrigin: Integer): Longint; stdcall;
  {$EXTERNALSYM TTabbedTextOutW}
  TTextOutW = function(DC: HDC; X, Y: Integer; Str: PWideChar; Count: Integer): BOOL; stdcall;
  {$EXTERNALSYM TTextOutW}
  TTranslateAcceleratorW = function(hWnd: HWND; hAccTable: HACCEL; var lpMsg: TMsg): Integer; stdcall;
  {$EXTERNALSYM TTranslateAcceleratorW}
  TUnregisterClassW = function(lpClassName: PWideChar; hInstance: HINST): BOOL; stdcall;
  {$EXTERNALSYM TUnregisterClassW}
  TUpdateICMRegKeyW = function(p1: DWORD; p2, p3: PWideChar; p4: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TUpdateICMRegKeyW}
  TUpdateResourceA = function(hUpdate: THandle; lpType, lpName: PChar; // 3703
    wLanguage: Word; lpData: Pointer; cbData: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TUpdateResourceA}
  TUpdateResourceW = function(hUpdate: THandle; lpType, lpName: PWideChar;
    wLanguage: Word; lpData: Pointer; cbData: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TUpdateResourceW}
  TVerFindFileW = function(uFlags: DWORD; szFileName, szWinDir, szAppDir, szCurDir: PWideChar;
    var lpuCurDirLen: DWORD; szDestDir: PWideChar; var lpuDestDirLen: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TVerFindFileW}
  TVerInstallFileW = function(uFlags: DWORD;
    szSrcFileName, szDestFileName, szSrcDir, szDestDir, szCurDir, szTmpFile: PWideChar;
    var lpuTmpFileLen: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TVerInstallFileW}
  TVerLanguageNameW = function(wLang: DWORD; szLang: PWideChar; nSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TVerLanguageNameW}
  TVerQueryValueW = function(pBlock: Pointer; lpSubBlock: PWideChar;
    var lplpBuffer: Pointer; var puLen: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TVerQueryValueW}
  TVkKeyScanExW = function(ch: WideChar; dwhkl: HKL): SHORT; stdcall;
  {$EXTERNALSYM TVkKeyScanExW}
  TVkKeyScanW = function(ch: WideChar): SHORT; stdcall;
  {$EXTERNALSYM TVkKeyScanW}
  TWNetAddConnection2W = function(var lpNetResource: TNetResourceW;
    lpPassword, lpUserName: PWideChar; dwFlags: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetAddConnection2W}
  TWNetAddConnection3W = function(hwndOwner: HWND; var lpNetResource: TNetResourceW;
    lpPassword, lpUserName: PWideChar; dwFlags: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetAddConnection3W}
  TWNetAddConnectionW = function(lpRemoteName, lpPassword, lpLocalName: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TWNetAddConnectionW}
  TWNetCancelConnection2W = function(lpName: PWideChar; dwFlags: DWORD; fForce: BOOL): DWORD; stdcall;
  {$EXTERNALSYM TWNetCancelConnection2W}
  TWNetCancelConnectionW = function(lpName: PWideChar; fForce: BOOL): DWORD; stdcall;
  {$EXTERNALSYM TWNetCancelConnectionW}
  TWNetConnectionDialog1W = function(var lpConnDlgStruct: TConnectDlgStruct): DWORD; stdcall;
  {$EXTERNALSYM TWNetConnectionDialog1W}
  TWNetDisconnectDialog1W = function(var lpConnDlgStruct: TDiscDlgStructW): DWORD; stdcall;
  {$EXTERNALSYM TWNetDisconnectDialog1W}
  TWNetEnumResourceW = function(hEnum: THandle; var lpcCount: DWORD;
    lpBuffer: Pointer; var lpBufferSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetEnumResourceW}
  TWNetGetConnectionW = function(lpLocalName: PWideChar;
    lpRemoteName: PWideChar; var lpnLength: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetGetConnectionW}
  TWNetGetLastErrorW = function(var lpError: DWORD; lpErrorBuf: PWideChar;
    nErrorBufSize: DWORD; lpNameBuf: PWideChar; nNameBufSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetGetLastErrorW}
  TWNetGetNetworkInformationW = function(lpProvider: PWideChar;
    var lpNetInfoStruct: TNetInfoStruct): DWORD; stdcall;
  {$EXTERNALSYM TWNetGetNetworkInformationW}
  TWNetGetProviderNameW = function(dwNetType: DWORD; lpProviderName: PWideChar;
    var lpBufferSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetGetProviderNameW}
  TWNetGetResourceInformationW = function(const lpNetResource: NETRESOURCEW;
    lpBuffer: Pointer; var lpcbBuffer: DWORD; var lplpSystem: PWideChar): DWORD; stdcall;
  {$EXTERNALSYM TWNetGetResourceInformationW}
  TWNetGetResourceParentW = function(lpNetResource: PNetResourceW;
    lpBuffer: Pointer; var cbBuffer: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetGetResourceParentW}
  TWNetGetUniversalNameW = function(lpLocalPath: PWideChar; dwInfoLevel: DWORD;
    lpBuffer: Pointer; var lpBufferSize: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetGetUniversalNameW}
  TWNetGetUserW = function(lpName: PWideChar; lpUserName: PWideChar; var lpnLength: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetGetUserW}
  TWNetOpenEnumW = function(dwScope, dwType, dwUsage: DWORD;
    lpNetResource: PNetResourceW; var lphEnum: THandle): DWORD; stdcall;
  {$EXTERNALSYM TWNetOpenEnumW}
  TWNetUseConnectionW = function(hwndOwner: HWND;
    var lpNetResource: TNetResourceW; lpUserID: PWideChar;
    lpPassword: PWideChar; dwFlags: DWORD; lpAccessName: PWideChar;
    var lpBufferSize: DWORD; var lpResult: DWORD): DWORD; stdcall;
  {$EXTERNALSYM TWNetUseConnectionW}
  TWaitNamedPipeW = function(lpNamedPipeName: PWideChar; nTimeOut: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TWaitNamedPipeW}
  TWideCharToMultiByte = function(CodePage: DWORD; dwFlags: DWORD;
    lpWideCharStr: PWideChar; cchWideChar: Integer; lpMultiByteStr: LPSTR;
    cchMultiByte: Integer; lpDefaultChar: PChar; lpUsedDefaultChar: PBOOL): Integer; stdcall;
  {$EXTERNALSYM TWideCharToMultiByte}
  TWinHelpW = function(hWndMain: HWND; lpszHelp: PWideChar; uCommand: DWORD; dwData: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TWinHelpW}
  TWriteConsoleInputW = function(hConsoleInput: THandle; const lpBuffer: TInputRecord;
    nLength: DWORD; var lpNumberOfEventsWritten: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TWriteConsoleInputW}
  TWriteConsoleOutputCharacterW = function(hConsoleOutput: THandle;lpCharacter: PWideChar;
    nLength: DWORD; dwWriteCoord: TCoord; var lpNumberOfCharsWritten: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TWriteConsoleOutputCharacterW}
  TWriteConsoleOutputW = function(hConsoleOutput: THandle; lpBuffer: Pointer;
    dwBufferSize, dwBufferCoord: TCoord; var lpWriteRegion: TSmallRect): BOOL; stdcall;
  {$EXTERNALSYM TWriteConsoleOutputW}
  TWriteConsoleW = function(hConsoleOutput: THandle; const lpBuffer: Pointer;
    nNumberOfCharsToWrite: DWORD; var lpNumberOfCharsWritten: DWORD; lpReserved: Pointer): BOOL; stdcall;
  {$EXTERNALSYM TWriteConsoleW}
  TWritePrivateProfileSectionW = function(lpAppName, lpString, lpFileName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TWritePrivateProfileSectionW}
  TWritePrivateProfileStringW = function(lpAppName, lpKeyName, lpString, lpFileName: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TWritePrivateProfileStringW}
  TWritePrivateProfileStructW = function(lpszSection, lpszKey: PWideChar;
    lpStruct: Pointer; uSizeStruct: DWORD; szFile: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TWritePrivateProfileStructW}
  TWriteProfileSectionW = function(lpAppName, lpString: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TWriteProfileSectionW}
  TWriteProfileStringW = function(lpAppName, lpKeyName, lpString: PWideChar): BOOL; stdcall;
  {$EXTERNALSYM TWriteProfileStringW}
  TauxGetDevCapsW = function(uDeviceID: DWORD; lpCaps: PAuxCapsW; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TauxGetDevCapsW}
  TcapCreateCaptureWindowW = function(lpszWindowName: PWideChar; dwStyle: DWORD;
    x, y, nWidth, nHeight: Integer; hWnd: HWND; nID: Integer): HWND; stdcall;
  {$EXTERNALSYM TcapCreateCaptureWindowW}
  TcapGetDriverDescriptionW = function(wDriverIndex: WORD; lpszName: PWideChar;
    cbName: Integer; lpszVer: PWideChar; cbVer: Integer): BOOL; stdcall;
  {$EXTERNALSYM TcapGetDriverDescriptionW}
  TjoyGetDevCapsW = function(uJoyID: DWORD; lpCaps: PJoyCapsW; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TjoyGetDevCapsW}
  TlstrcatW = function(lpString1, lpString2: PWideChar): PWideChar; stdcall;
  {$EXTERNALSYM TlstrcatW}
  TlstrcmpW = function(lpString1, lpString2: PWideChar): Integer; stdcall;
  {$EXTERNALSYM TlstrcmpW}
  TlstrcmpiW = function(lpString1, lpString2: PWideChar): Integer; stdcall;
  {$EXTERNALSYM TlstrcmpiW}
  TlstrcpyW = function(lpString1, lpString2: PWideChar): PWideChar; stdcall;
  {$EXTERNALSYM TlstrcpyW}
  TlstrcpynW = function(lpString1, lpString2: PWideChar; iMaxLength: Integer): PWideChar; stdcall;
  {$EXTERNALSYM TlstrcpynW}
  TlstrlenW = function(lpString: PWideChar): Integer; stdcall;
  {$EXTERNALSYM TlstrlenW}
  TmciGetDeviceIDW = function(pszDevice: PWideChar): MCIDEVICEID; stdcall;
  {$EXTERNALSYM TmciGetDeviceIDW}
  TmciGetErrorStringW = function(mcierr: MCIERROR; pszText: PWideChar; uLength: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TmciGetErrorStringW}
  TmciSendCommandW = function(mciId: MCIDEVICEID; uMessage: DWORD;
    dwParam1, dwParam2: DWORD): MCIERROR; stdcall;
  {$EXTERNALSYM TmciSendCommandW}
  TmciSendStringW = function(lpstrCommand, lpstrReturnString: PWideChar;
    uReturnLength: DWORD; hWndCallback: HWND): MCIERROR; stdcall;
  {$EXTERNALSYM TmciSendStringW}
  TmidiInGetDevCapsW = function(DeviceID: DWORD; lpCaps: PMidiInCapsW; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmidiInGetDevCapsW}
  TmidiInGetErrorTextW = function(mmrError: MMRESULT; pszText: PWideChar; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmidiInGetErrorTextW}
  TmidiOutGetDevCapsW = function(uDeviceID: DWORD; lpCaps: PMidiOutCapsW; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmidiOutGetDevCapsW}
  TmidiOutGetErrorTextW = function(mmrError: MMRESULT; pszText: PWideChar; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmidiOutGetErrorTextW}
  TmixerGetControlDetailsW = function(hmxobj: HMIXEROBJ; pmxcd: PMixerControlDetails; fdwDetails: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmixerGetControlDetailsW}
  TmixerGetDevCapsW = function(uMxId: DWORD; pmxcaps: PMixerCapsW; cbmxcaps: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmixerGetDevCapsW}
  TmixerGetLineControlsW = function(hmxobj: HMIXEROBJ; pmxlc: PMixerLineControlsW; fdwControls: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmixerGetLineControlsW}
  TmixerGetLineInfoW = function(hmxobj: HMIXEROBJ; pmxl: PMixerLineW;
    fdwInfo: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmixerGetLineInfoW}
  TmmioInstallIOProcW = function(fccIOProc: FOURCC; pIOProc: TFNMMIOProc;
    dwFlags: DWORD): TFNMMIOProc; stdcall;
  {$EXTERNALSYM TmmioInstallIOProcW}
  TmmioOpenW = function(szFileName: PChar; lpmmioinfo: PMMIOInfo;
    dwOpenFlags: DWORD): HMMIO; stdcall;
  {$EXTERNALSYM TmmioOpenW}
  TmmioRenameW = function(szFileName, szNewFileName: PWideChar;
    lpmmioinfo: PMMIOInfo; dwRenameFlags: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TmmioRenameW}
  TmmioStringToFOURCCW = function(sz: PWideChar; uFlags: DWORD): FOURCC; stdcall;
  {$EXTERNALSYM TmmioStringToFOURCCW}
  TsndPlaySoundW = function(lpszSoundName: PWideChar; uFlags: DWORD): BOOL; stdcall;
  {$EXTERNALSYM TsndPlaySoundW}
  TwaveInGetDevCapsW = function(hwo: HWAVEOUT; lpCaps: PWaveInCapsW; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TwaveInGetDevCapsW}
  TwaveInGetErrorTextW = function(mmrError: MMRESULT; lpText: PWideChar; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TwaveInGetErrorTextW}
  TwaveOutGetDevCapsW = function(uDeviceID: DWORD; lpCaps: PWaveOutCapsW; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TwaveOutGetDevCapsW}
  TwaveOutGetErrorTextW = function(mmrError: MMRESULT; lpText: PWideChar; uSize: DWORD): MMRESULT; stdcall;
  {$EXTERNALSYM TwaveOutGetErrorTextW}
  TwsprintfW = function(Output: PWideChar; Format: PWideChar): Integer; stdcall;
  {$EXTERNALSYM TwsprintfW}
  TwvsprintfW = function(Output: PWideChar; Format: PWideChar; arglist: va_list): Integer; stdcall;
  {$EXTERNALSYM TwvsprintfW}

var
{$IFDEF UNIJWA}
  AcquireCredentialsHandleW: TAcquireCredentialsHandleW = nil;
  {$EXTERNALSYM AcquireCredentialsHandleW}
{$ENDIF}
  AddAtomW: TAddAtomW = nil;
  {$EXTERNALSYM AddAtomW}
  AddFontResourceW: TAddFontResourceW = nil;
  {$EXTERNALSYM AddFontResourceW}
  AddJobW: TAddJobW = nil;
  {$EXTERNALSYM AddJobW}
  AddMonitorW: TAddMonitorW = nil;
  {$EXTERNALSYM AddMonitorW}
  AddPortW: TAddPortW = nil;
  {$EXTERNALSYM AddPortW}
  AddPrintProcessorW: TAddPrintProcessorW = nil;
  {$EXTERNALSYM AddPrintProcessorW}
  AddPrintProvidorW: TAddPrintProvidorW = nil;
  {$EXTERNALSYM AddPrintProvidorW}
  AddPrinterDriverW: TAddPrinterDriverW = nil;
  {$EXTERNALSYM AddPrinterDriverW}
  AddPrinterW: TAddPrinterW = nil;
  {$EXTERNALSYM AddPrinterW}
  AdvancedDocumentPropertiesW: TAdvancedDocumentPropertiesW = nil;
  {$EXTERNALSYM AdvancedDocumentPropertiesW}
  AppendMenuW: TAppendMenuW = nil;
  {$EXTERNALSYM AppendMenuW}
  BeginUpdateResourceA: TBeginUpdateResourceA = nil; // 3703
  {$EXTERNALSYM BeginUpdateResourceA}
  BeginUpdateResourceW: TBeginUpdateResourceW = nil;
  {$EXTERNALSYM BeginUpdateResourceW}
  BroadcastSystemMessageW: TBroadcastSystemMessageW = nil;
  {$EXTERNALSYM BroadcastSystemMessageW}
  BuildCommDCBAndTimeoutsW: TBuildCommDCBAndTimeoutsW = nil;
  {$EXTERNALSYM BuildCommDCBAndTimeoutsW}
  BuildCommDCBW: TBuildCommDCBW = nil;
  {$EXTERNALSYM BuildCommDCBW}
  CallMsgFilterW: TCallMsgFilterW = nil;
  {$EXTERNALSYM CallMsgFilterW}
  CallNamedPipeW: TCallNamedPipeW = nil;
  {$EXTERNALSYM CallNamedPipeW}
  CallWindowProcA: TCallWindowProcA = nil;
  {$EXTERNALSYM CallWindowProcA}
  CallWindowProcW: TCallWindowProcW = nil;
  {$EXTERNALSYM CallWindowProcW}
  ChangeDisplaySettingsExW: TChangeDisplaySettingsExW = nil;
  {$EXTERNALSYM ChangeDisplaySettingsExW}
  ChangeDisplaySettingsW: TChangeDisplaySettingsW = nil;
  {$EXTERNALSYM ChangeDisplaySettingsW}
  ChangeMenuW: TChangeMenuW = nil;
  {$EXTERNALSYM ChangeMenuW}
  CharLowerBuffW: TCharLowerBuffW = nil;
  {$EXTERNALSYM CharLowerBuffW}
  CharLowerW: TCharLowerW = nil;
  {$EXTERNALSYM CharLowerW}
  CharNextW: TCharNextW = nil;
  {$EXTERNALSYM CharNextW}
  CharPrevW: TCharPrevW = nil;
  {$EXTERNALSYM CharPrevW}
  CharToOemBuffW: TCharToOemBuffW = nil;
  {$EXTERNALSYM CharToOemBuffW}
  CharToOemW: TCharToOemW = nil;
  {$EXTERNALSYM CharToOemW}
  CharUpperBuffW: TCharUpperBuffW = nil;
  {$EXTERNALSYM CharUpperBuffW}
  CharUpperW: TCharUpperW = nil;
  {$EXTERNALSYM CharUpperW}
  ChooseColorW: TChooseColorWFunc = nil;
  {$EXTERNALSYM ChooseColorW}
  ChooseFontW: TChooseFontWFunc = nil;
  {$EXTERNALSYM ChooseFontW}
  CommConfigDialogW: TCommConfigDialogW = nil;
  {$EXTERNALSYM CommConfigDialogW}
  CompareStringW: TCompareStringW = nil;
  {$EXTERNALSYM CompareStringW}
  ConfigurePortW: TConfigurePortW = nil;
  {$EXTERNALSYM ConfigurePortW}
  CopyAcceleratorTableW: TCopyAcceleratorTableW = nil;
  {$EXTERNALSYM CopyAcceleratorTableW}
  CopyEnhMetaFileW: TCopyEnhMetaFileW = nil;
  {$EXTERNALSYM CopyEnhMetaFileW}
  CopyFileExW: TCopyFileExW = nil;
  {$EXTERNALSYM CopyFileExW}
  CopyFileW: TCopyFileW = nil;
  {$EXTERNALSYM CopyFileW}
  CopyMetaFileW: TCopyMetaFileW = nil;
  {$EXTERNALSYM CopyMetaFileW}
  CreateAcceleratorTableW: TCreateAcceleratorTableW = nil;
  {$EXTERNALSYM CreateAcceleratorTableW}
  CreateColorSpaceW: TCreateColorSpaceW = nil;
  {$EXTERNALSYM CreateColorSpaceW}
  CreateDCW: TCreateDCW = nil;
  {$EXTERNALSYM CreateDCW}
  CreateDialogIndirectParamW: TCreateDialogIndirectParamW = nil;
  {$EXTERNALSYM CreateDialogIndirectParamW}
  CreateDialogParamW: TCreateDialogParamW = nil;
  {$EXTERNALSYM CreateDialogParamW}
  CreateDirectoryExW: TCreateDirectoryExW = nil;
  {$EXTERNALSYM CreateDirectoryExW}
  CreateDirectoryW: TCreateDirectoryW = nil;
  {$EXTERNALSYM CreateDirectoryW}
  CreateEnhMetaFileW: TCreateEnhMetaFileW = nil;
  {$EXTERNALSYM CreateEnhMetaFileW}
  CreateEventW: TCreateEventW = nil;
  {$EXTERNALSYM CreateEventW}
  CreateFileMappingW: TCreateFileMappingW = nil;
  {$EXTERNALSYM CreateFileMappingW}
  CreateFileW: TCreateFileW = nil;
  {$EXTERNALSYM CreateFileW}
  CreateFontIndirectW: TCreateFontIndirectW = nil;
  {$EXTERNALSYM CreateFontIndirectW}
  CreateFontW: TCreateFontW = nil;
  {$EXTERNALSYM CreateFontW}
  CreateICW: TCreateICW = nil;
  {$EXTERNALSYM CreateICW}
  CreateMDIWindowW: TCreateMDIWindowW = nil;
  {$EXTERNALSYM CreateMDIWindowW}
  CreateMailslotW: TCreateMailslotW = nil;
  {$EXTERNALSYM CreateMailslotW}
  CreateMetaFileW: TCreateMetaFileW = nil;
  {$EXTERNALSYM CreateMetaFileW}
  CreateMutexW: TCreateMutexW = nil;
  {$EXTERNALSYM CreateMutexW}
  CreateNamedPipeW: TCreateNamedPipeW = nil;
  {$EXTERNALSYM CreateNamedPipeW}
  CreateProcessW: TCreateProcessW = nil;
  {$EXTERNALSYM CreateProcessW}
  CreateScalableFontResourceW: TCreateScalableFontResourceW = nil;
  {$EXTERNALSYM CreateScalableFontResourceW}
  CreateSemaphoreW: TCreateSemaphoreW = nil;
  {$EXTERNALSYM CreateSemaphoreW}
  CreateStdAccessibleProxyW: TCreateStdAccessibleProxyW = nil;
  {$EXTERNALSYM CreateStdAccessibleProxyW}
  CreateWaitableTimerW: TCreateWaitableTimerW = nil;
  {$EXTERNALSYM CreateWaitableTimerW}
  CreateWindowExW: TCreateWindowExW = nil;
  {$EXTERNALSYM CreateWindowExW}
{$IFDEF UNIJWA}
  CryptAcquireContextW: TCryptAcquireContextW = nil;
  {$EXTERNALSYM CryptAcquireContextW}
  CryptEnumProviderTypesW: TCryptEnumProviderTypesW = nil;
  {$EXTERNALSYM CryptEnumProviderTypesW}
  CryptEnumProvidersW: TCryptEnumProvidersW = nil;
  {$EXTERNALSYM CryptEnumProvidersW}
  CryptGetDefaultProviderW: TCryptGetDefaultProviderW = nil;
  {$EXTERNALSYM CryptGetDefaultProviderW}
  CryptSetProviderExW: TCryptSetProviderExW = nil;
  {$EXTERNALSYM CryptSetProviderExW}
  CryptSetProviderW: TCryptSetProviderW = nil;
  {$EXTERNALSYM CryptSetProviderW}
  CryptSignHashW: TCryptSignHashW = nil;
  {$EXTERNALSYM CryptSignHashW}
  CryptVerifySignatureW: TCryptVerifySignatureW = nil;
  {$EXTERNALSYM CryptVerifySignatureW}
{$ENDIF}
  DdeConnect: TDdeConnect = nil;
  {$EXTERNALSYM DdeConnect}
  DdeConnectList: TDdeConnectList = nil;
  {$EXTERNALSYM DdeConnectList}
  DdeCreateStringHandleW: TDdeCreateStringHandleW = nil;
  {$EXTERNALSYM DdeCreateStringHandleW}
  DdeInitializeW: TDdeInitializeW = nil;
  {$EXTERNALSYM DdeInitializeW}
  DdeQueryConvInfo: TDdeQueryConvInfo = nil;
  {$EXTERNALSYM DdeQueryConvInfo}
  DdeQueryStringW: TDdeQueryStringW = nil;
  {$EXTERNALSYM DdeQueryStringW}
  DefDlgProcW: TDefDlgProcW = nil;
  {$EXTERNALSYM DefDlgProcW}
  DefFrameProcW: TDefFrameProcW = nil;
  {$EXTERNALSYM DefFrameProcW}
  DefMDIChildProcW: TDefMDIChildProcW = nil;
  {$EXTERNALSYM DefMDIChildProcW}
  DefWindowProcW: TDefWindowProcW = nil;
  {$EXTERNALSYM DefWindowProcW}
  DeleteFileW: TDeleteFileW = nil;
  {$EXTERNALSYM DeleteFileW}
  DeleteMonitorW: TDeleteMonitorW = nil;
  {$EXTERNALSYM DeleteMonitorW}
  DeletePortW: TDeletePortW = nil;
  {$EXTERNALSYM DeletePortW}
  DeletePrintProcessorW: TDeletePrintProcessorW = nil;
  {$EXTERNALSYM DeletePrintProcessorW}
  DeletePrintProvidorW: TDeletePrintProvidorW = nil;
  {$EXTERNALSYM DeletePrintProvidorW}
  DeletePrinterDriverW: TDeletePrinterDriverW = nil;
  {$EXTERNALSYM DeletePrinterDriverW}
  DeviceCapabilitiesW: TDeviceCapabilitiesW = nil;
  {$EXTERNALSYM DeviceCapabilitiesW}
  DialogBoxIndirectParamW: TDialogBoxIndirectParamW = nil;
  {$EXTERNALSYM DialogBoxIndirectParamW}
  DialogBoxParamW: TDialogBoxParamW = nil;
  {$EXTERNALSYM DialogBoxParamW}
  DispatchMessageW: TDispatchMessageW = nil;
  {$EXTERNALSYM DispatchMessageW}
  DlgDirListComboBoxW: TDlgDirListComboBoxW = nil;
  {$EXTERNALSYM DlgDirListComboBoxW}
  DlgDirListW: TDlgDirListW = nil;
  {$EXTERNALSYM DlgDirListW}
  DlgDirSelectComboBoxExW: TDlgDirSelectComboBoxExW = nil;
  {$EXTERNALSYM DlgDirSelectComboBoxExW}
  DlgDirSelectExW: TDlgDirSelectExW = nil;
  {$EXTERNALSYM DlgDirSelectExW}
  DocumentPropertiesW: TDocumentPropertiesW = nil;
  {$EXTERNALSYM DocumentPropertiesW}
  DragQueryFileW: TDragQueryFileW = nil;
  {$EXTERNALSYM DragQueryFileW}
  DrawStateW: TDrawStateW = nil;
  {$EXTERNALSYM DrawStateW}
  DrawTextExW: TDrawTextExW = nil;
  {$EXTERNALSYM DrawTextExW}
  DrawTextW: TDrawTextW = nil;
  {$EXTERNALSYM DrawTextW}
  EnableWindow: TEnableWindoW = nil;
  {$EXTERNALSYM EnableWindow}
  EndUpdateResourceA: TEndUpdateResourceA = nil; // 3703
  {$EXTERNALSYM EndUpdateResourceA}
  EndUpdateResourceW: TEndUpdateResourceW = nil;
  {$EXTERNALSYM EndUpdateResourceW}
{$IFDEF UNIJWA}
  EnumCalendarInfoExW: TEnumCalendarInfoExW = nil;
  {$EXTERNALSYM EnumCalendarInfoExW}
{$ENDIF}
  EnumCalendarInfoW: TEnumCalendarInfoW = nil;
  {$EXTERNALSYM EnumCalendarInfoW}
  EnumClipboardFormats: TEnumClipboardFormats = nil;
  {$EXTERNALSYM EnumClipboardFormats}
{$IFDEF UNIJWA}
  EnumDateFormatsExW: TEnumDateFormatsExW = nil;
  {$EXTERNALSYM EnumDateFormatsExW}
{$ENDIF}
  EnumDateFormatsW: TEnumDateFormatsW = nil;
  {$EXTERNALSYM EnumDateFormatsW}
  EnumDisplayDevicesW: TEnumDisplayDevicesW = nil;
  {$EXTERNALSYM EnumDisplayDevicesW}
  EnumDisplaySettingsExW: TEnumDisplaySettingsExW = nil;
  {$EXTERNALSYM EnumDisplaySettingsExW}
  EnumDisplaySettingsW: TEnumDisplaySettingsW = nil;
  {$EXTERNALSYM EnumDisplaySettingsW}
  EnumFontFamiliesExW: TEnumFontFamiliesExW = nil;
  {$EXTERNALSYM EnumFontFamiliesExW}
  EnumFontFamiliesW: TEnumFontFamiliesW = nil;
  {$EXTERNALSYM EnumFontFamiliesW}
  EnumFontsW: TEnumFontsW = nil;
  {$EXTERNALSYM EnumFontsW}
  EnumICMProfilesW: TEnumICMProfilesW = nil;
  {$EXTERNALSYM EnumICMProfilesW}
  EnumMonitorsW: TEnumMonitorsW = nil;
  {$EXTERNALSYM EnumMonitorsW}
  EnumPortsW: TEnumPortsW = nil;
  {$EXTERNALSYM EnumPortsW}
  EnumPrintProcessorDatatypesW: TEnumPrintProcessorDatatypesW = nil;
  {$EXTERNALSYM EnumPrintProcessorDatatypesW}
  EnumPrintProcessorsW: TEnumPrintProcessorsW = nil;
  {$EXTERNALSYM EnumPrintProcessorsW}
  EnumPrinterDriversW: TEnumPrinterDriversW = nil;
  {$EXTERNALSYM EnumPrinterDriversW}
  EnumPrintersW: TEnumPrintersW = nil;
  {$EXTERNALSYM EnumPrintersW}
  EnumPropsA: TEnumPropsA = nil;
  {$EXTERNALSYM EnumPropsA}
  EnumPropsExA: TEnumPropsExA = nil;
  {$EXTERNALSYM EnumPropsExA}
  EnumPropsExW: TEnumPropsExW = nil;
  {$EXTERNALSYM EnumPropsExW}
  EnumPropsW: TEnumPropsW = nil;
  {$EXTERNALSYM EnumPropsW}
  EnumSystemCodePagesW: TEnumSystemCodePagesW = nil;
  {$EXTERNALSYM EnumSystemCodePagesW}
  EnumSystemLocalesW: TEnumSystemLocalesW = nil;
  {$EXTERNALSYM EnumSystemLocalesW}
  EnumTimeFormatsW: TEnumTimeFormatsW = nil;
  {$EXTERNALSYM EnumTimeFormatsW}
{$IFDEF UNIJWA}
  EnumerateSecurityPackagesW: TEnumerateSecurityPackagesW = nil;
  {$EXTERNALSYM EnumerateSecurityPackagesW}
{$ENDIF}
  ExpandEnvironmentStringsW: TExpandEnvironmentStringsW = nil;
  {$EXTERNALSYM ExpandEnvironmentStringsW}
  ExtTextOutW: TExtTextOutW = nil;
  {$EXTERNALSYM ExtTextOutW}
  ExtractIconExW: TExtractIconExW = nil;
  {$EXTERNALSYM ExtractIconExW}
  ExtractIconW: TExtractIconW = nil;
  {$EXTERNALSYM ExtractIconW}
  FatalAppExitW: TFatalAppExitW = nil;
  {$EXTERNALSYM FatalAppExitW}
  FillConsoleOutputCharacterW: TFillConsoleOutputCharacterW = nil;
  {$EXTERNALSYM FillConsoleOutputCharacterW}
  FindAtomW: TFindAtomW = nil;
  {$EXTERNALSYM FindAtomW}
  FindExecutableW: TFindExecutableW = nil;
  {$EXTERNALSYM FindExecutableW}
  FindFirstChangeNotificationW: TFindFirstChangeNotificationW = nil;
  {$EXTERNALSYM FindFirstChangeNotificationW}
  FindFirstFileW: TFindFirstFileW = nil;
  {$EXTERNALSYM FindFirstFileW}
  FindNextFileW: TFindNextFileW = nil;
  {$EXTERNALSYM FindNextFileW}
  FindResourceExW: TFindResourceExW = nil;
  {$EXTERNALSYM FindResourceExW}
  FindResourceW: TFindResourceW = nil;
  {$EXTERNALSYM FindResourceW}
  FindTextW: TFindTextW = nil;
  {$EXTERNALSYM FindTextW}
  FindWindowExW: TFindWindowExW = nil;
  {$EXTERNALSYM FindWindowExW}
  FindWindowW: TFindWindowW = nil;
  {$EXTERNALSYM FindWindowW}
  FormatMessageW: TFormatMessageW = nil;
  {$EXTERNALSYM FormatMessageW}
{$IFDEF UNIJWA}
  FreeContextBuffer: TFreeContextBuffer = nil;
  {$EXTERNALSYM FreeContextBuffer}
{$ENDIF}
  FreeEnvironmentStringsW: TFreeEnvironmentStringsW = nil;
  {$EXTERNALSYM FreeEnvironmentStringsW}
  GetAltTabInfoW: TGetAltTabInfoW = nil;
  {$EXTERNALSYM GetAltTabInfoW}
  GetAtomNameW: TGetAtomNameW = nil;
  {$EXTERNALSYM GetAtomNameW}
  GetCPInfo: TGetCPInfo = nil;
  {$EXTERNALSYM GetCPInfo}
{$IFDEF UNIJWA}
  GetCPInfoExW: TGetCPInfoExW = nil;
  {$EXTERNALSYM GetCPInfoExW}
{$ENDIF}
  GetCalendarInfoW: TGetCalendarInfoW = nil;
  {$EXTERNALSYM GetCalendarInfoW}
  GetCharABCWidthsFloatW: TGetCharABCWidthsFloatW = nil;
  {$EXTERNALSYM GetCharABCWidthsFloatW}
  GetCharABCWidthsW: TGetCharABCWidthsW = nil;
  {$EXTERNALSYM GetCharABCWidthsW}
  GetCharWidth32W: TGetCharWidth32W = nil;
  {$EXTERNALSYM GetCharWidth32W}
  GetCharWidthFloatW: TGetCharWidthFloatW = nil;
  {$EXTERNALSYM GetCharWidthFloatW}
  GetCharWidthW: TGetCharWidthW = nil;
  {$EXTERNALSYM GetCharWidthW}
  GetCharacterPlacementW: TGetCharacterPlacementW = nil;
  {$EXTERNALSYM GetCharacterPlacementW}
  GetClassInfoExW: TGetClassInfoExW = nil;
  {$EXTERNALSYM GetClassInfoExW}
  GetClassInfoW: TGetClassInfoW = nil;
  {$EXTERNALSYM GetClassInfoW}
  GetClassLongW: TGetClassLongW = nil;
  {$EXTERNALSYM GetClassLongW}
  GetClassNameW: TGetClassNameW = nil;
  {$EXTERNALSYM GetClassNameW}
  GetClipboardData: TGetClipboardData = nil;
  {$EXTERNALSYM GetClipboardData}
  GetClipboardFormatNameW: TGetClipboardFormatNameW = nil;
  {$EXTERNALSYM GetClipboardFormatNameW}
  GetComputerNameW: TGetComputerNameW = nil;
  {$EXTERNALSYM GetComputerNameW}
  GetConsoleTitleW: TGetConsoleTitleW = nil;
  {$EXTERNALSYM GetConsoleTitleW}
  GetCurrencyFormatW: TGetCurrencyFormatW = nil;
  {$EXTERNALSYM GetCurrencyFormatW}
  GetCurrentDirectoryW: TGetCurrentDirectoryW = nil;
  {$EXTERNALSYM GetCurrentDirectoryW}
  GetCurrentHwProfileW: TGetCurrentHwProfileW = nil;
  {$EXTERNALSYM GetCurrentHwProfileW}
  GetDateFormatW: TGetDateFormatW = nil;
  {$EXTERNALSYM GetDateFormatW}
  GetDefaultCommConfigW: TGetDefaultCommConfigW = nil;
  {$EXTERNALSYM GetDefaultCommConfigW}
  GetDiskFreeSpaceExW: TGetDiskFreeSpaceExW = nil;
  {$EXTERNALSYM GetDiskFreeSpaceExW}
  GetDiskFreeSpaceW: TGetDiskFreeSpaceW = nil;
  {$EXTERNALSYM GetDiskFreeSpaceW}
  GetDlgItemTextW: TGetDlgItemTextW = nil;
  {$EXTERNALSYM GetDlgItemTextW}
  GetDriveTypeW: TGetDriveTypeW = nil;
  {$EXTERNALSYM GetDriveTypeW}
  GetEnhMetaFileDescriptionW: TGetEnhMetaFileDescriptionW = nil;
  {$EXTERNALSYM GetEnhMetaFileDescriptionW}
  GetEnhMetaFileW: TGetEnhMetaFileW = nil;
  {$EXTERNALSYM GetEnhMetaFileW}
  GetEnvironmentStringsW: TGetEnvironmentStringsW = nil;
  {$EXTERNALSYM GetEnvironmentStringsW}
  GetEnvironmentVariableW: TGetEnvironmentVariableW = nil;
  {$EXTERNALSYM GetEnvironmentVariableW}
  GetFileAttributesExW: TGetFileAttributesExW = nil;
  {$EXTERNALSYM GetFileAttributesExW}
  GetFileAttributesW: TGetFileAttributesW = nil;
  {$EXTERNALSYM GetFileAttributesW}
  GetFileTitleW: TGetFileTitleW = nil;
  {$EXTERNALSYM GetFileTitleW}
  GetFileVersionInfoSizeW: TGetFileVersionInfoSizeW = nil;
  {$EXTERNALSYM GetFileVersionInfoSizeW}
  GetFileVersionInfoW: TGetFileVersionInfoW = nil;
  {$EXTERNALSYM GetFileVersionInfoW}
  GetFullPathNameW: TGetFullPathNameW = nil;
  {$EXTERNALSYM GetFullPathNameW}
  GetGlyphOutlineW: TGetGlyphOutlineW = nil;
  {$EXTERNALSYM GetGlyphOutlineW}
  GetICMProfileW: TGetICMProfileW = nil;
  {$EXTERNALSYM GetICMProfileW}
  GetJobW: TGetJobW = nil;
  {$EXTERNALSYM GetJobW}
{$IFDEF UNIJWA}
  GetKerningPairsW: TGetKerningPairsW = nil;
  {$EXTERNALSYM GetKerningPairsW}
{$ENDIF}
  GetKeyNameTextW: TGetKeyNameTextW = nil;
  {$EXTERNALSYM GetKeyNameTextW}
  GetKeyboardLayoutNameW: TGetKeyboardLayoutNameW = nil;
  {$EXTERNALSYM GetKeyboardLayoutNameW}
  GetLocaleInfoW: TGetLocaleInfoW = nil;
  {$EXTERNALSYM GetLocaleInfoW}
  GetLogColorSpaceW: TGetLogColorSpaceW = nil;
  {$EXTERNALSYM GetLogColorSpaceW}
  GetLogicalDriveStringsW: TGetLogicalDriveStringsW = nil;
  {$EXTERNALSYM GetLogicalDriveStringsW}
  GetLongPathNameW: TGetLongPathNameW = nil;
  {$EXTERNALSYM GetLongPathNameW}
  GetMenuItemInfoW: TGetMenuItemInfoW = nil;
  {$EXTERNALSYM GetMenuItemInfoW}
  GetMenuStringW: TGetMenuStringW = nil;
  {$EXTERNALSYM GetMenuStringW}
  GetMessageW: TGetMessageW = nil;
  {$EXTERNALSYM GetMessageW}
  GetMetaFileW: TGetMetaFileW = nil;
  {$EXTERNALSYM GetMetaFileW}
  GetModuleFileNameW: TGetModuleFileNameW = nil;
  {$EXTERNALSYM GetModuleFileNameW}
  GetModuleHandleW: TGetModuleHandleW = nil;
  {$EXTERNALSYM GetModuleHandleW}
  GetMonitorInfoW: TGetMonitorInfoW = nil;
  {$EXTERNALSYM GetMonitorInfoW}
  GetNamedPipeHandleStateW: TGetNamedPipeHandleStateW = nil;
  {$EXTERNALSYM GetNamedPipeHandleStateW}
  GetNumberFormatW: TGetNumberFormatW = nil;
  {$EXTERNALSYM GetNumberFormatW}
  GetObjectW: TGetObjectW = nil;
  {$EXTERNALSYM GetObjectW}
  GetOpenFileNamePreviewW: TGetOpenFileNamePreviewW = nil;
  {$EXTERNALSYM GetOpenFileNamePreviewW}
  GetOpenFileNameW: TGetOpenFileNameW = nil;
  {$EXTERNALSYM GetOpenFileNameW}
  GetOutlineTextMetricsW: TGetOutlineTextMetricsW = nil;
  {$EXTERNALSYM GetOutlineTextMetricsW}
  GetPrintProcessorDirectoryW: TGetPrintProcessorDirectoryW = nil;
  {$EXTERNALSYM GetPrintProcessorDirectoryW}
  GetPrinterDataW: TGetPrinterDataW = nil;
  {$EXTERNALSYM GetPrinterDataW}
  GetPrinterDriverDirectoryW: TGetPrinterDriverDirectoryW = nil;
  {$EXTERNALSYM GetPrinterDriverDirectoryW}
  GetPrinterDriverW: TGetPrinterDriverW = nil;
  {$EXTERNALSYM GetPrinterDriverW}
  GetPrinterW: TGetPrinterW = nil;
  {$EXTERNALSYM GetPrinterW}
  GetPrivateProfileIntW: TGetPrivateProfileIntW = nil;
  {$EXTERNALSYM GetPrivateProfileIntW}
  GetPrivateProfileSectionNamesW: TGetPrivateProfileSectionNamesW = nil;
  {$EXTERNALSYM GetPrivateProfileSectionNamesW}
  GetPrivateProfileSectionW: TGetPrivateProfileSectionW = nil;
  {$EXTERNALSYM GetPrivateProfileSectionW}
  GetPrivateProfileStringW: TGetPrivateProfileStringW = nil;
  {$EXTERNALSYM GetPrivateProfileStringW}
  GetPrivateProfileStructW: TGetPrivateProfileStructW = nil;
  {$EXTERNALSYM GetPrivateProfileStructW}
  GetProcAddress: TGetProcAddress;
  {$EXTERNALSYM GetProcAddress}
  GetProfileIntW: TGetProfileIntW = nil;
  {$EXTERNALSYM GetProfileIntW}
  GetProfileSectionW: TGetProfileSectionW = nil;
  {$EXTERNALSYM GetProfileSectionW}
  GetProfileStringW: TGetProfileStringW = nil;
  {$EXTERNALSYM GetProfileStringW}
  GetPropA: TGetPropA = nil;
  {$EXTERNALSYM GetPropA}
  GetPropW: TGetPropW = nil;
  {$EXTERNALSYM GetPropW}
  GetRoleTextW: TGetRoleTextW = nil;
  {$EXTERNALSYM GetRoleTextW}
  GetSaveFileNamePreviewW: TGetSaveFileNamePreviewW = nil;
  {$EXTERNALSYM GetSaveFileNamePreviewW}
  GetSaveFileNameW: TGetSaveFileNameW = nil;
  {$EXTERNALSYM GetSaveFileNameW}
  GetShortPathNameW: TGetShortPathNameW = nil;
  {$EXTERNALSYM GetShortPathNameW}
  GetStartupInfoW: TGetStartupInfoW = nil;
  {$EXTERNALSYM GetStartupInfoW}
  GetStateTextW: TGetStateTextW = nil;
  {$EXTERNALSYM GetStateTextW}
  GetStringTypeExW: TGetStringTypeExW = nil;
  {$EXTERNALSYM GetStringTypeExW}
  GetStringTypeW: TGetStringTypeW = nil;
  {$EXTERNALSYM GetStringTypeW}
  GetSystemDirectoryW: TGetSystemDirectoryW = nil;
  {$EXTERNALSYM GetSystemDirectoryW}
  GetSystemWindowsDirectoryW: TGetSystemWindowsDirectoryW = nil;
  {$EXTERNALSYM GetSystemWindowsDirectoryW}
  GetTabbedTextExtentW: TGetTabbedTextExtentW = nil;
  {$EXTERNALSYM GetTabbedTextExtentW}
  GetTempFileNameW: TGetTempFileNameW = nil;
  {$EXTERNALSYM GetTempFileNameW}
  GetTempPathW: TGetTempPathW = nil;
  {$EXTERNALSYM GetTempPathW}
  GetTextExtentExPointW: TGetTextExtentExPointW = nil;
  {$EXTERNALSYM GetTextExtentExPointW}
  GetTextExtentPoint32W: TGetTextExtentPoint32W = nil;
  {$EXTERNALSYM GetTextExtentPoint32W}
  GetTextExtentPointW: TGetTextExtentPointW = nil;
  {$EXTERNALSYM GetTextExtentPointW}
  GetTextFaceW: TGetTextFaceW = nil;
  {$EXTERNALSYM GetTextFaceW}
  GetTextMetricsW: TGetTextMetricsW = nil;
  {$EXTERNALSYM GetTextMetricsW}
  GetTimeFormatW: TGetTimeFormatW = nil;
  {$EXTERNALSYM GetTimeFormatW}
  GetUserNameW: TGetUserNameW = nil;
  {$EXTERNALSYM GetUserNameW}
  GetVersionExW: TGetVersionExW = nil;
  {$EXTERNALSYM GetVersionExW}
  GetVolumeInformationW: TGetVolumeInformationW = nil;
  {$EXTERNALSYM GetVolumeInformationW}
  GetWindowLongA: TGetWindowLongA = nil;
  {$EXTERNALSYM GetWindowLongA}
  GetWindowLongW: TGetWindowLongW = nil;
  {$EXTERNALSYM GetWindowLongW}
  GetWindowModuleFileNameW: TGetWindowModuleFileNameW = nil;
  {$EXTERNALSYM GetWindowModuleFileNameW}
  GetWindowTextLengthW: TGetWindowTextLengthW = nil;
  {$EXTERNALSYM GetWindowTextLengthW}
  GetWindowTextW: TGetWindowTextW = nil;
  {$EXTERNALSYM GetWindowTextW}
  GetWindowsDirectoryW: TGetWindowsDirectoryW = nil;
  {$EXTERNALSYM GetWindowsDirectoryW}
  GlobalAddAtomW: TGlobalAddAtomW = nil;
  {$EXTERNALSYM GlobalAddAtomW}
  GlobalFindAtomW: TGlobalFindAtomW = nil;
  {$EXTERNALSYM GlobalFindAtomW}
  GlobalGetAtomNameW: TGlobalGetAtomNameW = nil;
  {$EXTERNALSYM GlobalGetAtomNameW}
  GrayStringW: TGrayStringW = nil;
  {$EXTERNALSYM GrayStringW}
{$IFDEF UNIJWA}
  InitSecurityInterfaceW: TInitSecurityInterfaceW = nil;
  {$EXTERNALSYM InitSecurityInterfaceW}
  InitializeSecurityContextW: TInitializeSecurityContextW = nil;
  {$EXTERNALSYM InitializeSecurityContextW}
{$ENDIF}
  InsertMenuItemW: TInsertMenuItemW = nil;
  {$EXTERNALSYM InsertMenuItemW}
  InsertMenuW: TInsertMenuW = nil;
  {$EXTERNALSYM InsertMenuW}
  IsBadStringPtrW: TIsBadStringPtrW = nil;
  {$EXTERNALSYM IsBadStringPtrW}
  IsCharAlphaNumericW: TIsCharAlphaNumericW = nil;
  {$EXTERNALSYM IsCharAlphaNumericW}
  IsCharAlphaW: TIsCharAlphaW = nil;
  {$EXTERNALSYM IsCharAlphaW}
  IsCharLowerW: TIsCharLowerW = nil;
  {$EXTERNALSYM IsCharLowerW}
  IsCharUpperW: TIsCharUpperW = nil;
  {$EXTERNALSYM IsCharUpperW}
  IsClipboardFormatAvailable: TIsClipboardFormatAvailable = nil;
  {$EXTERNALSYM IsClipboardFormatAvailable}
  IsDestinationReachableW: TIsDestinationReachableW = nil;
  {$EXTERNALSYM IsDestinationReachableW}
  IsDialogMessageW: TIsDialogMessageW = nil;
  {$EXTERNALSYM IsDialogMessageW}
  IsTextUnicode: TIsTextUnicode = nil;
  {$EXTERNALSYM IsTextUnicode}
  IsValidCodePage: TIsValidCodePage = nil;
  {$EXTERNALSYM IsValidCodePage}
  IsWindowUnicode: TIsWindowUnicode = nil;
  {$EXTERNALSYM IsWindowUnicode}
  LCMapStringW: TLCMapStringW = nil;
  {$EXTERNALSYM LCMapStringW}
  LoadAcceleratorsW: TLoadAcceleratorsW = nil;
  {$EXTERNALSYM LoadAcceleratorsW}
  LoadBitmapW: TLoadBitmapW = nil;
  {$EXTERNALSYM LoadBitmapW}
  LoadCursorFromFileW: TLoadCursorFromFileW = nil;
  {$EXTERNALSYM LoadCursorFromFileW}
  LoadCursorW: TLoadCursorW = nil;
  {$EXTERNALSYM LoadCursorW}
  LoadIconW: TLoadIconW = nil;
  {$EXTERNALSYM LoadIconW}
  LoadImageW: TLoadImageW = nil;
  {$EXTERNALSYM LoadImageW}
  LoadKeyboardLayoutW: TLoadKeyboardLayoutW = nil;
  {$EXTERNALSYM LoadKeyboardLayoutW}
  LoadLibraryExW: TLoadLibraryExW = nil;
  {$EXTERNALSYM LoadLibraryExW}
  LoadLibraryW: TLoadLibraryW = nil;
  {$EXTERNALSYM LoadLibraryW}
  LoadMenuIndirectW: TLoadMenuIndirectW = nil;
  {$EXTERNALSYM LoadMenuIndirectW}
  LoadMenuW: TLoadMenuW = nil;
  {$EXTERNALSYM LoadMenuW}
  LoadStringW: TLoadStringW = nil;
  {$EXTERNALSYM LoadStringW}
  MCIWndCreateW: TMCIWndCreateW = nil;
  {$EXTERNALSYM MCIWndCreateW}
  MapVirtualKeyExW: TMapVirtualKeyExW = nil;
  {$EXTERNALSYM MapVirtualKeyExW}
  MapVirtualKeyW: TMapVirtualKeyW = nil;
  {$EXTERNALSYM MapVirtualKeyW}
  MessageBoxExW: TMessageBoxExW = nil;
  {$EXTERNALSYM MessageBoxExW}
  MessageBoxIndirectW: TMessageBoxIndirectW = nil;
  {$EXTERNALSYM MessageBoxIndirectW}
  MessageBoxW: TMessageBoxW = nil;
  {$EXTERNALSYM MessageBoxW}
  ModifyMenuW: TModifyMenuW = nil;
  {$EXTERNALSYM ModifyMenuW}
  MoveFileW: TMoveFileW = nil;
  {$EXTERNALSYM MoveFileW}
  MultiByteToWideChar: TMultiByteToWideChar = nil;
  {$EXTERNALSYM MultiByteToWideChar}
  MultinetGetConnectionPerformanceW: TMultinetGetConnectionPerformanceW = nil;
  {$EXTERNALSYM MultinetGetConnectionPerformanceW}
  OemToCharBuffW: TOemToCharBuffW = nil;
  {$EXTERNALSYM OemToCharBuffW}
  OemToCharW: TOemToCharW = nil;
  {$EXTERNALSYM OemToCharW}
  OleUIAddVerbMenuW: TOleUIAddVerbMenuW = nil;
  {$EXTERNALSYM OleUIAddVerbMenuW}
  OleUIBusyW: TOleUIBusyWFunc = nil;
  {$EXTERNALSYM OleUIBusyW}
  OleUIChangeIconW: TOleUIChangeIconWFunc = nil;
  {$EXTERNALSYM OleUIChangeIconW}
  OleUIChangeSourceW: TOleUIChangeSourceWFunc = nil;
  {$EXTERNALSYM OleUIChangeSourceW}
  OleUIConvertW: TOleUIConvertWFunc = nil;
  {$EXTERNALSYM OleUIConvertW}
  OleUIEditLinksW: TOleUIEditLinksWFunc = nil;
  {$EXTERNALSYM OleUIEditLinksW}
  OleUIInsertObjectW: TOleUIInsertObjectWFunc = nil;
  {$EXTERNALSYM OleUIInsertObjectW}
  OleUIObjectPropertiesW: TOleUIObjectPropertiesW = nil;
  {$EXTERNALSYM OleUIObjectPropertiesW}
  OleUIPasteSpecialW: TOleUIPasteSpecialWFunc = nil;
  {$EXTERNALSYM OleUIPasteSpecialW}
  OleUIPromptUserW: TOleUIPromptUserW = nil;
  {$EXTERNALSYM OleUIPromptUserW}
  OleUIUpdateLinksW: TOleUIUpdateLinksW = nil;
  {$EXTERNALSYM OleUIUpdateLinksW}
  OpenEventW: TOpenEventW = nil;
  {$EXTERNALSYM OpenEventW}
  OpenFileMappingW: TOpenFileMappingW = nil;
  {$EXTERNALSYM OpenFileMappingW}
  OpenMutexW: TOpenMutexW = nil;
  {$EXTERNALSYM OpenMutexW}
  OpenPrinterW: TOpenPrinterW = nil;
  {$EXTERNALSYM OpenPrinterW}
  OpenSemaphoreW: TOpenSemaphoreW = nil;
  {$EXTERNALSYM OpenSemaphoreW}
  OpenWaitableTimerW: TOpenWaitableTimerW = nil;
  {$EXTERNALSYM OpenWaitableTimerW}
  OutputDebugStringW: TOutputDebugStringW = nil;
  {$EXTERNALSYM OutputDebugStringW}
  PageSetupDlgW: TPageSetupDlgWFunc = nil;
  {$EXTERNALSYM PageSetupDlgW}
  PeekConsoleInputW: TPeekConsoleInputW = nil;
  {$EXTERNALSYM PeekConsoleInputW}
  PeekMessageW: TPeekMessageW = nil;
  {$EXTERNALSYM PeekMessageW}
  PlaySoundW: TPlaySoundW = nil;
  {$EXTERNALSYM PlaySoundW}
  PolyTextOutW: TPolyTextOutW = nil;
  {$EXTERNALSYM PolyTextOutW}
  PostMessageW: TPostMessageW = nil;
  {$EXTERNALSYM PostMessageW}
  PostThreadMessageW: TPostThreadMessageW = nil;
  {$EXTERNALSYM PostThreadMessageW}
  PrintDlgW: TPrintDlgWFunc = nil;
  {$EXTERNALSYM PrintDlgW}
{$IFDEF UNIJWA}
  QueryContextAttributesW: TQueryContextAttributesW = nil;
  {$EXTERNALSYM QueryContextAttributesW}
  QueryCredentialsAttributesW: TQueryCredentialsAttributesW = nil;
  {$EXTERNALSYM QueryCredentialsAttributesW}
{$ENDIF}
  QueryDosDeviceW: TQueryDosDeviceW = nil;
  {$EXTERNALSYM QueryDosDeviceW}
{$IFDEF UNIJWA}
  QuerySecurityPackageInfoW: TQuerySecurityPackageInfoW = nil;
  {$EXTERNALSYM QuerySecurityPackageInfoW}
{$ENDIF}
{$IFDEF UNIJRA}
  RasConnectionNotificationW: TRasConnectionNotificationW = nil;
  {$EXTERNALSYM RasConnectionNotificationW}
  RasCreatePhonebookEntryW: TRasCreatePhonebookEntryW = nil;
  {$EXTERNALSYM RasCreatePhonebookEntryW}
  RasDeleteEntryW: TRasDeleteEntryW = nil;
  {$EXTERNALSYM RasDeleteEntryW}
  RasDeleteSubEntryW: TRasDeleteSubEntryW = nil;
  {$EXTERNALSYM RasDeleteSubEntryW}
  RasDialW: TRasDialW = nil;
  {$EXTERNALSYM RasDialW}
  RasEditPhonebookEntryW: TRasEditPhonebookEntryW = nil;
  {$EXTERNALSYM RasEditPhonebookEntryW}
  RasEnumConnectionsW: TRasEnumConnectionsW = nil;
  {$EXTERNALSYM RasEnumConnectionsW}
  RasEnumDevicesW: TRasEnumDevicesW = nil;
  {$EXTERNALSYM RasEnumDevicesW}
  RasEnumEntriesW: TRasEnumEntriesW = nil;
  {$EXTERNALSYM RasEnumEntriesW}
  RasGetConnectStatusW: TRasGetConnectStatusW = nil;
  {$EXTERNALSYM RasGetConnectStatusW}
  RasGetEntryDialParamsW: TRasGetEntryDialParamsW = nil;
  {$EXTERNALSYM RasGetEntryDialParamsW}
  RasGetEntryPropertiesW: TRasGetEntryPropertiesW = nil;
  {$EXTERNALSYM RasGetEntryPropertiesW}
  RasGetErrorStringW: TRasGetErrorStringW = nil;
  {$EXTERNALSYM RasGetErrorStringW}
  RasGetProjectionInfoW: TRasGetProjectionInfoW; // 3703
  {$EXTERNALSYM RasGetProjectionInfoW}
  RasHangUpW: TRasHangUpW = nil;
  {$EXTERNALSYM RasHangUpW}
  RasRenameEntryW: TRasRenameEntryW = nil;
  {$EXTERNALSYM RasRenameEntryW}
  RasSetEntryDialParamsW: TRasSetEntryDialParamsW = nil;
  {$EXTERNALSYM RasSetEntryDialParamsW}
  RasSetEntryPropertiesW: TRasSetEntryPropertiesW = nil;
  {$EXTERNALSYM RasSetEntryPropertiesW}
  RasSetSubEntryPropertiesW: TRasSetSubEntryPropertiesW = nil;
  {$EXTERNALSYM RasSetSubEntryPropertiesW}
  RasValidateEntryNameW: TRasValidateEntryNameW = nil;
  {$EXTERNALSYM RasValidateEntryNameW}
{$ENDIF}
  ReadConsoleInputW: TReadConsoleInputW = nil;
  {$EXTERNALSYM ReadConsoleInputW}
  ReadConsoleOutputCharacterW: TReadConsoleOutputCharacterW = nil;
  {$EXTERNALSYM ReadConsoleOutputCharacterW}
  ReadConsoleOutputW: TReadConsoleOutputW = nil;
  {$EXTERNALSYM ReadConsoleOutputW}
  ReadConsoleW: TReadConsoleW = nil;
  {$EXTERNALSYM ReadConsoleW}
  RegConnectRegistryW: TRegConnectRegistryW = nil;
  {$EXTERNALSYM RegConnectRegistryW}
  RegCreateKeyExW: TRegCreateKeyExW = nil;
  {$EXTERNALSYM RegCreateKeyExW}
  RegCreateKeyW: TRegCreateKeyW = nil;
  {$EXTERNALSYM RegCreateKeyW}
  RegDeleteKeyW: TRegDeleteKeyW = nil;
  {$EXTERNALSYM RegDeleteKeyW}
  RegDeleteValueW: TRegDeleteValueW = nil;
  {$EXTERNALSYM RegDeleteValueW}
  RegEnumKeyExW: TRegEnumKeyExW = nil;
  {$EXTERNALSYM RegEnumKeyExW}
  RegEnumKeyW: TRegEnumKeyW = nil;
  {$EXTERNALSYM RegEnumKeyW}
  RegEnumValueW: TRegEnumValueW = nil;
  {$EXTERNALSYM RegEnumValueW}
  RegLoadKeyW: TRegLoadKeyW = nil;
  {$EXTERNALSYM RegLoadKeyW}
  RegOpenKeyExW: TRegOpenKeyExW = nil;
  {$EXTERNALSYM RegOpenKeyExW}
  RegOpenKeyW: TRegOpenKeyW = nil;
  {$EXTERNALSYM RegOpenKeyW}
  RegQueryInfoKeyW: TRegQueryInfoKeyW = nil;
  {$EXTERNALSYM RegQueryInfoKeyW}
  RegQueryMultipleValuesW: TRegQueryMultipleValuesW = nil;
  {$EXTERNALSYM RegQueryMultipleValuesW}
  RegQueryValueExW: TRegQueryValueExW = nil;
  {$EXTERNALSYM RegQueryValueExW}
  RegQueryValueW: TRegQueryValueW = nil;
  {$EXTERNALSYM RegQueryValueW}
  RegReplaceKeyW: TRegReplaceKeyW = nil;
  {$EXTERNALSYM RegReplaceKeyW}
  RegSaveKeyW: TRegSaveKeyW = nil;
  {$EXTERNALSYM RegSaveKeyW}
  RegSetValueExW: TRegSetValueExW = nil;
  {$EXTERNALSYM RegSetValueExW}
  RegSetValueW: TRegSetValueW = nil;
  {$EXTERNALSYM RegSetValueW}
  RegUnLoadKeyW: TRegUnLoadKeyW = nil;
  {$EXTERNALSYM RegUnLoadKeyW}
  RegisterClassExW: TRegisterClassExW = nil;
  {$EXTERNALSYM RegisterClassExW}
  RegisterClassW: TRegisterClassW = nil;
  {$EXTERNALSYM RegisterClassW}
  RegisterClipboardFormatW: TRegisterClipboardFormatW = nil;
  {$EXTERNALSYM RegisterClipboardFormatW}
  RegisterDeviceNotificationW: TRegisterDeviceNotificationW = nil;
  {$EXTERNALSYM RegisterDeviceNotificationW}
  RegisterWindowMessageW: TRegisterWindowMessageW = nil;
  {$EXTERNALSYM RegisterWindowMessageW}
  RemoveDirectoryW: TRemoveDirectoryW = nil;
  {$EXTERNALSYM RemoveDirectoryW}
  RemoveFontResourceW: TRemoveFontResourceW = nil;
  {$EXTERNALSYM RemoveFontResourceW}
  RemovePropA: TRemovePropA = nil;
  {$EXTERNALSYM RemovePropA}
  RemovePropW: TRemovePropW = nil;
  {$EXTERNALSYM RemovePropW}
  ReplaceTextW: TReplaceTextW = nil;
  {$EXTERNALSYM ReplaceTextW}
  ResetDCW: TResetDCW = nil;
  {$EXTERNALSYM ResetDCW}
  ResetPrinterW: TResetPrinterW = nil;
  {$EXTERNALSYM ResetPrinterW}
  SHBrowseForFolderW: TSHBrowseForFolderW = nil;
  {$EXTERNALSYM SHBrowseForFolderW}
  SHChangeNotify: TSHChangeNotify = nil;
  {$EXTERNALSYM SHChangeNotify}
  SHFileOperationW: TSHFileOperationW = nil;
  {$EXTERNALSYM SHFileOperationW}
  SHGetFileInfoW: TSHGetFileInfoW = nil;
  {$EXTERNALSYM SHGetFileInfoW}
  SHGetNewLinkInfoW: TSHGetNewLinkInfoW = nil;
  {$EXTERNALSYM SHGetNewLinkInfoW}
  SHGetPathFromIDListW: TSHGetPathFromIDListW = nil;
  {$EXTERNALSYM SHGetPathFromIDListW}
  ScrollConsoleScreenBufferW: TScrollConsoleScreenBufferW = nil;
  {$EXTERNALSYM ScrollConsoleScreenBufferW}
  SearchPathW: TSearchPathW = nil;
  {$EXTERNALSYM SearchPathW}
  SendDlgItemMessageW: TSendDlgItemMessageW = nil;
  {$EXTERNALSYM SendDlgItemMessageW}
  SendMessageCallbackW: TSendMessageCallbackW = nil;
  {$EXTERNALSYM SendMessageCallbackW}
  SendMessageTimeoutW: TSendMessageTimeoutW = nil;
  {$EXTERNALSYM SendMessageTimeoutW}
  SendMessageW: TSendMessageW = nil;
  {$EXTERNALSYM SendMessageW}
  SendNotifyMessageW: TSendNotifyMessageW = nil;
  {$EXTERNALSYM SendNotifyMessageW}
  SetCalendarInfoW: TSetCalendarInfoW = nil;
  {$EXTERNALSYM SetCalendarInfoW}
  SetClassLongW: TSetClassLongW = nil;
  {$EXTERNALSYM SetClassLongW}
  SetComputerNameW: TSetComputerNameW = nil;
  {$EXTERNALSYM SetComputerNameW}
  SetConsoleTitleW: TSetConsoleTitleW = nil;
  {$EXTERNALSYM SetConsoleTitleW}
  SetCurrentDirectoryW: TSetCurrentDirectoryW = nil;
  {$EXTERNALSYM SetCurrentDirectoryW}
  SetDefaultCommConfigW: TSetDefaultCommConfigW = nil;
  {$EXTERNALSYM SetDefaultCommConfigW}
  SetDlgItemTextW: TSetDlgItemTextW = nil;
  {$EXTERNALSYM SetDlgItemTextW}
  SetEnvironmentVariableW: TSetEnvironmentVariableW = nil;
  {$EXTERNALSYM SetEnvironmentVariableW}
  SetFileAttributesW: TSetFileAttributesW = nil;
  {$EXTERNALSYM SetFileAttributesW}
  SetICMProfileW: TSetICMProfileW = nil;
  {$EXTERNALSYM SetICMProfileW}
  SetJobW: TSetJobW = nil;
  {$EXTERNALSYM SetJobW}
  SetLocaleInfoW: TSetLocaleInfoW = nil;
  {$EXTERNALSYM SetLocaleInfoW}
  SetMenuItemInfoW: TSetMenuItemInfoW = nil;
  {$EXTERNALSYM SetMenuItemInfoW}
  SetPrinterDataW: TSetPrinterDataW = nil;
  {$EXTERNALSYM SetPrinterDataW}
  SetPrinterW: TSetPrinterW = nil;
  {$EXTERNALSYM SetPrinterW}
  SetPropA: TSetPropA = nil;
  {$EXTERNALSYM SetPropA}
  SetPropW: TSetPropW = nil;
  {$EXTERNALSYM SetPropW}
  SetVolumeLabelW: TSetVolumeLabelW = nil;
  {$EXTERNALSYM SetVolumeLabelW}
  SetWindowLongA: TSetWindowLongA = nil;
  {$EXTERNALSYM SetWindowLongA}
  SetWindowLongW: TSetWindowLongW = nil;
  {$EXTERNALSYM SetWindowLongW}
  SetWindowTextW: TSetWindowTextW = nil;
  {$EXTERNALSYM SetWindowTextW}
  SetWindowsHookExW: TSetWindowsHookExW = nil;
  {$EXTERNALSYM SetWindowsHookExW}
  SetWindowsHookW: TSetWindowsHookW = nil;
  {$EXTERNALSYM SetWindowsHookW}
  ShellAboutW: TShellAboutW = nil;
  {$EXTERNALSYM ShellAboutW}
  ShellExecuteExW: TShellExecuteExW = nil;
  {$EXTERNALSYM ShellExecuteExW}
  ShellExecuteW: TShellExecuteW = nil;
  {$EXTERNALSYM ShellExecuteW}
  Shell_NotifyIconW: TShell_NotifyIconW = nil;
  {$EXTERNALSYM Shell_NotifyIconW}
  StartDocPrinterW: TStartDocPrinterW = nil;
  {$EXTERNALSYM StartDocPrinterW}
  StartDocW: TStartDocW = nil;
  {$EXTERNALSYM StartDocW}
  SystemParametersInfoW: TSystemParametersInfoW = nil;
  {$EXTERNALSYM SystemParametersInfoW}
  TabbedTextOutW: TTabbedTextOutW = nil;
  {$EXTERNALSYM TabbedTextOutW}
  TextOutW: TTextOutW = nil;
  {$EXTERNALSYM TextOutW}
  TranslateAcceleratorW: TTranslateAcceleratorW = nil;
  {$EXTERNALSYM TranslateAcceleratorW}
  UnregisterClassW: TUnregisterClassW = nil;
  {$EXTERNALSYM UnregisterClassW}
  UpdateICMRegKeyW: TUpdateICMRegKeyW = nil;
  {$EXTERNALSYM UpdateICMRegKeyW}
  UpdateResourceA: TUpdateResourceA = nil; // 3703
  {$EXTERNALSYM UpdateResourceA}
  UpdateResourceW: TUpdateResourceW = nil;
  {$EXTERNALSYM UpdateResourceW}
  VerFindFileW: TVerFindFileW = nil;
  {$EXTERNALSYM VerFindFileW}
  VerInstallFileW: TVerInstallFileW = nil;
  {$EXTERNALSYM VerInstallFileW}
  VerLanguageNameW: TVerLanguageNameW = nil;
  {$EXTERNALSYM VerLanguageNameW}
  VerQueryValueW: TVerQueryValueW = nil;
  {$EXTERNALSYM VerQueryValueW}
  VkKeyScanExW: TVkKeyScanExW = nil;
  {$EXTERNALSYM VkKeyScanExW}
  VkKeyScanW: TVkKeyScanW = nil;
  {$EXTERNALSYM VkKeyScanW}
  WNetAddConnection2W: TWNetAddConnection2W = nil;
  {$EXTERNALSYM WNetAddConnection2W}
  WNetAddConnection3W: TWNetAddConnection3W = nil;
  {$EXTERNALSYM WNetAddConnection3W}
  WNetAddConnectionW: TWNetAddConnectionW = nil;
  {$EXTERNALSYM WNetAddConnectionW}
  WNetCancelConnection2W: TWNetCancelConnection2W = nil;
  {$EXTERNALSYM WNetCancelConnection2W}
  WNetCancelConnectionW: TWNetCancelConnectionW = nil;
  {$EXTERNALSYM WNetCancelConnectionW}
  WNetConnectionDialog1W: TWNetConnectionDialog1W = nil;
  {$EXTERNALSYM WNetConnectionDialog1W}
  WNetDisconnectDialog1W: TWNetDisconnectDialog1W = nil;
  {$EXTERNALSYM WNetDisconnectDialog1W}
  WNetEnumResourceW: TWNetEnumResourceW = nil;
  {$EXTERNALSYM WNetEnumResourceW}
  WNetGetConnectionW: TWNetGetConnectionW = nil;
  {$EXTERNALSYM WNetGetConnectionW}
  WNetGetLastErrorW: TWNetGetLastErrorW = nil;
  {$EXTERNALSYM WNetGetLastErrorW}
  WNetGetNetworkInformationW: TWNetGetNetworkInformationW = nil;
  {$EXTERNALSYM WNetGetNetworkInformationW}
  WNetGetProviderNameW: TWNetGetProviderNameW = nil;
  {$EXTERNALSYM WNetGetProviderNameW}
  WNetGetResourceInformationW: TWNetGetResourceInformationW = nil;
  {$EXTERNALSYM WNetGetResourceInformationW}
  WNetGetResourceParentW: TWNetGetResourceParentW = nil;
  {$EXTERNALSYM WNetGetResourceParentW}
  WNetGetUniversalNameW: TWNetGetUniversalNameW = nil;
  {$EXTERNALSYM WNetGetUniversalNameW}
  WNetGetUserW: TWNetGetUserW = nil;
  {$EXTERNALSYM WNetGetUserW}
  WNetOpenEnumW: TWNetOpenEnumW = nil;
  {$EXTERNALSYM WNetOpenEnumW}
  WNetUseConnectionW: TWNetUseConnectionW = nil;
  {$EXTERNALSYM WNetUseConnectionW}
  WaitNamedPipeW: TWaitNamedPipeW = nil;
  {$EXTERNALSYM WaitNamedPipeW}
  WideCharToMultiByte: TWideCharToMultiByte = nil;
  {$EXTERNALSYM WideCharToMultiByte}
  WinHelpW: TWinHelpW = nil;
  {$EXTERNALSYM WinHelpW}
  WriteConsoleInputW: TWriteConsoleInputW = nil;
  {$EXTERNALSYM WriteConsoleInputW}
  WriteConsoleOutputCharacterW: TWriteConsoleOutputCharacterW = nil;
  {$EXTERNALSYM WriteConsoleOutputCharacterW}
  WriteConsoleOutputW: TWriteConsoleOutputW = nil;
  {$EXTERNALSYM WriteConsoleOutputW}
  WriteConsoleW: TWriteConsoleW = nil;
  {$EXTERNALSYM WriteConsoleW}
  WritePrivateProfileSectionW: TWritePrivateProfileSectionW = nil;
  {$EXTERNALSYM WritePrivateProfileSectionW}
  WritePrivateProfileStringW: TWritePrivateProfileStringW = nil;
  {$EXTERNALSYM WritePrivateProfileStringW}
  WritePrivateProfileStructW: TWritePrivateProfileStructW = nil;
  {$EXTERNALSYM WritePrivateProfileStructW}
  WriteProfileSectionW: TWriteProfileSectionW = nil;
  {$EXTERNALSYM WriteProfileSectionW}
  WriteProfileStringW: TWriteProfileStringW = nil;
  {$EXTERNALSYM WriteProfileStringW}
  auxGetDevCapsW: TauxGetDevCapsW = nil;
  {$EXTERNALSYM auxGetDevCapsW}
  capCreateCaptureWindowW: TcapCreateCaptureWindowW = nil;
  {$EXTERNALSYM capCreateCaptureWindowW}
  capGetDriverDescriptionW: TcapGetDriverDescriptionW = nil;
  {$EXTERNALSYM capGetDriverDescriptionW}
  joyGetDevCapsW: TjoyGetDevCapsW = nil;
  {$EXTERNALSYM joyGetDevCapsW}
  lstrcatW: TlstrcatW = nil;
  {$EXTERNALSYM lstrcatW}
  lstrcmpW: TlstrcmpW = nil;
  {$EXTERNALSYM lstrcmpW}
  lstrcmpiW: TlstrcmpiW = nil;
  {$EXTERNALSYM lstrcmpiW}
  lstrcpyW: TlstrcpyW = nil;
  {$EXTERNALSYM lstrcpyW}
  lstrcpynW: TlstrcpynW = nil;
  {$EXTERNALSYM lstrcpynW}
  lstrlenW: TlstrlenW = nil;
  {$EXTERNALSYM lstrlenW}
  mciGetDeviceIDW: TmciGetDeviceIDW = nil;
  {$EXTERNALSYM mciGetDeviceIDW}
  mciGetErrorStringW: TmciGetErrorStringW = nil;
  {$EXTERNALSYM mciGetErrorStringW}
  mciSendCommandW: TmciSendCommandW = nil;
  {$EXTERNALSYM mciSendCommandW}
  mciSendStringW: TmciSendStringW = nil;
  {$EXTERNALSYM mciSendStringW}
  midiInGetDevCapsW: TmidiInGetDevCapsW = nil;
  {$EXTERNALSYM midiInGetDevCapsW}
  midiInGetErrorTextW: TmidiInGetErrorTextW = nil;
  {$EXTERNALSYM midiInGetErrorTextW}
  midiOutGetDevCapsW: TmidiOutGetDevCapsW = nil;
  {$EXTERNALSYM midiOutGetDevCapsW}
  midiOutGetErrorTextW: TmidiOutGetErrorTextW = nil;
  {$EXTERNALSYM midiOutGetErrorTextW}
  mixerGetControlDetailsW: TmixerGetControlDetailsW = nil;
  {$EXTERNALSYM mixerGetControlDetailsW}
  mixerGetDevCapsW: TmixerGetDevCapsW = nil;
  {$EXTERNALSYM mixerGetDevCapsW}
  mixerGetLineControlsW: TmixerGetLineControlsW = nil;
  {$EXTERNALSYM mixerGetLineControlsW}
  mixerGetLineInfoW: TmixerGetLineInfoW = nil;
  {$EXTERNALSYM mixerGetLineInfoW}
  mmioInstallIOProcW: TmmioInstallIOProcW = nil;
  {$EXTERNALSYM mmioInstallIOProcW}
  mmioOpenW: TmmioOpenW = nil;
  {$EXTERNALSYM mmioOpenW}
  mmioRenameW: TmmioRenameW = nil;
  {$EXTERNALSYM mmioRenameW}
  mmioStringToFOURCCW: TmmioStringToFOURCCW = nil;
  {$EXTERNALSYM mmioStringToFOURCCW}
  sndPlaySoundW: TsndPlaySoundW = nil;
  {$EXTERNALSYM sndPlaySoundW}
  waveInGetDevCapsW: TwaveInGetDevCapsW = nil;
  {$EXTERNALSYM waveInGetDevCapsW}
  waveInGetErrorTextW: TwaveInGetErrorTextW = nil;
  {$EXTERNALSYM waveInGetErrorTextW}
  waveOutGetDevCapsW: TwaveOutGetDevCapsW = nil;
  {$EXTERNALSYM waveOutGetDevCapsW}
  waveOutGetErrorTextW: TwaveOutGetErrorTextW = nil;
  {$EXTERNALSYM waveOutGetErrorTextW}
  wsprintfW: TwsprintfW = nil;
  {$EXTERNALSYM wsprintfW}
  wvsprintfW: TwvsprintfW = nil;
  {$EXTERNALSYM wvsprintfW}

function IsMSLULoaded: Boolean;
{$EXTERNALSYM IsMSLULoaded}
procedure LoadMSLU;
{$EXTERNALSYM LoadMSLU}

implementation

uses SysUtils;

var
  hUnicoWS: HMODULE = 0;

function IsMSLULoaded: Boolean;
begin
  Result := (hUnicoWS <> 0);
end;

procedure LoadMSLU;
var
  FN: string;
begin
  if hUnicoWS = 0 then
  begin
    SetLength(FN,MAX_PATH);
    if GetModuleFileName(SysInit.HInstance,PChar(FN),MAX_PATH) > 0
    then
{$IFDEF VER130}
      hUnicoWS := LoadLibraryA(PChar(IncludeTrailingBackslash(ExtractFilePath(
        string(PChar(FN)))) + 'unicows.dll'));
{$ELSE}
      hUnicoWS := LoadLibraryA(PChar(IncludeTrailingPathDelimiter(ExtractFilePath(
        string(PChar(FN)))) + 'unicows.dll'));
{$ENDIF}
    if hUnicoWS = 0 then hUnicoWS := LoadLibraryA('unicows.dll');
    if hUnicoWS <> 0 then
    begin
      GetProcAddress := Windows.GetProcAddress(hUnicoWS,'GetProcAddress');
{$IFDEF UNIJWA}
      AcquireCredentialsHandleW := GetProcAddress(hUnicoWS,'AcquireCredentialsHandleW');
{$ENDIF}
      AddAtomW := GetProcAddress(hUnicoWS,'AddAtomW');
      AddFontResourceW := GetProcAddress(hUnicoWS,'AddFontResourceW');
      AddJobW := GetProcAddress(hUnicoWS,'AddJobW');
      AddMonitorW := GetProcAddress(hUnicoWS,'AddMonitorW');
      AddPortW := GetProcAddress(hUnicoWS,'AddPortW');
      AddPrintProcessorW := GetProcAddress(hUnicoWS,'AddPrintProcessorW');
      AddPrintProvidorW := GetProcAddress(hUnicoWS,'AddPrintProvidorW');
      AddPrinterDriverW := GetProcAddress(hUnicoWS,'AddPrinterDriverW');
      AddPrinterW := GetProcAddress(hUnicoWS,'AddPrinterW');
      AdvancedDocumentPropertiesW := GetProcAddress(hUnicoWS,'AdvancedDocumentPropertiesW');
      AppendMenuW := GetProcAddress(hUnicoWS,'AppendMenuW');
      BeginUpdateResourceA := GetProcAddress(hUnicoWS,'BeginUpdateResourceA'); // 3703
      BeginUpdateResourceW := GetProcAddress(hUnicoWS,'BeginUpdateResourceW');
      BroadcastSystemMessageW := GetProcAddress(hUnicoWS,'BroadcastSystemMessageW');
      BuildCommDCBAndTimeoutsW := GetProcAddress(hUnicoWS,'BuildCommDCBAndTimeoutsW');
      BuildCommDCBW := GetProcAddress(hUnicoWS,'BuildCommDCBW');
      CallMsgFilterW := GetProcAddress(hUnicoWS,'CallMsgFilterW');
      CallNamedPipeW := GetProcAddress(hUnicoWS,'CallNamedPipeW');
      CallWindowProcA := GetProcAddress(hUnicoWS,'CallWindowProcA');
      CallWindowProcW := GetProcAddress(hUnicoWS,'CallWindowProcW');
      ChangeDisplaySettingsExW := GetProcAddress(hUnicoWS,'ChangeDisplaySettingsExW');
      ChangeDisplaySettingsW := GetProcAddress(hUnicoWS,'ChangeDisplaySettingsW');
      ChangeMenuW := GetProcAddress(hUnicoWS,'ChangeMenuW');
      CharLowerBuffW := GetProcAddress(hUnicoWS,'CharLowerBuffW');
      CharLowerW := GetProcAddress(hUnicoWS,'CharLowerW');
      CharNextW := GetProcAddress(hUnicoWS,'CharNextW');
      CharPrevW := GetProcAddress(hUnicoWS,'CharPrevW');
      CharToOemBuffW := GetProcAddress(hUnicoWS,'CharToOemBuffW');
      CharToOemW := GetProcAddress(hUnicoWS,'CharToOemW');
      CharUpperBuffW := GetProcAddress(hUnicoWS,'CharUpperBuffW');
      CharUpperW := GetProcAddress(hUnicoWS,'CharUpperW');
      ChooseColorW := GetProcAddress(hUnicoWS,'ChooseColorW');
      ChooseFontW := GetProcAddress(hUnicoWS,'ChooseFontW');
      CommConfigDialogW := GetProcAddress(hUnicoWS,'CommConfigDialogW');
      CompareStringW := GetProcAddress(hUnicoWS,'CompareStringW');
      ConfigurePortW := GetProcAddress(hUnicoWS,'ConfigurePortW');
      CopyAcceleratorTableW := GetProcAddress(hUnicoWS,'CopyAcceleratorTableW');
      CopyEnhMetaFileW := GetProcAddress(hUnicoWS,'CopyEnhMetaFileW');
      CopyFileExW := GetProcAddress(hUnicoWS,'CopyFileExW');
      CopyFileW := GetProcAddress(hUnicoWS,'CopyFileW');
      CopyMetaFileW := GetProcAddress(hUnicoWS,'CopyMetaFileW');
      CreateAcceleratorTableW := GetProcAddress(hUnicoWS,'CreateAcceleratorTableW');
      CreateColorSpaceW := GetProcAddress(hUnicoWS,'CreateColorSpaceW');
      CreateDCW := GetProcAddress(hUnicoWS,'CreateDCW');
      CreateDialogIndirectParamW := GetProcAddress(hUnicoWS,'CreateDialogIndirectParamW');
      CreateDialogParamW := GetProcAddress(hUnicoWS,'CreateDialogParamW');
      CreateDirectoryExW := GetProcAddress(hUnicoWS,'CreateDirectoryExW');
      CreateDirectoryW := GetProcAddress(hUnicoWS,'CreateDirectoryW');
      CreateEnhMetaFileW := GetProcAddress(hUnicoWS,'CreateEnhMetaFileW');
      CreateEventW := GetProcAddress(hUnicoWS,'CreateEventW');
      CreateFileMappingW := GetProcAddress(hUnicoWS,'CreateFileMappingW');
      CreateFileW := GetProcAddress(hUnicoWS,'CreateFileW');
      CreateFontIndirectW := GetProcAddress(hUnicoWS,'CreateFontIndirectW');
      CreateFontW := GetProcAddress(hUnicoWS,'CreateFontW');
      CreateICW := GetProcAddress(hUnicoWS,'CreateICW');
      CreateMDIWindowW := GetProcAddress(hUnicoWS,'CreateMDIWindowW');
      CreateMailslotW := GetProcAddress(hUnicoWS,'CreateMailslotW');
      CreateMetaFileW := GetProcAddress(hUnicoWS,'CreateMetaFileW');
      CreateMutexW := GetProcAddress(hUnicoWS,'CreateMutexW');
      CreateNamedPipeW := GetProcAddress(hUnicoWS,'CreateNamedPipeW');
      CreateProcessW := GetProcAddress(hUnicoWS,'CreateProcessW');
      CreateScalableFontResourceW := GetProcAddress(hUnicoWS,'CreateScalableFontResourceW');
      CreateSemaphoreW := GetProcAddress(hUnicoWS,'CreateSemaphoreW');
      CreateStdAccessibleProxyW := GetProcAddress(hUnicoWS,'CreateStdAccessibleProxyW');
      CreateWaitableTimerW := GetProcAddress(hUnicoWS,'CreateWaitableTimerW');
      CreateWindowExW := GetProcAddress(hUnicoWS,'CreateWindowExW');
{$IFDEF UNIJWA}
      CryptAcquireContextW := GetProcAddress(hUnicoWS,'CryptAcquireContextW');
      CryptEnumProviderTypesW := GetProcAddress(hUnicoWS,'CryptEnumProviderTypesW');
      CryptEnumProvidersW := GetProcAddress(hUnicoWS,'CryptEnumProvidersW');
      CryptGetDefaultProviderW := GetProcAddress(hUnicoWS,'CryptGetDefaultProviderW');
      CryptSetProviderExW := GetProcAddress(hUnicoWS,'CryptSetProviderExW');
      CryptSetProviderW := GetProcAddress(hUnicoWS,'CryptSetProviderW');
      CryptSignHashW := GetProcAddress(hUnicoWS,'CryptSignHashW');
      CryptVerifySignatureW := GetProcAddress(hUnicoWS,'CryptVerifySignatureW');
{$ENDIF}
      DdeConnect := GetProcAddress(hUnicoWS,'DdeConnect');
      DdeConnectList := GetProcAddress(hUnicoWS,'DdeConnectList');
      DdeCreateStringHandleW := GetProcAddress(hUnicoWS,'DdeCreateStringHandleW');
      DdeInitializeW := GetProcAddress(hUnicoWS,'DdeInitializeW');
      DdeQueryConvInfo := GetProcAddress(hUnicoWS,'DdeQueryConvInfo');
      DdeQueryStringW := GetProcAddress(hUnicoWS,'DdeQueryStringW');
      DefDlgProcW := GetProcAddress(hUnicoWS,'DefDlgProcW');
      DefFrameProcW := GetProcAddress(hUnicoWS,'DefFrameProcW');
      DefMDIChildProcW := GetProcAddress(hUnicoWS,'DefMDIChildProcW');
      DefWindowProcW := GetProcAddress(hUnicoWS,'DefWindowProcW');
      DeleteFileW := GetProcAddress(hUnicoWS,'DeleteFileW');
      DeleteMonitorW := GetProcAddress(hUnicoWS,'DeleteMonitorW');
      DeletePortW := GetProcAddress(hUnicoWS,'DeletePortW');
      DeletePrintProcessorW := GetProcAddress(hUnicoWS,'DeletePrintProcessorW');
      DeletePrintProvidorW := GetProcAddress(hUnicoWS,'DeletePrintProvidorW');
      DeletePrinterDriverW := GetProcAddress(hUnicoWS,'DeletePrinterDriverW');
      DeviceCapabilitiesW := GetProcAddress(hUnicoWS,'DeviceCapabilitiesW');
      DialogBoxIndirectParamW := GetProcAddress(hUnicoWS,'DialogBoxIndirectParamW');
      DialogBoxParamW := GetProcAddress(hUnicoWS,'DialogBoxParamW');
      DispatchMessageW := GetProcAddress(hUnicoWS,'DispatchMessageW');
      DlgDirListComboBoxW := GetProcAddress(hUnicoWS,'DlgDirListComboBoxW');
      DlgDirListW := GetProcAddress(hUnicoWS,'DlgDirListW');
      DlgDirSelectComboBoxExW := GetProcAddress(hUnicoWS,'DlgDirSelectComboBoxExW');
      DlgDirSelectExW := GetProcAddress(hUnicoWS,'DlgDirSelectExW');
      DocumentPropertiesW := GetProcAddress(hUnicoWS,'DocumentPropertiesW');
      DragQueryFileW := GetProcAddress(hUnicoWS,'DragQueryFileW');
      DrawStateW := GetProcAddress(hUnicoWS,'DrawStateW');
      DrawTextExW := GetProcAddress(hUnicoWS,'DrawTextExW');
      DrawTextW := GetProcAddress(hUnicoWS,'DrawTextW');
      EnableWindow := GetProcAddress(hUnicoWS,'EnableWindow');
      EndUpdateResourceA := GetProcAddress(hUnicoWS,'EndUpdateResourceA'); // 3703
      EndUpdateResourceW := GetProcAddress(hUnicoWS,'EndUpdateResourceW');
{$IFDEF UNIJWA}
      EnumCalendarInfoExW := GetProcAddress(hUnicoWS,'EnumCalendarInfoExW');
{$ENDIF}
      EnumCalendarInfoW := GetProcAddress(hUnicoWS,'EnumCalendarInfoW');
      EnumClipboardFormats := GetProcAddress(hUnicoWS,'EnumClipboardFormats');
{$IFDEF UNIJWA}
      EnumDateFormatsExW := GetProcAddress(hUnicoWS,'EnumDateFormatsExW');
{$ENDIF}
      EnumDateFormatsW := GetProcAddress(hUnicoWS,'EnumDateFormatsW');
      EnumDisplayDevicesW := GetProcAddress(hUnicoWS,'EnumDisplayDevicesW');
      EnumDisplaySettingsExW := GetProcAddress(hUnicoWS,'EnumDisplaySettingsExW');
      EnumDisplaySettingsW := GetProcAddress(hUnicoWS,'EnumDisplaySettingsW');
      EnumFontFamiliesExW := GetProcAddress(hUnicoWS,'EnumFontFamiliesExW');
      EnumFontFamiliesW := GetProcAddress(hUnicoWS,'EnumFontFamiliesW');
      EnumFontsW := GetProcAddress(hUnicoWS,'EnumFontsW');
      EnumICMProfilesW := GetProcAddress(hUnicoWS,'EnumICMProfilesW');
      EnumMonitorsW := GetProcAddress(hUnicoWS,'EnumMonitorsW');
      EnumPortsW := GetProcAddress(hUnicoWS,'EnumPortsW');
      EnumPrintProcessorDatatypesW := GetProcAddress(hUnicoWS,'EnumPrintProcessorDatatypesW');
      EnumPrintProcessorsW := GetProcAddress(hUnicoWS,'EnumPrintProcessorsW');
      EnumPrinterDriversW := GetProcAddress(hUnicoWS,'EnumPrinterDriversW');
      EnumPrintersW := GetProcAddress(hUnicoWS,'EnumPrintersW');
      EnumPropsA := GetProcAddress(hUnicoWS,'EnumPropsA');
      EnumPropsExA := GetProcAddress(hUnicoWS,'EnumPropsExA');
      EnumPropsExW := GetProcAddress(hUnicoWS,'EnumPropsExW');
      EnumPropsW := GetProcAddress(hUnicoWS,'EnumPropsW');
      EnumSystemCodePagesW := GetProcAddress(hUnicoWS,'EnumSystemCodePagesW');
      EnumSystemLocalesW := GetProcAddress(hUnicoWS,'EnumSystemLocalesW');
      EnumTimeFormatsW := GetProcAddress(hUnicoWS,'EnumTimeFormatsW');
{$IFDEF UNIJWA}
      EnumerateSecurityPackagesW := GetProcAddress(hUnicoWS,'EnumerateSecurityPackagesW');
{$ENDIF}
      ExpandEnvironmentStringsW := GetProcAddress(hUnicoWS,'ExpandEnvironmentStringsW');
      ExtTextOutW := GetProcAddress(hUnicoWS,'ExtTextOutW');
      ExtractIconExW := GetProcAddress(hUnicoWS,'ExtractIconExW');
      ExtractIconW := GetProcAddress(hUnicoWS,'ExtractIconW');
      FatalAppExitW := GetProcAddress(hUnicoWS,'FatalAppExitW');
      FillConsoleOutputCharacterW := GetProcAddress(hUnicoWS,'FillConsoleOutputCharacterW');
      FindAtomW := GetProcAddress(hUnicoWS,'FindAtomW');
      FindExecutableW := GetProcAddress(hUnicoWS,'FindExecutableW');
      FindFirstChangeNotificationW := GetProcAddress(hUnicoWS,'FindFirstChangeNotificationW');
      FindFirstFileW := GetProcAddress(hUnicoWS,'FindFirstFileW');
      FindNextFileW := GetProcAddress(hUnicoWS,'FindNextFileW');
      FindResourceExW := GetProcAddress(hUnicoWS,'FindResourceExW');
      FindResourceW := GetProcAddress(hUnicoWS,'FindResourceW');
      FindTextW := GetProcAddress(hUnicoWS,'FindTextW');
      FindWindowExW := GetProcAddress(hUnicoWS,'FindWindowExW');
      FindWindowW := GetProcAddress(hUnicoWS,'FindWindowW');
      FormatMessageW := GetProcAddress(hUnicoWS,'FormatMessageW');
{$IFDEF UNIJWA}
      FreeContextBuffer := GetProcAddress(hUnicoWS,'FreeContextBuffer');
{$ENDIF}
      FreeEnvironmentStringsW := GetProcAddress(hUnicoWS,'FreeEnvironmentStringsW');
      GetAltTabInfoW := GetProcAddress(hUnicoWS,'GetAltTabInfoW');
      GetAtomNameW := GetProcAddress(hUnicoWS,'GetAtomNameW');
      GetCPInfo := GetProcAddress(hUnicoWS,'GetCPInfo');
{$IFDEF UNIJWA}
      GetCPInfoExW := GetProcAddress(hUnicoWS,'GetCPInfoExW');
{$ENDIF}
      GetCalendarInfoW := GetProcAddress(hUnicoWS,'GetCalendarInfoW');
      GetCharABCWidthsFloatW := GetProcAddress(hUnicoWS,'GetCharABCWidthsFloatW');
      GetCharABCWidthsW := GetProcAddress(hUnicoWS,'GetCharABCWidthsW');
      GetCharWidth32W := GetProcAddress(hUnicoWS,'GetCharWidth32W');
      GetCharWidthFloatW := GetProcAddress(hUnicoWS,'GetCharWidthFloatW');
      GetCharWidthW := GetProcAddress(hUnicoWS,'GetCharWidthW');
      GetCharacterPlacementW := GetProcAddress(hUnicoWS,'GetCharacterPlacementW');
      GetClassInfoExW := GetProcAddress(hUnicoWS,'GetClassInfoExW');
      GetClassInfoW := GetProcAddress(hUnicoWS,'GetClassInfoW');
      GetClassLongW := GetProcAddress(hUnicoWS,'GetClassLongW');
      GetClassNameW := GetProcAddress(hUnicoWS,'GetClassNameW');
      GetClipboardData := GetProcAddress(hUnicoWS,'GetClipboardData');
      GetClipboardFormatNameW := GetProcAddress(hUnicoWS,'GetClipboardFormatNameW');
      GetComputerNameW := GetProcAddress(hUnicoWS,'GetComputerNameW');
      GetConsoleTitleW := GetProcAddress(hUnicoWS,'GetConsoleTitleW');
      GetCurrencyFormatW := GetProcAddress(hUnicoWS,'GetCurrencyFormatW');
      GetCurrentDirectoryW := GetProcAddress(hUnicoWS,'GetCurrentDirectoryW');
      GetCurrentHwProfileW := GetProcAddress(hUnicoWS,'GetCurrentHwProfileW');
      GetDateFormatW := GetProcAddress(hUnicoWS,'GetDateFormatW');
      GetDefaultCommConfigW := GetProcAddress(hUnicoWS,'GetDefaultCommConfigW');
      GetDiskFreeSpaceExW := GetProcAddress(hUnicoWS,'GetDiskFreeSpaceExW');
      GetDiskFreeSpaceW := GetProcAddress(hUnicoWS,'GetDiskFreeSpaceW');
      GetDlgItemTextW := GetProcAddress(hUnicoWS,'GetDlgItemTextW');
      GetDriveTypeW := GetProcAddress(hUnicoWS,'GetDriveTypeW');
      GetEnhMetaFileDescriptionW := GetProcAddress(hUnicoWS,'GetEnhMetaFileDescriptionW');
      GetEnhMetaFileW := GetProcAddress(hUnicoWS,'GetEnhMetaFileW');
      GetEnvironmentStringsW := GetProcAddress(hUnicoWS,'GetEnvironmentStringsW');
      GetEnvironmentVariableW := GetProcAddress(hUnicoWS,'GetEnvironmentVariableW');
      GetFileAttributesExW := GetProcAddress(hUnicoWS,'GetFileAttributesExW');
      GetFileAttributesW := GetProcAddress(hUnicoWS,'GetFileAttributesW');
      GetFileTitleW := GetProcAddress(hUnicoWS,'GetFileTitleW');
      GetFileVersionInfoSizeW := GetProcAddress(hUnicoWS,'GetFileVersionInfoSizeW');
      GetFileVersionInfoW := GetProcAddress(hUnicoWS,'GetFileVersionInfoW');
      GetFullPathNameW := GetProcAddress(hUnicoWS,'GetFullPathNameW');
      GetGlyphOutlineW := GetProcAddress(hUnicoWS,'GetGlyphOutlineW');
      GetICMProfileW := GetProcAddress(hUnicoWS,'GetICMProfileW');
      GetJobW := GetProcAddress(hUnicoWS,'GetJobW');
{$IFDEF UNIJWA}
      GetKerningPairsW := GetProcAddress(hUnicoWS,'GetKerningPairsW');
{$ENDIF}
      GetKeyNameTextW := GetProcAddress(hUnicoWS,'GetKeyNameTextW');
      GetKeyboardLayoutNameW := GetProcAddress(hUnicoWS,'GetKeyboardLayoutNameW');
      GetLocaleInfoW := GetProcAddress(hUnicoWS,'GetLocaleInfoW');
      GetLogColorSpaceW := GetProcAddress(hUnicoWS,'GetLogColorSpaceW');
      GetLogicalDriveStringsW := GetProcAddress(hUnicoWS,'GetLogicalDriveStringsW');
      GetLongPathNameW := GetProcAddress(hUnicoWS,'GetLongPathNameW');
      GetMenuItemInfoW := GetProcAddress(hUnicoWS,'GetMenuItemInfoW');
      GetMenuStringW := GetProcAddress(hUnicoWS,'GetMenuStringW');
      GetMessageW := GetProcAddress(hUnicoWS,'GetMessageW');
      GetMetaFileW := GetProcAddress(hUnicoWS,'GetMetaFileW');
      GetModuleFileNameW := GetProcAddress(hUnicoWS,'GetModuleFileNameW');
      GetModuleHandleW := GetProcAddress(hUnicoWS,'GetModuleHandleW');
      GetMonitorInfoW := GetProcAddress(hUnicoWS,'GetMonitorInfoW');
      GetNamedPipeHandleStateW := GetProcAddress(hUnicoWS,'GetNamedPipeHandleStateW');
      GetNumberFormatW := GetProcAddress(hUnicoWS,'GetNumberFormatW');
      GetObjectW := GetProcAddress(hUnicoWS,'GetObjectW');
      GetOpenFileNamePreviewW := GetProcAddress(hUnicoWS,'GetOpenFileNamePreviewW');
      GetOpenFileNameW := GetProcAddress(hUnicoWS,'GetOpenFileNameW');
      GetOutlineTextMetricsW := GetProcAddress(hUnicoWS,'GetOutlineTextMetricsW');
      GetPrintProcessorDirectoryW := GetProcAddress(hUnicoWS,'GetPrintProcessorDirectoryW');
      GetPrinterDataW := GetProcAddress(hUnicoWS,'GetPrinterDataW');
      GetPrinterDriverDirectoryW := GetProcAddress(hUnicoWS,'GetPrinterDriverDirectoryW');
      GetPrinterDriverW := GetProcAddress(hUnicoWS,'GetPrinterDriverW');
      GetPrinterW := GetProcAddress(hUnicoWS,'GetPrinterW');
      GetPrivateProfileIntW := GetProcAddress(hUnicoWS,'GetPrivateProfileIntW');
      GetPrivateProfileSectionNamesW := GetProcAddress(hUnicoWS,'GetPrivateProfileSectionNamesW');
      GetPrivateProfileSectionW := GetProcAddress(hUnicoWS,'GetPrivateProfileSectionW');
      GetPrivateProfileStringW := GetProcAddress(hUnicoWS,'GetPrivateProfileStringW');
      GetPrivateProfileStructW := GetProcAddress(hUnicoWS,'GetPrivateProfileStructW');
      GetProfileIntW := GetProcAddress(hUnicoWS,'GetProfileIntW');
      GetProfileSectionW := GetProcAddress(hUnicoWS,'GetProfileSectionW');
      GetProfileStringW := GetProcAddress(hUnicoWS,'GetProfileStringW');
      GetPropA := GetProcAddress(hUnicoWS,'GetPropA');
      GetPropW := GetProcAddress(hUnicoWS,'GetPropW');
      GetRoleTextW := GetProcAddress(hUnicoWS,'GetRoleTextW');
      GetSaveFileNamePreviewW := GetProcAddress(hUnicoWS,'GetSaveFileNamePreviewW');
      GetSaveFileNameW := GetProcAddress(hUnicoWS,'GetSaveFileNameW');
      GetShortPathNameW := GetProcAddress(hUnicoWS,'GetShortPathNameW');
      GetStartupInfoW := GetProcAddress(hUnicoWS,'GetStartupInfoW');
      GetStateTextW := GetProcAddress(hUnicoWS,'GetStateTextW');
      GetStringTypeExW := GetProcAddress(hUnicoWS,'GetStringTypeExW');
      GetStringTypeW := GetProcAddress(hUnicoWS,'GetStringTypeW');
      GetSystemDirectoryW := GetProcAddress(hUnicoWS,'GetSystemDirectoryW');
      GetSystemWindowsDirectoryW := GetProcAddress(hUnicoWS,'GetSystemWindowsDirectoryW');
      GetTabbedTextExtentW := GetProcAddress(hUnicoWS,'GetTabbedTextExtentW');
      GetTempFileNameW := GetProcAddress(hUnicoWS,'GetTempFileNameW');
      GetTempPathW := GetProcAddress(hUnicoWS,'GetTempPathW');
      GetTextExtentExPointW := GetProcAddress(hUnicoWS,'GetTextExtentExPointW');
      GetTextExtentPoint32W := GetProcAddress(hUnicoWS,'GetTextExtentPoint32W');
      GetTextExtentPointW := GetProcAddress(hUnicoWS,'GetTextExtentPointW');
      GetTextFaceW := GetProcAddress(hUnicoWS,'GetTextFaceW');
      GetTextMetricsW := GetProcAddress(hUnicoWS,'GetTextMetricsW');
      GetTimeFormatW := GetProcAddress(hUnicoWS,'GetTimeFormatW');
      GetUserNameW := GetProcAddress(hUnicoWS,'GetUserNameW');
      GetVersionExW := GetProcAddress(hUnicoWS,'GetVersionExW');
      GetVolumeInformationW := GetProcAddress(hUnicoWS,'GetVolumeInformationW');
      GetWindowLongA := GetProcAddress(hUnicoWS,'GetWindowLongA');
      GetWindowLongW := GetProcAddress(hUnicoWS,'GetWindowLongW');
      GetWindowModuleFileNameW := GetProcAddress(hUnicoWS,'GetWindowModuleFileNameW');
      GetWindowTextLengthW := GetProcAddress(hUnicoWS,'GetWindowTextLengthW');
      GetWindowTextW := GetProcAddress(hUnicoWS,'GetWindowTextW');
      GetWindowsDirectoryW := GetProcAddress(hUnicoWS,'GetWindowsDirectoryW');
      GlobalAddAtomW := GetProcAddress(hUnicoWS,'GlobalAddAtomW');
      GlobalFindAtomW := GetProcAddress(hUnicoWS,'GlobalFindAtomW');
      GlobalGetAtomNameW := GetProcAddress(hUnicoWS,'GlobalGetAtomNameW');
      GrayStringW := GetProcAddress(hUnicoWS,'GrayStringW');
{$IFDEF UNIJWA}
      InitSecurityInterfaceW := GetProcAddress(hUnicoWS,'InitSecurityInterfaceW');
      InitializeSecurityContextW := GetProcAddress(hUnicoWS,'InitializeSecurityContextW');
{$ENDIF}
      InsertMenuItemW := GetProcAddress(hUnicoWS,'InsertMenuItemW');
      InsertMenuW := GetProcAddress(hUnicoWS,'InsertMenuW');
      IsBadStringPtrW := GetProcAddress(hUnicoWS,'IsBadStringPtrW');
      IsCharAlphaNumericW := GetProcAddress(hUnicoWS,'IsCharAlphaNumericW');
      IsCharAlphaW := GetProcAddress(hUnicoWS,'IsCharAlphaW');
      IsCharLowerW := GetProcAddress(hUnicoWS,'IsCharLowerW');
      IsCharUpperW := GetProcAddress(hUnicoWS,'IsCharUpperW');
      IsClipboardFormatAvailable := GetProcAddress(hUnicoWS,'IsClipboardFormatAvailable');
      IsDestinationReachableW := GetProcAddress(hUnicoWS,'IsDestinationReachableW');
      IsDialogMessageW := GetProcAddress(hUnicoWS,'IsDialogMessageW');
      IsTextUnicode := GetProcAddress(hUnicoWS,'IsTextUnicode');
      IsValidCodePage := GetProcAddress(hUnicoWS,'IsValidCodePage');
      IsWindowUnicode := GetProcAddress(hUnicoWS,'IsWindowUnicode');
      LCMapStringW := GetProcAddress(hUnicoWS,'LCMapStringW');
      LoadAcceleratorsW := GetProcAddress(hUnicoWS,'LoadAcceleratorsW');
      LoadBitmapW := GetProcAddress(hUnicoWS,'LoadBitmapW');
      LoadCursorFromFileW := GetProcAddress(hUnicoWS,'LoadCursorFromFileW');
      LoadCursorW := GetProcAddress(hUnicoWS,'LoadCursorW');
      LoadIconW := GetProcAddress(hUnicoWS,'LoadIconW');
      LoadImageW := GetProcAddress(hUnicoWS,'LoadImageW');
      LoadKeyboardLayoutW := GetProcAddress(hUnicoWS,'LoadKeyboardLayoutW');
      LoadLibraryExW := GetProcAddress(hUnicoWS,'LoadLibraryExW');
      LoadLibraryW := GetProcAddress(hUnicoWS,'LoadLibraryW');
      LoadMenuIndirectW := GetProcAddress(hUnicoWS,'LoadMenuIndirectW');
      LoadMenuW := GetProcAddress(hUnicoWS,'LoadMenuW');
      LoadStringW := GetProcAddress(hUnicoWS,'LoadStringW');
      MCIWndCreateW := GetProcAddress(hUnicoWS,'MCIWndCreateW');
      MapVirtualKeyExW := GetProcAddress(hUnicoWS,'MapVirtualKeyExW');
      MapVirtualKeyW := GetProcAddress(hUnicoWS,'MapVirtualKeyW');
      MessageBoxExW := GetProcAddress(hUnicoWS,'MessageBoxExW');
      MessageBoxIndirectW := GetProcAddress(hUnicoWS,'MessageBoxIndirectW');
      MessageBoxW := GetProcAddress(hUnicoWS,'MessageBoxW');
      ModifyMenuW := GetProcAddress(hUnicoWS,'ModifyMenuW');
      MoveFileW := GetProcAddress(hUnicoWS,'MoveFileW');
      MultiByteToWideChar := GetProcAddress(hUnicoWS,'MultiByteToWideChar');
      MultinetGetConnectionPerformanceW := GetProcAddress(hUnicoWS,'MultinetGetConnectionPerformanceW');
      OemToCharBuffW := GetProcAddress(hUnicoWS,'OemToCharBuffW');
      OemToCharW := GetProcAddress(hUnicoWS,'OemToCharW');
      OleUIAddVerbMenuW := GetProcAddress(hUnicoWS,'OleUIAddVerbMenuW');
      OleUIBusyW := GetProcAddress(hUnicoWS,'OleUIBusyW');
      OleUIChangeIconW := GetProcAddress(hUnicoWS,'OleUIChangeIconW');
      OleUIChangeSourceW := GetProcAddress(hUnicoWS,'OleUIChangeSourceW');
      OleUIConvertW := GetProcAddress(hUnicoWS,'OleUIConvertW');
      OleUIEditLinksW := GetProcAddress(hUnicoWS,'OleUIEditLinksW');
      OleUIInsertObjectW := GetProcAddress(hUnicoWS,'OleUIInsertObjectW');
      OleUIObjectPropertiesW := GetProcAddress(hUnicoWS,'OleUIObjectPropertiesW');
      OleUIPasteSpecialW := GetProcAddress(hUnicoWS,'OleUIPasteSpecialW');
      OleUIPromptUserW := GetProcAddress(hUnicoWS,'OleUIPromptUserW');
      OleUIUpdateLinksW := GetProcAddress(hUnicoWS,'OleUIUpdateLinksW');
      OpenEventW := GetProcAddress(hUnicoWS,'OpenEventW');
      OpenFileMappingW := GetProcAddress(hUnicoWS,'OpenFileMappingW');
      OpenMutexW := GetProcAddress(hUnicoWS,'OpenMutexW');
      OpenPrinterW := GetProcAddress(hUnicoWS,'OpenPrinterW');
      OpenSemaphoreW := GetProcAddress(hUnicoWS,'OpenSemaphoreW');
      OpenWaitableTimerW := GetProcAddress(hUnicoWS,'OpenWaitableTimerW');
      OutputDebugStringW := GetProcAddress(hUnicoWS,'OutputDebugStringW');
      PageSetupDlgW := GetProcAddress(hUnicoWS,'PageSetupDlgW');
      PeekConsoleInputW := GetProcAddress(hUnicoWS,'PeekConsoleInputW');
      PeekMessageW := GetProcAddress(hUnicoWS,'PeekMessageW');
      PlaySoundW := GetProcAddress(hUnicoWS,'PlaySoundW');
      PolyTextOutW := GetProcAddress(hUnicoWS,'PolyTextOutW');
      PostMessageW := GetProcAddress(hUnicoWS,'PostMessageW');
      PostThreadMessageW := GetProcAddress(hUnicoWS,'PostThreadMessageW');
      PrintDlgW := GetProcAddress(hUnicoWS,'PrintDlgW');
{$IFDEF UNIJWA}
      QueryContextAttributesW := GetProcAddress(hUnicoWS,'QueryContextAttributesW');
      QueryCredentialsAttributesW := GetProcAddress(hUnicoWS,'QueryCredentialsAttributesW');
{$ENDIF}
      QueryDosDeviceW := GetProcAddress(hUnicoWS,'QueryDosDeviceW');
{$IFDEF UNIJWA}
      QuerySecurityPackageInfoW := GetProcAddress(hUnicoWS,'QuerySecurityPackageInfoW');
{$ENDIF}
{$IFDEF UNIJRA}
      RasConnectionNotificationW := GetProcAddress(hUnicoWS,'RasConnectionNotificationW');
      RasCreatePhonebookEntryW := GetProcAddress(hUnicoWS,'RasCreatePhonebookEntryW');
      RasDeleteEntryW := GetProcAddress(hUnicoWS,'RasDeleteEntryW');
      RasDeleteSubEntryW := GetProcAddress(hUnicoWS,'RasDeleteSubEntryW');
      RasDialW := GetProcAddress(hUnicoWS,'RasDialW');
      RasEditPhonebookEntryW := GetProcAddress(hUnicoWS,'RasEditPhonebookEntryW');
      RasEnumConnectionsW := GetProcAddress(hUnicoWS,'RasEnumConnectionsW');
      RasEnumDevicesW := GetProcAddress(hUnicoWS,'RasEnumDevicesW');
      RasEnumEntriesW := GetProcAddress(hUnicoWS,'RasEnumEntriesW');
      RasGetConnectStatusW := GetProcAddress(hUnicoWS,'RasGetConnectStatusW');
      RasGetEntryDialParamsW := GetProcAddress(hUnicoWS,'RasGetEntryDialParamsW');
      RasGetEntryPropertiesW := GetProcAddress(hUnicoWS,'RasGetEntryPropertiesW');
      RasGetErrorStringW := GetProcAddress(hUnicoWS,'RasGetErrorStringW');
      RasGetProjectionInfoW := GetProcAddress(hUnicoWS,'RasGetProjectionInfoW'); // 3703
      RasHangUpW := GetProcAddress(hUnicoWS,'RasHangUpW');
      RasRenameEntryW := GetProcAddress(hUnicoWS,'RasRenameEntryW');
      RasSetEntryDialParamsW := GetProcAddress(hUnicoWS,'RasSetEntryDialParamsW');
      RasSetEntryPropertiesW := GetProcAddress(hUnicoWS,'RasSetEntryPropertiesW');
      RasSetSubEntryPropertiesW := GetProcAddress(hUnicoWS,'RasSetSubEntryPropertiesW');
      RasValidateEntryNameW := GetProcAddress(hUnicoWS,'RasValidateEntryNameW');
{$ENDIF}
      ReadConsoleInputW := GetProcAddress(hUnicoWS,'ReadConsoleInputW');
      ReadConsoleOutputCharacterW := GetProcAddress(hUnicoWS,'ReadConsoleOutputCharacterW');
      ReadConsoleOutputW := GetProcAddress(hUnicoWS,'ReadConsoleOutputW');
      ReadConsoleW := GetProcAddress(hUnicoWS,'ReadConsoleW');
      RegConnectRegistryW := GetProcAddress(hUnicoWS,'RegConnectRegistryW');
      RegCreateKeyExW := GetProcAddress(hUnicoWS,'RegCreateKeyExW');
      RegCreateKeyW := GetProcAddress(hUnicoWS,'RegCreateKeyW');
      RegDeleteKeyW := GetProcAddress(hUnicoWS,'RegDeleteKeyW');
      RegDeleteValueW := GetProcAddress(hUnicoWS,'RegDeleteValueW');
      RegEnumKeyExW := GetProcAddress(hUnicoWS,'RegEnumKeyExW');
      RegEnumKeyW := GetProcAddress(hUnicoWS,'RegEnumKeyW');
      RegEnumValueW := GetProcAddress(hUnicoWS,'RegEnumValueW');
      RegLoadKeyW := GetProcAddress(hUnicoWS,'RegLoadKeyW');
      RegOpenKeyExW := GetProcAddress(hUnicoWS,'RegOpenKeyExW');
      RegOpenKeyW := GetProcAddress(hUnicoWS,'RegOpenKeyW');
      RegQueryInfoKeyW := GetProcAddress(hUnicoWS,'RegQueryInfoKeyW');
      RegQueryMultipleValuesW := GetProcAddress(hUnicoWS,'RegQueryMultipleValuesW');
      RegQueryValueExW := GetProcAddress(hUnicoWS,'RegQueryValueExW');
      RegQueryValueW := GetProcAddress(hUnicoWS,'RegQueryValueW');
      RegReplaceKeyW := GetProcAddress(hUnicoWS,'RegReplaceKeyW');
      RegSaveKeyW := GetProcAddress(hUnicoWS,'RegSaveKeyW');
      RegSetValueExW := GetProcAddress(hUnicoWS,'RegSetValueExW');
      RegSetValueW := GetProcAddress(hUnicoWS,'RegSetValueW');
      RegUnLoadKeyW := GetProcAddress(hUnicoWS,'RegUnLoadKeyW');
      RegisterClassExW := GetProcAddress(hUnicoWS,'RegisterClassExW');
      RegisterClassW := GetProcAddress(hUnicoWS,'RegisterClassW');
      RegisterClipboardFormatW := GetProcAddress(hUnicoWS,'RegisterClipboardFormatW');
      RegisterDeviceNotificationW := GetProcAddress(hUnicoWS,'RegisterDeviceNotificationW');
      RegisterWindowMessageW := GetProcAddress(hUnicoWS,'RegisterWindowMessageW');
      RemoveDirectoryW := GetProcAddress(hUnicoWS,'RemoveDirectoryW');
      RemoveFontResourceW := GetProcAddress(hUnicoWS,'RemoveFontResourceW');
      RemovePropA := GetProcAddress(hUnicoWS,'RemovePropA');
      RemovePropW := GetProcAddress(hUnicoWS,'RemovePropW');
      ReplaceTextW := GetProcAddress(hUnicoWS,'ReplaceTextW');
      ResetDCW := GetProcAddress(hUnicoWS,'ResetDCW');
      ResetPrinterW := GetProcAddress(hUnicoWS,'ResetPrinterW');
      SHBrowseForFolderW := GetProcAddress(hUnicoWS,'SHBrowseForFolderW');
      SHChangeNotify := GetProcAddress(hUnicoWS,'SHChangeNotify');
      SHFileOperationW := GetProcAddress(hUnicoWS,'SHFileOperationW');
      SHGetFileInfoW := GetProcAddress(hUnicoWS,'SHGetFileInfoW');
      SHGetNewLinkInfoW := GetProcAddress(hUnicoWS,'SHGetNewLinkInfoW');
      SHGetPathFromIDListW := GetProcAddress(hUnicoWS,'SHGetPathFromIDListW');
      ScrollConsoleScreenBufferW := GetProcAddress(hUnicoWS,'ScrollConsoleScreenBufferW');
      SearchPathW := GetProcAddress(hUnicoWS,'SearchPathW');
      SendDlgItemMessageW := GetProcAddress(hUnicoWS,'SendDlgItemMessageW');
      SendMessageCallbackW := GetProcAddress(hUnicoWS,'SendMessageCallbackW');
      SendMessageTimeoutW := GetProcAddress(hUnicoWS,'SendMessageTimeoutW');
      SendMessageW := GetProcAddress(hUnicoWS,'SendMessageW');
      SendNotifyMessageW := GetProcAddress(hUnicoWS,'SendNotifyMessageW');
      SetCalendarInfoW := GetProcAddress(hUnicoWS,'SetCalendarInfoW');
      SetClassLongW := GetProcAddress(hUnicoWS,'SetClassLongW');
      SetComputerNameW := GetProcAddress(hUnicoWS,'SetComputerNameW');
      SetConsoleTitleW := GetProcAddress(hUnicoWS,'SetConsoleTitleW');
      SetCurrentDirectoryW := GetProcAddress(hUnicoWS,'SetCurrentDirectoryW');
      SetDefaultCommConfigW := GetProcAddress(hUnicoWS,'SetDefaultCommConfigW');
      SetDlgItemTextW := GetProcAddress(hUnicoWS,'SetDlgItemTextW');
      SetEnvironmentVariableW := GetProcAddress(hUnicoWS,'SetEnvironmentVariableW');
      SetFileAttributesW := GetProcAddress(hUnicoWS,'SetFileAttributesW');
      SetICMProfileW := GetProcAddress(hUnicoWS,'SetICMProfileW');
      SetJobW := GetProcAddress(hUnicoWS,'SetJobW');
      SetLocaleInfoW := GetProcAddress(hUnicoWS,'SetLocaleInfoW');
      SetMenuItemInfoW := GetProcAddress(hUnicoWS,'SetMenuItemInfoW');
      SetPrinterDataW := GetProcAddress(hUnicoWS,'SetPrinterDataW');
      SetPrinterW := GetProcAddress(hUnicoWS,'SetPrinterW');
      SetPropA := GetProcAddress(hUnicoWS,'SetPropA');
      SetPropW := GetProcAddress(hUnicoWS,'SetPropW');
      SetVolumeLabelW := GetProcAddress(hUnicoWS,'SetVolumeLabelW');
      SetWindowLongA := GetProcAddress(hUnicoWS,'SetWindowLongA');
      SetWindowLongW := GetProcAddress(hUnicoWS,'SetWindowLongW');
      SetWindowTextW := GetProcAddress(hUnicoWS,'SetWindowTextW');
      SetWindowsHookExW := GetProcAddress(hUnicoWS,'SetWindowsHookExW');
      SetWindowsHookW := GetProcAddress(hUnicoWS,'SetWindowsHookW');
      ShellAboutW := GetProcAddress(hUnicoWS,'ShellAboutW');
      ShellExecuteExW := GetProcAddress(hUnicoWS,'ShellExecuteExW');
      ShellExecuteW := GetProcAddress(hUnicoWS,'ShellExecuteW');
      Shell_NotifyIconW := GetProcAddress(hUnicoWS,'Shell_NotifyIconW');
      StartDocPrinterW := GetProcAddress(hUnicoWS,'StartDocPrinterW');
      StartDocW := GetProcAddress(hUnicoWS,'StartDocW');
      SystemParametersInfoW := GetProcAddress(hUnicoWS,'SystemParametersInfoW');
      TabbedTextOutW := GetProcAddress(hUnicoWS,'TabbedTextOutW');
      TextOutW := GetProcAddress(hUnicoWS,'TextOutW');
      TranslateAcceleratorW := GetProcAddress(hUnicoWS,'TranslateAcceleratorW');
      UnregisterClassW := GetProcAddress(hUnicoWS,'UnregisterClassW');
      UpdateICMRegKeyW := GetProcAddress(hUnicoWS,'UpdateICMRegKeyW');
      UpdateResourceA := GetProcAddress(hUnicoWS,'UpdateResourceA'); // 3703
      UpdateResourceW := GetProcAddress(hUnicoWS,'UpdateResourceW');
      VerFindFileW := GetProcAddress(hUnicoWS,'VerFindFileW');
      VerInstallFileW := GetProcAddress(hUnicoWS,'VerInstallFileW');
      VerLanguageNameW := GetProcAddress(hUnicoWS,'VerLanguageNameW');
      VerQueryValueW := GetProcAddress(hUnicoWS,'VerQueryValueW');
      VkKeyScanExW := GetProcAddress(hUnicoWS,'VkKeyScanExW');
      VkKeyScanW := GetProcAddress(hUnicoWS,'VkKeyScanW');
      WNetAddConnection2W := GetProcAddress(hUnicoWS,'WNetAddConnection2W');
      WNetAddConnection3W := GetProcAddress(hUnicoWS,'WNetAddConnection3W');
      WNetAddConnectionW := GetProcAddress(hUnicoWS,'WNetAddConnectionW');
      WNetCancelConnection2W := GetProcAddress(hUnicoWS,'WNetCancelConnection2W');
      WNetCancelConnectionW := GetProcAddress(hUnicoWS,'WNetCancelConnectionW');
      WNetConnectionDialog1W := GetProcAddress(hUnicoWS,'WNetConnectionDialog1W');
      WNetDisconnectDialog1W := GetProcAddress(hUnicoWS,'WNetDisconnectDialog1W');
      WNetEnumResourceW := GetProcAddress(hUnicoWS,'WNetEnumResourceW');
      WNetGetConnectionW := GetProcAddress(hUnicoWS,'WNetGetConnectionW');
      WNetGetLastErrorW := GetProcAddress(hUnicoWS,'WNetGetLastErrorW');
      WNetGetNetworkInformationW := GetProcAddress(hUnicoWS,'WNetGetNetworkInformationW');
      WNetGetProviderNameW := GetProcAddress(hUnicoWS,'WNetGetProviderNameW');
      WNetGetResourceInformationW := GetProcAddress(hUnicoWS,'WNetGetResourceInformationW');
      WNetGetResourceParentW := GetProcAddress(hUnicoWS,'WNetGetResourceParentW');
      WNetGetUniversalNameW := GetProcAddress(hUnicoWS,'WNetGetUniversalNameW');
      WNetGetUserW := GetProcAddress(hUnicoWS,'WNetGetUserW');
      WNetOpenEnumW := GetProcAddress(hUnicoWS,'WNetOpenEnumW');
      WNetUseConnectionW := GetProcAddress(hUnicoWS,'WNetUseConnectionW');
      WaitNamedPipeW := GetProcAddress(hUnicoWS,'WaitNamedPipeW');
      WideCharToMultiByte := GetProcAddress(hUnicoWS,'WideCharToMultiByte');
      WinHelpW := GetProcAddress(hUnicoWS,'WinHelpW');
      WriteConsoleInputW := GetProcAddress(hUnicoWS,'WriteConsoleInputW');
      WriteConsoleOutputCharacterW := GetProcAddress(hUnicoWS,'WriteConsoleOutputCharacterW');
      WriteConsoleOutputW := GetProcAddress(hUnicoWS,'WriteConsoleOutputW');
      WriteConsoleW := GetProcAddress(hUnicoWS,'WriteConsoleW');
      WritePrivateProfileSectionW := GetProcAddress(hUnicoWS,'WritePrivateProfileSectionW');
      WritePrivateProfileStringW := GetProcAddress(hUnicoWS,'WritePrivateProfileStringW');
      WritePrivateProfileStructW := GetProcAddress(hUnicoWS,'WritePrivateProfileStructW');
      WriteProfileSectionW := GetProcAddress(hUnicoWS,'WriteProfileSectionW');
      WriteProfileStringW := GetProcAddress(hUnicoWS,'WriteProfileStringW');
      auxGetDevCapsW := GetProcAddress(hUnicoWS,'auxGetDevCapsW');
      capCreateCaptureWindowW := GetProcAddress(hUnicoWS,'capCreateCaptureWindowW');
      capGetDriverDescriptionW := GetProcAddress(hUnicoWS,'capGetDriverDescriptionW');
      joyGetDevCapsW := GetProcAddress(hUnicoWS,'joyGetDevCapsW');
      lstrcatW := GetProcAddress(hUnicoWS,'lstrcatW');
      lstrcmpW := GetProcAddress(hUnicoWS,'lstrcmpW');
      lstrcmpiW := GetProcAddress(hUnicoWS,'lstrcmpiW');
      lstrcpyW := GetProcAddress(hUnicoWS,'lstrcpyW');
      lstrcpynW := GetProcAddress(hUnicoWS,'lstrcpynW');
      lstrlenW := GetProcAddress(hUnicoWS,'lstrlenW');
      mciGetDeviceIDW := GetProcAddress(hUnicoWS,'mciGetDeviceIDW');
      mciGetErrorStringW := GetProcAddress(hUnicoWS,'mciGetErrorStringW');
      mciSendCommandW := GetProcAddress(hUnicoWS,'mciSendCommandW');
      mciSendStringW := GetProcAddress(hUnicoWS,'mciSendStringW');
      midiInGetDevCapsW := GetProcAddress(hUnicoWS,'midiInGetDevCapsW');
      midiInGetErrorTextW := GetProcAddress(hUnicoWS,'midiInGetErrorTextW');
      midiOutGetDevCapsW := GetProcAddress(hUnicoWS,'midiOutGetDevCapsW');
      midiOutGetErrorTextW := GetProcAddress(hUnicoWS,'midiOutGetErrorTextW');
      mixerGetControlDetailsW := GetProcAddress(hUnicoWS,'mixerGetControlDetailsW');
      mixerGetDevCapsW := GetProcAddress(hUnicoWS,'mixerGetDevCapsW');
      mixerGetLineControlsW := GetProcAddress(hUnicoWS,'mixerGetLineControlsW');
      mixerGetLineInfoW := GetProcAddress(hUnicoWS,'mixerGetLineInfoW');
      mmioInstallIOProcW := GetProcAddress(hUnicoWS,'mmioInstallIOProcW');
      mmioOpenW := GetProcAddress(hUnicoWS,'mmioOpenW');
      mmioRenameW := GetProcAddress(hUnicoWS,'mmioRenameW');
      mmioStringToFOURCCW := GetProcAddress(hUnicoWS,'mmioStringToFOURCCW');
      sndPlaySoundW := GetProcAddress(hUnicoWS,'sndPlaySoundW');
      waveInGetDevCapsW := GetProcAddress(hUnicoWS,'waveInGetDevCapsW');
      waveInGetErrorTextW := GetProcAddress(hUnicoWS,'waveInGetErrorTextW');
      waveOutGetDevCapsW := GetProcAddress(hUnicoWS,'waveOutGetDevCapsW');
      waveOutGetErrorTextW := GetProcAddress(hUnicoWS,'waveOutGetErrorTextW');
      wsprintfW := GetProcAddress(hUnicoWS,'wsprintfW');
      wvsprintfW := GetProcAddress(hUnicoWS,'wvsprintfW');
    end
  end
end;

initialization
  LoadMSLU;
finalization
  if hUnicoWS <> 0 then FreeLibrary(hUnicoWS);
end.
